// (C) Folkert van Heusden. AGPL v3.0.
// $Revision: 4635 $
#include "Tweaks.h"
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "Debug.h"
#include "Exception.h"
#include "Board.h"
#include "Utils.h"

char REV_BOARD[] = "Board $Revision: 4635 $";

Board::Board()
{
	clearBoard();
}

Board::Board(const Board & bIn)
{
	memcpy(b, bIn.b, sizeof b);

	const unsigned long int bs_size = bIn.bs.size();

	for(unsigned long int idx=0; idx<bs_size; idx++)
		bs.push_back(bIn.bs.at(idx));
}

Board::~Board()
{
}

void Board::clearBoard()
{
	memset(b, 0x00, sizeof b);
}

void Board::replace(ChessPiece *const piece)
{
	piece -> validate();

	int x, y;
	piece -> getXY(&x, &y);

	my_assert(piece -> getAlive(), "Board::replace: must be alive");
	my_assert(b[y][x] != NULL, "Board::replace: must not be empty");

	b[y][x] = piece;
}

void Board::move(const int fx, const int fy, const int tx, const int ty)
{
	my_assert(b[fy][fx] != NULL, "Board::move: from must not be empty");
	my_assert(b[fy][fx] -> getAlive(), "Board::move: from must be alive");
	my_assert(fx != tx || fy != ty, "Board::move: destination must differ from source");
	my_assert(b[ty][tx] == NULL || b[ty][tx] -> getColor() != b[fy][fx] -> getColor(), "Board::move must be either empty or different color");
	my_assert(b[ty][tx] == NULL || b[ty][tx] -> getType() != KING, "Board::move target piece cannot be king");
	// cannot do verify here which checks if piece-location is same as real location
	// as this method is part of the moving process

	b[ty][tx] = b[fy][fx];
	b[fy][fx] = NULL;
}

const ChessPiece *Board::findVictim(const Move & m) const
{
	my_assert(isValidMove(m), "findVictim");
	my_assert(m.u.m.isCatchMove, "victim sanity");
	my_assert(m.u.m.mt == REGULAR || m.u.m.mt == EN_PASSANT, "victim sanity");

	if (likely(m.u.m.mt == REGULAR))
		return b[m.u.m.ty][m.u.m.tx];

	/* must be en-passant */
	const int delX = m.u.m.tx;
	const int delY = m.u.m.ty == 5 ? 4 : 3;

	return b[delY][delX];
}

void Board::doMove(const Move & m, BoardState *const cbs)
{
	my_assert(isValidMove(m), "doMove");
	my_assert(b[m.u.m.fy][m.u.m.fx] != NULL, "Board::doMove from must not be empty");
	my_assert(b[m.u.m.fy][m.u.m.fx] -> getX() == m.u.m.fx, "Board::doMove: x of piece == x of board");
	my_assert(b[m.u.m.fy][m.u.m.fx] -> getY() == m.u.m.fy, "Board::doMove: y of piece == y of board");

	cbs -> mt = m.u.m.mt;
	cbs -> isPromotion = false;

	ChessPiece *fromPiece = b[m.u.m.fy][m.u.m.fx];
	my_assert(fromPiece != NULL, "Board::doMove source field cannot be empty");

	ChessPiece *toPiece = b[m.u.m.ty][m.u.m.tx];
	my_assert(toPiece == NULL || toPiece -> getColor() != fromPiece -> getColor(), "Board::doMove must be either empty or different color");
	my_assert(toPiece == NULL || toPiece -> getType() != KING, "Board::doMove target piece cannot be king");

	my_assert(toPiece == NULL || m.u.m.isCatchMove, "Board::doMove toPiece != NULL, then isCatchMove must be true");

	switch(m.u.m.mt)
	{
		case REGULAR:
			cbs -> fx = m.u.m.fx;
			cbs -> fy = m.u.m.fy;
			cbs -> p1 = b[m.u.m.fy][m.u.m.fx];
			cbs -> tx = m.u.m.tx;
			cbs -> ty = m.u.m.ty;
			cbs -> p2 = b[m.u.m.ty][m.u.m.tx];

			if (toPiece != NULL)
			{
				toPiece -> pushState();

				toPiece -> doKill();
			}

			move(m.u.m.fx, m.u.m.fy, m.u.m.tx, m.u.m.ty);

			fromPiece -> pushState();
			fromPiece -> doMove(m.u.m.tx, m.u.m.ty);

			if (unlikely(m.u.m.promoteTo != NONE)) {
				my_assert(m.u.m.promoteTo != PAWN, "Board::doMove cannot promote to pawn");
				my_assert(m.u.m.promoteTo != NONE, "Board::doMove pawn on 1/8 must promote");

				fromPiece -> promote(m.u.m.promoteTo);
				cbs -> isPromotion = true;
			}

			break;

		case CASTLING:
			my_assert(m.u.m.tx == 2 || m.u.m.tx == 6, "Board::doMove parameter sanity 1");
			my_assert(m.u.m.fy == m.u.m.ty, "Board::doMove parameter sanity 2");

			cbs -> fx = 4;
			cbs -> fy = m.u.m.fy;
			cbs -> p1 = b[m.u.m.fy][4];
			b[m.u.m.fy][4] -> pushState();

			cbs -> ty = m.u.m.ty;

			if (m.u.m.tx == 6)	// SHORT
			{
				b[m.u.m.fy][4] -> doMove(6, m.u.m.fy);
				move(4, m.u.m.fy, 6, m.u.m.fy);

				cbs -> tx = 6; // SHORT CASTLING
				cbs -> p2 = b[m.u.m.ty][7];

				b[m.u.m.fy][7] -> pushState();
				b[m.u.m.fy][7] -> doMove(5, m.u.m.fy);
				move(7, m.u.m.fy, 5, m.u.m.fy);
			}
			else		// LONG (2)
			{
				b[m.u.m.fy][4] -> doMove(2, m.u.m.fy);
				move(4, m.u.m.fy, 2, m.u.m.fy);

				cbs -> tx = 2; // LONG CASTLING
				cbs -> p2 = b[m.u.m.ty][0];

				b[m.u.m.fy][0] -> pushState();
				b[m.u.m.fy][0] -> doMove(3, m.u.m.fy);
				move(0, m.u.m.fy, 3, m.u.m.fy);
			}
			break;

		case EN_PASSANT:
			{
				cbs -> fx = m.u.m.fx;
				cbs -> fy = m.u.m.fy;
				cbs -> p1 = b[m.u.m.fy][m.u.m.fx];
				b[m.u.m.fy][m.u.m.fx] -> pushState();

				int delX = m.u.m.tx;
				int delY = m.u.m.ty == 5 ? 4 : 3;

				cbs -> tx = delX;
				cbs -> ty = delY;
				cbs -> p2 = b[delY][delX];

				b[delY][delX] -> pushState();
				b[delY][delX] -> doKill();
				b[delY][delX] = NULL;

				fromPiece -> doMove(m.u.m.tx, m.u.m.ty);
				move(m.u.m.fx, m.u.m.fy, m.u.m.tx, m.u.m.ty);
			}
			break;

		case INVALID_MT:
		default:
			my_assert(false, "Board::doMove invalid move type");
			break;
	}

	bs.push_back(*cbs); 
}

void Board::undoMove()
{
	BoardState cbs = bs.back();
#ifdef _DEBUG
	bs.back().fx = bs.back().fy = -1;
	bs.back().tx = bs.back().ty = -1;
	bs.back().p1 = bs.back().p2 = NULL;
	bs.back().mt = INVALID_MT;
#endif
	bs.pop_back();

	switch(cbs.mt)
	{
		case REGULAR:
			b[cbs.fy][cbs.fx] = cbs.p1;
			b[cbs.ty][cbs.tx] = cbs.p2;
			break;

		case CASTLING:
			if (cbs.tx == 6)	// SHORT
			{
				move(6, cbs.fy, 4, cbs.fy);

				move(5, cbs.fy, 7, cbs.fy);
			}
			else		// LONG (2)
			{
				move(2, cbs.fy, 4, cbs.fy);

				move(3, cbs.fy, 0, cbs.fy);
			}
			break;

		case EN_PASSANT:
			b[cbs.p1 -> getY()][cbs.p1 -> getX()] = NULL;
			b[cbs.fy][cbs.fx] = cbs.p1;
			b[cbs.ty][cbs.tx] = cbs.p2;
			break;

		case INVALID_MT:
		default:
			my_assert(false, "Board::undoMove invalid move type");
			break;
	}

	cbs.p1 -> popState();

	if (cbs.p2)
		cbs.p2 -> popState();
}

void Board::setAt(const int x, const int y, ChessPiece *const p)
{
	p -> validate();
	my_assert(y >= 0 && y < 8, "Board::setAt: y within limits");
	my_assert(x >= 0 && x < 8, "Board::setAt: x within limits");

	b[y][x] = p;
}

void Board::put(ChessPiece * const p)
{
	p -> validate();
	my_assert(p -> getAlive(), "Board::put: piece to add alive");

	int x, y;
	p -> getXY(&x, &y);
	my_assert(y >= 0 && y < 8, "Board::put: y within limits");
	my_assert(x >= 0 && x < 8, "Board::put: x within limits");

	my_assert(b[y][x] == NULL, "Board::put: empty spot");

	b[y][x] = p;
}

bool isEpMove(const Board *const b, const Move & m)
{
	my_assert(isValidMove(m), "isEpMove");

	return b -> getAt(m.u.m.fx, m.u.m.fy) -> getType() == PAWN && abs(m.u.m.ty - m.u.m.fy) == 2;
}

void Board::verifyBoard() const
{
#ifdef _DEBUG
	int n = 0, nw = 0, nb = 0;

	for(int y=0; y<8; y++) {
		for(int x=0; x<8; x++) {
			if (!b[y][x])
				continue;

			b[y][x] -> validate();

			if (b[y][x] -> getColor() == WHITE)
				nw++;
			else
				nb++;
		}
	}

	my_assert(n <= 32, "verifyBoard");
	my_assert(nw <= 16, "verifyBoard");
	my_assert(nb <= 16, "verifyBoard");
#endif
}
