// (C) Folkert van Heusden. AGPL v3.0.
// $Revision: 4872 $
// WDL book
#include "Tweaks.h"
#include "config.h"
#include "Book.h"
#include "Utils.h"
#include "Move.h"
#include "MoveType.h"

Book::Book(IO *const ioIn) : io(ioIn), th(NULL), terminate(false), id(rand())
{
	rm_lock = alloc_rw_lock();
	id = 0;

	mt = new MERS_RNG(rd());
}

Book::~Book()
{
	terminate = true;
	workAdded.notify_one();

	if (th) {
		th -> join();
		delete th;
	}

	auto iq_it = results.begin();
	while(iq_it != results.end()) {
		delete iq_it -> second.pv;

		iq_it++;
	}

	free_rw_lock(rm_lock);

	delete mt;
}

int Book::getRandom(int max_excl)
{
	int mask = upper_power_of_two(max_excl) - 1;

	for(;;)
	{
		int v = (*mt)() & mask;

		if (v < max_excl)
			return v;
	}

	return -1;
}

int Book::ask(Scene & s, const PlayerColor c)
{
	int cur_id = id.fetch_add(1);

	// queue fen string
	std::string fen = genFen(&s, true, c);

	log(ll_ok, "\"%s\" scheduled for book look-up (id: %d)", fen.c_str(), cur_id);

	bquestion_t q = { fen, c };

	w_lock(rm_lock);
	in_queue.insert(std::pair<int, bquestion_t>(cur_id, q));
	w_ulock(rm_lock);

	workAdded.notify_one();

	return cur_id;
}

void Book::putResult(const int cur_id, const Move & m, const int eval, const bool isFinalAnswer, const bool valid)
{
	VectorType<PVnode_t> *ml = new VectorType<PVnode_t>();

	PVnode_t t = { 0, m, eval, MS_BOOK };
	ml -> push_back(t);

	putResult(cur_id, ml, isFinalAnswer, valid);
}

void Book::putResult(const int cur_id, VectorType<PVnode_t> *const ml, const bool isFinalAnswer, const bool valid)
{
	log(ll_ok, "BOOK: Storing %d", cur_id);
	w_lock(rm_lock);

	auto iq_it = in_queue.find(cur_id);
	if (iq_it != in_queue.end())
		in_queue.erase(iq_it);

	results.insert(std::pair<int, book_answer_t>(cur_id, { ml, isFinalAnswer }));

	w_ulock(rm_lock);
}

void Book::stopAllAsks()
{
	log(ll_ok, "Aborting all book searches");

	w_lock(rm_lock);
	in_queue.clear();

	auto re_it = results.begin();
	for(; re_it != results.end(); re_it++)
		delete re_it -> second.pv;
	results.clear();

	w_ulock(rm_lock);
}

book_result_t Book::getResult(const int idIn, const Scene *const s, const PlayerColor c, VectorType<PVnode_t> **const ml, bool *const is_final_answer)
{
	log(ll_ok, "BOOK: checking for %d", idIn);

	book_result_t br = BR_ERR;

	r_lock(rm_lock);

	auto iq_it = in_queue.find(idIn);
	auto re_it = results.find(idIn);

	if (iq_it != in_queue.end())
	{
		log(ll_ok, "Still in input-queue");
		br = BR_BUSY;
	}
	else if (re_it != results.end())
	{
		log(ll_ok, "Result found in result-queue");
		*ml = re_it -> second.pv;
		*is_final_answer = re_it -> second.is_final_answer;

		if (isValidMove((*ml) -> at(0).m))
			br = BR_OK;
		else
		{
			log(ll_ok, "Result found in result-queue indicates nothing found");
		 	br = BR_ERR;
		}
	}
	else
	{
		log(ll_err, "Not in input and not in output-queue");
		br = BR_BUSY;
	}

	r_ulock(rm_lock);

	if (br == BR_OK)
	{
		log(ll_ok, "Verify %s", pvToString(**ml, false, false, false).c_str());

		w_lock(rm_lock);
		results.erase(re_it);
		w_ulock(rm_lock);

		const Board *const b = s -> getBoardConst();
		VectorType<Move> *const mlVerify = s -> getMoveList(c);
		Move *const m = (Move *)(&(*ml) -> atPointer(0) -> m);

		if (b -> getAt(m -> u.m.fx, m -> u.m.fy) == NULL)
		{
			log(ll_err, "ERR: Source-field is empty");
			return BR_ERR;
		}

		if (b -> getAt(m -> u.m.tx, m -> u.m.ty))
		{
			m -> u.m.isCatchMove = true;

			if (b -> getAt(m -> u.m.tx, m -> u.m.ty) -> getType() == KING)
			{
				log(ll_err, "ERR: Target is king");
				return BR_ERR;
			}
		}

		if (m -> u.m.mt == REGULAR)
		{
			if (b -> getAt(m -> u.m.fx, m -> u.m.fy) -> getType() == PAWN)
			{
				if (abs(m -> u.m.fx - m -> u.m.tx) == 1 && (m -> u.m.ty == 2 || m -> u.m.ty == 5)) // possibly en-passant
				{
					ChessPiece *possibleVictim = b -> getAt(m -> u.m.tx, m -> u.m.fy);

					if (possibleVictim && possibleVictim -> getType() == PAWN)
					{
						log(ll_ok, "Move is an en-passant move");
						m -> u.m.mt = EN_PASSANT;
					}
				}
				// FIXME update move for promote
				else
				{
				}
			}
		}

		bool found = indexOf(*mlVerify, *m) != -1;
		if (!found)
		{
			log(ll_err, "Move (%s) NOT found! (in %s)", moveToString(*m, true).c_str(), toString(*mlVerify, false).c_str());
			br = BR_ERR;
		}
	}

	return br;
}

// default is do nothing
void Book::learn(const Board & b, const PlayerColor c, const Move & epMove, const Move & bestMove, const int depth, const winner_t win)
{
}

Move moveStringToMove(const std::string & str, const PlayerColor pc)
{
	Move m = invalidMove;

	if (str == "O-O" || str == "O-O-O" || str == "O-O+" || str == "O-O-O+")
	{
		if (pc == BLACK)
			m.u.m.fy = m.u.m.ty = 7;
		else
			m.u.m.fy = m.u.m.ty = 0;

		m.u.m.fx = 4;
		m.u.m.mt = CASTLING;

		m.u.m.promoteTo = NONE;

		m.u.m.isCatchMove = false;

		if (str == "O-O" || str == "O-O+")
			m.u.m.tx = 6;
		else
			m.u.m.tx = 2;
	}
	else
	{
		m.u.m.fx = str.at(0) - 'a';
		m.u.m.fy = str.at(1) - '1';
		m.u.m.tx = str.at(2) - 'a';
		m.u.m.ty = str.at(3) - '1';
		m.u.m.mt = REGULAR; // FIXME catch & en-passant

		m.u.m.promoteTo = NONE;
		if (str.size() >= 5) {
			if (str.at(4) == 'q')
				m.u.m.promoteTo = QUEEN;
			else if (str.at(4) == 'b')
				m.u.m.promoteTo = BISHOP;
			else if (str.at(4) == 'r')
				m.u.m.promoteTo = ROOK;
			else if (str.at(4) == 'n')
				m.u.m.promoteTo = KNIGHT;
		}

		m.u.m.isCatchMove = false;
	}

	return m;
}
