// (C) Folkert van Heusden. AGPL v3.0.
// $Revision: 5136 $
#include "Tweaks.h"
#include <assert.h>
#include <algorithm>
#include <string.h>
#include <thread>
#include <unistd.h>
#include <sys/resource.h>
#include <sys/time.h>

#include "config.h"
#include "Debug.h"
#include "Exception.h"
#include "Utils.h"
#include "Brain.h"
#include "Tpt.h"
#include "Move.h"
#include "IO_stderr.h"
#include "IO.h"
#include "MoveSort.h"
#include "ALG_position.h"
#include "Main.h"
#include "Rng.h"
#include "History.h"
#ifdef W_SYZYGY
#include "syzygy_etb/fathom.h"
#else
int get_syzygy_etb_piece_count() { return 0; }
#endif
#include "Tune.h"
#include "Gen.h"

char REV_BRAIN[] = "Brain $Revision: 5136 $";

constexpr int depth_sanity_limit = 127;
constexpr int value_checkmate = 10000, value_checkmate_min = 9800;
constexpr int infinite = INFINITE;
constexpr int maxNThreadsPerGroup = MAX_N_THREADS_PER_GROUP;

const Move invalidBestSibblings[] = { invalidMove, invalidMove };

std::string perftCountsToStr(const perft_counts_t *const pc)
{
	return myformat("captures: %lu, ep: %lu, castles: %lu, promotions: %lu, checks: %lu, checkmates: %lu, stalemates: %lu, insufficientmaterialdraw: %lu, white wins: %lu, black wins: %lu", pc -> captures, pc -> ep, pc -> castles, pc -> promotions, pc -> checks, pc -> checkmates, pc -> stalemates, pc -> imd, pc -> whiteWin, pc -> checkmates - pc -> whiteWin);
}

void sumPerftCounts(perft_counts_t *const tgt, const perft_counts_t *const src)
{
	tgt -> captures += src -> captures;
	tgt -> ep += src -> ep;
	tgt -> castles += src -> castles;
	tgt -> promotions += src -> promotions;
	tgt -> checks += src -> checks;
	tgt -> checkmates += src -> checkmates;
	tgt -> stalemates += src -> stalemates;
	tgt -> imd += src -> imd;
	tgt -> whiteWin += src -> whiteWin;
}

uint64_t getSysTimeUsed()
{
	struct rusage ru;

	if (getrusage(RUSAGE_SELF, &ru) == -1)
		return -1;

	return ru.ru_stime.tv_sec * 1000 * 1000 + ru.ru_stime.tv_usec;
}

uint64_t getUserTimeUsed()
{
	struct rusage ru;

	if (getrusage(RUSAGE_SELF, &ru) == -1)
		return -1;

	return ru.ru_utime.tv_sec * 1000 * 1000 + ru.ru_utime.tv_usec;
}

void statsThread(end_indicator_t *const ei, Brain *const b)
{
	uint64_t prevUser = getUserTimeUsed();
	uint64_t prevSys = getSysTimeUsed();

	for(;!ei -> flag;) {
		usleep(1000 * 1000);

		uint64_t curUser = getUserTimeUsed();
		uint64_t curSys = getSysTimeUsed();

		log(ll_debug, "TS %lu %lu", curUser - prevUser, curSys - prevSys);

		prevUser = curUser;
		prevSys = curSys;
	}
}

void timerThread(const uint64_t deadlineTs, end_indicator_t *const ei, IO *const io)
{
	int ms = int(deadlineTs - get_ts_ms());

	log(ll_ok, "timerThread: sleep %dms", ms);

	if (ms > 0)
	{
		std::mutex m;
		std::unique_lock<std::mutex> lk(m);

		for(;;)
		{
			if (ei -> cv.wait_for(lk, std::chrono::milliseconds(ms)) == std::cv_status::timeout)
			{
				log(ll_ok, "timeout");
				ei -> flag = true;
				break;
			}

			if (ei -> flag)
			{
				log(ll_ok, "timerThread stopped prematurely");
				break;
			}
		}
	}
	else {
		log(ll_ok, "immediate timeout");
		ei -> flag = true;
	}

	log(ll_ok, "timerThread end");
}

Brain::Brain(IO *ioIn, Tpt *const ttIn, Ptt *const pttIn, const int nThreadsIn, std::vector<Book *> *const booksIn) : tt(ttIn), ptt(pttIn), sepc(get_syzygy_etb_piece_count()), ponderEPMove(invalidMove), io(ioIn), books(booksIn), nThreads(nThreadsIn)
{
	statsLogging = false;

	resetStats();

	ponderScene = NULL;
	ponderColor = WHITE;

	ponderThreadHandle = NULL;
}

Brain::~Brain()
{
	stopPonder(NULL);
}

void Brain::registerIO(IO *const ioIn)
{
	io = ioIn;
}

const std::map<std::string, std::string> & Brain::getParameters()
{
	static const std::map<std::string, std::string> pars = { { "stats_logging", "thread usage counts & cpu usage" } };

	return pars;
}

bool Brain::setParam(const std::string & key, const std::string & value)
{
	std::string keyl = mytolower(key);

	if (keyl == "stats_logging")
		statsLogging = yesOrNo(value);
	else if (!tune -> setParam(key, atoi(value.c_str()))) {
		log(ll_err, "setParam(%s, %s): key not known", key.c_str(), value.c_str());
		return false;
	}

	return true;
}

void Brain::setNTHreads(const int  n)
{
	nThreads = n;
}

unsigned long int Brain::perft(Scene & s, const int depthLeft, const PlayerColor c, perft_counts_t *const pc, const bool isDivide)
{
	if (depthLeft == 0)
		return 1;

	s.generateMovelists();
	s.partialValidateMoves(c);

	if (s.isKingUnderAttack(c))
		pc -> checks++;
	if (s.isCheckMate(c)) {
		if (c != WHITE)
			pc -> whiteWin++;

		pc -> checkmates++;
	}
	if (s.isStaleMate(c))
		pc -> stalemates++;
	if (s.isInsufficientMaterialDraw())
		pc -> imd++;

	VectorType<Move> *const moves = s.getMoveList(c);

	const int newDepth = depthLeft - 1;
	const PlayerColor co = opponentColor(c);

	unsigned long int ccnt = 0;

	BoardState bs;
 	for(size_t idx=0; idx<moves -> size(); idx++)
 	{
		if (!s.validateMoves(c, idx))
			continue;

		Move m = moves -> at(idx);
 
//		printf("%d %s %s\n", depthLeft, genFen(&s, true, c).c_str(), moveToString(m).c_str());

		if (m.u.m.isCatchMove)
			pc -> captures++;

		if (m.u.m.mt == EN_PASSANT)
			pc -> ep++;
		else if (m.u.m.mt == CASTLING)
			pc -> castles++;

		if (m.u.m.promoteTo != NONE)
			pc -> promotions++;
 
		s.doMove(m, &bs);

		unsigned long int cur = perft(s, newDepth, co, pc, false);
		ccnt += cur;

 		s.undoMove();

		if (isDivide)
			io -> emitCmd("# %s %lu", moveToString(m).c_str(), cur);
 	}

	return ccnt;
}

int Brain::countTotalObjectsValue(const int counts[])
{
	int total = 0;

	for(int index=0; index<6; index++)
		total += ChessPiece::evalVal[index] * counts[index];

	if (counts[BISHOP] >= 2)
		total += tune -> moreBishops;

	if (unlikely(counts[PAWN] == 0))
		total += tune -> noPawns;

#ifdef TOO_MUCH_PAWNS
	if (counts[PAWN] == 8)
		total += tune -> tooMuchPawns;
#endif

	// total maximum: 4075 (material)

	return total;
}

int sumPositionValue(const bool isEndGame, const ChessPiece *const objs, const PlayerColor c)
{
	int score = 0;

	for(int index=0; index<16; index++)
	{
		const ChessPiece *const obj = &objs[index];

		if (!obj -> getAlive())
			continue;

		score += evalPositions(obj -> getX(), obj -> getY(), obj -> getType(), c, isEndGame);
	}

	return score;
}

void getNumberOfPawnsPerFile(const ChessPiece * const *objs, int *const counts)
{
	for(int index=0; index<16; index++)
	{
		if (!objs[index] -> getAlive() || objs[index] -> getType() != PAWN)
			continue;

		const int x = objs[index] -> getX();
		my_assert(x >= 0 && x < 8, "getNumberOfPawnsPerFile x in range");

		counts[x]++;
	}
}

int scoreIsolatedPawns(const ChessPiece * const objs)
{
	int np[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };

	for(int i=0; i<16; i++)
	{
		const ChessPiece & cur = objs[i];
		if (!cur.getAlive() || cur.getType() != PAWN)
			continue;

		np[cur.getX()]++;
	}

	int isolatedCount = 0;

	for(int x=0; x<8; x++) {
		int left = x > 0 ? np[x - 1] : 0;
		int right = x < 7 ? np[x + 1] : 0;

		if (left == 0 && right == 0)
			isolatedCount++;
	}

	return isolatedCount;
}

int countNDoublePawns(const int obCountsMe[6][8])
{
	int n = 0;

	for(int x=0; x<8; x++)
	{
		if (unlikely(obCountsMe[PAWN][x] >= 2))
			n += obCountsMe[PAWN][x] - 1;
	}

	my_assert(n <= 7, "countNDoublePawns");

	return n;
}

int scorePassedPawns(const bool isEndGame, const Scene *const s)
{
	const ChessPiece *const whiteObjects = s -> getPieces(WHITE);
	const ChessPiece *const blackObjects = s -> getPieces(BLACK);

	int8_t whiteYmax[8] = { -1, -1, -1, -1, -1, -1, -1, -1 };

	int8_t blackYmin[8] = { 8, 8, 8, 8, 8, 8, 8, 8 };

	for(int i=0; i<16; i++)
	{
		if (whiteObjects[i].getAlive() && whiteObjects[i].getType() == PAWN)
		{
			int whitex = whiteObjects[i].getX();
			whiteYmax[whitex] = std::max(whiteYmax[whitex], int8_t(whiteObjects[i].getY()));
		}

		if (blackObjects[i].getAlive() && blackObjects[i].getType() == PAWN)
		{
			int blackx = blackObjects[i].getX();
			blackYmin[blackx] = std::min(blackYmin[blackx], int8_t(blackObjects[i].getY()));
		}
	}

#ifdef _DEBUG
	bool ba = false, wa = false;
	for(int i=0; i<8; i++) {
		wa |= whiteYmax[i] != -1;
		ba |= blackYmin[i] != 8;
	}
	my_assert(wa || s -> getMaterialCounts(WHITE)[PAWN] == 0, "scorePassedPawns");
	my_assert(ba || s -> getMaterialCounts(BLACK)[PAWN] == 0, "scorePassedPawns");
#endif

	// these values where suggested by mkchan (irc) and are from
	// stockfish
	constexpr int scores[2][8] = { { 0, 5, 20, 30, 40, 50, 80, 0 }, { 0, 5, 20, 40, 70, 120, 200, 0 } };

	int scoreWhite = 0, scoreBlack = 0;
	int nw = 0, nb = 0;
	for(int i=0; i<16; i++)
	{
		if (whiteObjects[i].getAlive() && whiteObjects[i].getType() == PAWN)
		{
			int whitex = whiteObjects[i].getX();
			int whitey = whiteObjects[i].getY();

			bool left = (whitex > 0 && (blackYmin[whitex - 1] <= whitey || blackYmin[whitex - 1] == 8)) || whitex == 0;
			bool front = blackYmin[whitex] < whitey || blackYmin[whitex] == 8;
			bool right = (whitex < 7 && (blackYmin[whitex + 1] <= whitey || blackYmin[whitex + 1] == 8)) || whitex == 7;

			if (left && front && right) {
				//printf("PPw: %c%c\n", whitex + 'a', whitey + '1');
				scoreWhite += scores[isEndGame][whitey];
				nw++;
			}
		}

		if (blackObjects[i].getAlive() && blackObjects[i].getType() == PAWN)
		{
			int blackx = blackObjects[i].getX();
			int blacky = blackObjects[i].getY();

			bool left = (blackx > 0 && (whiteYmax[blackx - 1] >= blacky || whiteYmax[blackx - 1] == -1)) || blackx == 0;
			bool front = whiteYmax[blackx] > blacky || whiteYmax[blackx] == -1;
			bool right = (blackx < 7 && (whiteYmax[blackx + 1] >= blacky || whiteYmax[blackx + 1] == -1)) || blackx == 7;

			if (left && front && right) {
				//printf("PPb: %c%c\n", blackx + 'a', blacky + '1');
				scoreBlack += scores[isEndGame][7 - blacky];
				nb++;
			}
		}
	}

	my_assert(nw <= 8, "scorePassedPawns");
	my_assert(nb <= 8, "scorePassedPawns");

	return scoreWhite - scoreBlack;
}

void countObjects(const ChessPiece * const myObjects, int counts[6][8], int *const moveCountScore)
{
#ifdef _DEBUG
	int n = 0;
#endif

	int tuneMul[] = { tune -> pawnMobilMul, tune -> knightMobilMul, tune -> bishopMobilMul, tune -> rookMobilMul, tune -> queenMobilMul, tune -> kingMobilMul };

	for(int i=0; i<16; i++) {
		if (!myObjects[i].getAlive())
			continue;

		const int type = myObjects[i].getType();

		counts[type][myObjects[i].getX()]++;

		*moveCountScore += myObjects[i].getMoveCount() * tuneMul[type];

#ifdef _DEBUG
		my_assert(myObjects[i].getType() < 6, "countObjects");
		my_assert(myObjects[i].getType() >= 0, "countObjects");
		my_assert(myObjects[i].getX() >= 0, "countObjects");
		my_assert(myObjects[i].getX() < 8, "countObjects");

		n++;
#endif
	}

	my_assert(n >= 1 && n <= 16, "countObjects");
}

int nRookOnOpenFile(int obCountsMe[6][8])
{
	int cnt = 0;

	for(int i=0; i<8; i++)
		cnt += obCountsMe[PAWN][i] == 0 && obCountsMe[ROOK][i] > 0;

	my_assert(cnt <= 10, "2 rooks + 8 promotions");

	return cnt;
}

int kingPawnShield(const Scene *const restrict s, const PlayerColor me)
{
	const Board *const b = s -> getBoardConst();
	const ChessPiece *const king = s -> getKing(me);
	int checkx = king -> getX();

	int checky, kingy = king -> getY();

	if (me == WHITE)
		checky = kingy == 7 ? 6 : kingy + 1;
	else
		checky = kingy == 0 ? 1 : kingy - 1;

	int cnt = 0;

	if (checkx) {
		const ChessPiece *chk = b -> getAt(checkx - 1, checky);

		if (chk && chk -> getColor() == me && chk -> getType() == PAWN && !s -> getIsUnderAttackByPawn(chk))
			cnt++;

		chk = b -> getAt(checkx - 1, kingy);

		if (chk && chk -> getColor() == me && chk -> getType() == PAWN && !s -> getIsUnderAttackByPawn(chk))
			cnt++;
	}

	{
		const ChessPiece *const chk = b -> getAt(checkx, checky);

		if (chk && chk -> getColor() == me && chk -> getType() == PAWN && !s -> getIsUnderAttackByPawn(chk))
			cnt++;
	}

	if (checkx < 7) {
		const ChessPiece *chk = b -> getAt(checkx + 1, checky);

		if (chk && chk -> getColor() == me && chk -> getType() == PAWN && !s -> getIsUnderAttackByPawn(chk))
			cnt++;

		chk = b -> getAt(checkx + 1, kingy);

		if (chk && chk -> getColor() == me && chk -> getType() == PAWN && !s -> getIsUnderAttackByPawn(chk))
			cnt++;
	}

	return cnt;
}

int kingSafety(const Scene *const s, const PlayerColor me)
{
	int kx, ky;
	s -> getKing(me) -> getXY(&kx, &ky);

	const PlayerColor him = opponentColor(me);

	int cnt = 0;

	if (ky) {
		if (kx)
			cnt += s -> isFieldUnderAttack(him, kx - 1, ky - 1);
		cnt += s -> isFieldUnderAttack(him, kx, ky - 1);
		if (kx < 7)
			cnt += s -> isFieldUnderAttack(him, kx + 1, ky - 1);
	}

	if (kx)
		cnt += s -> isFieldUnderAttack(him, kx - 1, ky);
	if (kx < 7)
		cnt += s -> isFieldUnderAttack(him, kx + 1, ky);

	if (ky < 7) {
		if (kx)
			cnt += s -> isFieldUnderAttack(him, kx - 1, ky + 1);
		cnt += s -> isFieldUnderAttack(him, kx, ky + 1);
		if (kx < 7)
			cnt += s -> isFieldUnderAttack(him, kx + 1, ky + 1);
	}

	return cnt;
}

int centerControlHelper(const Board *const b, const int x, const int y, const PlayerColor me)
{
	if (b -> getAt(x, y) == NULL)
		return 0;

	return b -> getAt(x, y) -> getColor() == me ? 1 : -1;
}

int centerControl(const Board *const b, const PlayerColor me)
{
	return centerControlHelper(b, 3, 3, me) +
		centerControlHelper(b, 4, 3, me) +
		centerControlHelper(b, 3, 4, me) +
		centerControlHelper(b, 4, 4, me);
}

int rookOnEndRanks(const ChessPiece * const obs, const PlayerColor c)
{
	int cnt = 0;

	for(int i=0; i<16; i++) {
		if (obs[i].getAlive() == false)
			continue;

		const int y = obs[i].getY();

		cnt += ((y == 0 || y == 1) && c == BLACK);
		cnt += ((y == 6 || y == 7) && c == WHITE);
	}

	return cnt;
}

int distanceRBQtoKing(const Scene *const s, const PlayerColor me)
{
        const PlayerColor opponent = opponentColor(me);
	const ChessPiece * const myObjects = s -> getPieces(me);
	const ChessPiece *const oppKing = s -> getKing(opponent);
	int kx = oppKing->getX();
	int ky = oppKing->getY();

	int score = 0;

	for(int i=0; i<16; i++) {
		if (!myObjects[i].getAlive())
			continue;

		const ChessPieceType cpt = myObjects[i].getType();
		const int x = myObjects[i].getX();
		const int y = myObjects[i].getY();

		if (cpt == ROOK || cpt == QUEEN) {
			int dx = abs(x - kx);
			int dy = abs(y - ky);

			if ((dy == 0 && dx <= 2) || (dx == 0 && dy <=2 ))
				score += 5;
		}

		if (cpt == BISHOP || cpt == QUEEN) {
			int dx = abs(x - kx);
			int dy = abs(y - ky);

			if (dx <= 2 && dy <= 2 && dx == dy)
				score += 5;
		}
	}

	return score;
}

int Brain::calcShannon(const zobrist_t & nh, const Scene *const s, calc_stats_t *const cs, const PlayerColor me, const PlayerColor localColor)
{
        const PlayerColor opponent = opponentColor(me);
	const ChessPiece * const myObjects = s -> getPieces(me);
        const ChessPiece * const hisObjects = s -> getPieces(opponent);

        int obCounts[2][6][8];
        memset(obCounts, 0x00, sizeof obCounts);

	int moveScoreMe = 0, moveScoreHis = 0;
	countObjects(myObjects, obCounts[me], &moveScoreMe);
	countObjects(hisObjects, obCounts[opponent], &moveScoreHis);

        const int *const typeCountsMe = s -> getMaterialCounts(me);
        const int *const typeCountsOpponent = s -> getMaterialCounts(opponent);

	const bool endGame = s -> isEndGame();

        int value =
#ifdef PSQ
                (sumPositionValue(endGame, myObjects, me) - sumPositionValue(endGame, hisObjects, opponent)) / tune -> divSumPositionValue
#endif
#ifdef MATERIAL
                + (countTotalObjectsValue(typeCountsMe) - countTotalObjectsValue(typeCountsOpponent)) * tune -> mulObjValue
#endif
		+ 0
                ;

#ifdef CENTER_CONTROL
	value += centerControl(s -> getBoardConst(), me) * tune -> centerControl;
#endif

#ifdef MOBILITY
	value += (moveScoreMe - moveScoreHis) * tune -> mobilityMul;
#endif

#ifdef ROOK_ON_OPEN_FILE
	value += (nRookOnOpenFile(obCounts[me]) - nRookOnOpenFile(obCounts[opponent])) * tune -> mulROF;
#endif

#ifdef PAWN_SHIELD
	value += (kingPawnShield(s, me) - kingPawnShield(s, opponent)) * tune -> pawnShield;
#endif

#ifdef KING_SAFETY
	value += (kingSafety(s, me) - kingSafety(s, opponent)) * tune -> kingSafety;
#endif

	int pp = 0;
#ifdef PAWN_HASH
	pttEntry pte;
	if (ptt -> lookup(nh.pawnNewHash, &pte)) {
		if (cs)
			cs -> ptt_hit++;

		pp = pte.data_._data.value;

#if _DEBUG
		int v = scorePassedPawns(endGame, s);
		v -= (countNDoublePawns(obCounts[WHITE]) - countNDoublePawns(obCounts[BLACK])) * tune -> mulDoublePawn;
		v -= (scoreIsolatedPawns(myObjects) - scoreIsolatedPawns(hisObjects)) * tune -> mulIsolatedPawns;

		if (pp != v) {
			printf("FAIL %d %d\n", pp, v);
			my_assert(0, "scorePassedPawns");
		}
#endif
	}
	else
#endif
	{
		if (cs)
			cs -> ptt_mis++;

#ifdef PASSED_PAWNS
		pp = scorePassedPawns(endGame, s);
#endif
#ifdef DOUBLE_PAWNS
		pp -= (countNDoublePawns(obCounts[WHITE]) - countNDoublePawns(obCounts[BLACK])) * tune -> mulDoublePawn;
#endif
#ifdef ISOLATED_PAWNS
		pp -= (scoreIsolatedPawns(myObjects) - scoreIsolatedPawns(hisObjects)) * tune -> mulIsolatedPawns;
#endif

#ifdef PAWN_HASH
		ptt -> store(nh.pawnNewHash, -1, pp);
#endif
	}

#ifdef ROER
	if (endGame)
		value += (rookOnEndRanks(myObjects, me) - rookOnEndRanks(hisObjects, opponent)) * tune -> roer;
#endif

	if (me != WHITE)
		pp = -pp;

#ifdef DISTANCE_RBQ
	value += distanceRBQtoKing(s, me) - distanceRBQtoKing(s, opponent);
#endif

#ifdef TEMPO
	if (me == localColor && !endGame)
		value += tune -> tempo; // tempo
#endif

	value += pp;

        return value;
}

int calcMateScore(const int currentSearchDistance)
{
	return -(value_checkmate - currentSearchDistance);
}

int Brain::evaluate(const zobrist_t & nh, const Scene *const s, calc_stats_t *const cs, const PlayerColor me, const int currentSearchDistance, const PlayerColor localColor)
{
	my_assert(s -> getMoveList(opponentColor(me)) -> size() > 0, "evaluate: opponent-list is not empty");

	cs -> evals++;

	int score = calcShannon(nh, s, cs, me, localColor);
	my_assert(score == -calcShannon(nh, s, cs, opponentColor(me), localColor), "evaluate");

	my_assert(score >= -value_checkmate && score <= value_checkmate, "evaluate");
	return score;
}

zobrist_t Brain::calcZobrist(const Board * const restrict b, const PlayerColor c, const Move epMove)
{
#ifdef USE_TT
	return tt -> calcZobrist(b, c, epMove);
#else
	zobrist_t dummy;
	return dummy;
#endif
}

zobrist_t Brain::updateZobrist(const zobrist_t & in, const BoardState & bs, const Board *const restrict b, const PlayerColor c, const Move epMove)
{
#ifdef USE_TT
	zobrist_t zh = tt -> updateZobrist(in, bs, b, c, epMove);

#if 0
	if (tt -> calcZobrist(b, c, epMove).pawnNewHash != zh.pawnNewHash) {
		printf("FAIL %lx %lx %s\n", zh.pawnNewHash, tt -> calcZobrist(b, c, epMove).pawnNewHash, genFen(s, true, c).c_str());
		*fail = true;
	}
#endif

	return zh;
#else
	zobrist_t dummy;
	return dummy;
#endif
}

bool Brain::evalTT(search_data_t *const sd, const PlayerColor c, Tpt *const restrict curtt, const int depthLeft, const uint64_t hash, const int alpha, const int beta, int *const restrict useValue, Move *const restrict m, bool *const useNm, const int csd) const
{
#ifdef USE_TT
	tptEntry ttEntry;
	if (!curtt -> lookup(hash, &ttEntry))
		return false;

	*useNm = ttEntry.data_._data.nm;

	m -> u.m = ttEntry.data_._data.m;

	// check for hash collisions
	if (isValidMove(*m)) {
		if (m->u.m.mt == REGULAR) {
			const Board *const b = sd -> s -> getBoardConst();

			ChessPiece *const from = b->getAt(m->u.m.fx, m->u.m.fy);
			if (from == nullptr || from->getColor() != c)
				return false;


			ChessPiece *const to = b->getAt(m->u.m.tx, m->u.m.ty);
			if (to != nullptr && (to->getColor() == c || to->getType() == KING))
				return false;
		}
		else {
			VectorType<Move> *const moves = sd -> s -> getMoveList(c);
			ssize_t idx = indexOf(*moves, *m);

			if (idx == -1) {
				sd -> cs.tt_err++;
				m -> u.m = invalidMove.u.m;
				return false;
			}
		}
	}

	if (ttEntry.data_._data.depth >= depthLeft) {
		sd -> cs.tt_depth_ok++;

		const tptEntryFlag flags = (tptEntryFlag)ttEntry.data_._data.flags;

		sd -> cs.tt_hit_type[ttEntry.data_._data.flags]++;

		int cur_v = abs(ttEntry.data_._data.value) > value_checkmate_min ? (ttEntry.data_._data.value < 0 ? ttEntry.data_._data.value + csd : ttEntry.data_._data.value - csd) : ttEntry.data_._data.value;

		if (flags == EXACT) {
			*useValue = cur_v;
			return true;
		}

		if (flags == LOWERBOUND) {
			if (cur_v>=beta) {
				*useValue=cur_v;
				return true;
			}
		}
		else if (flags == UPPERBOUND) {
			if(cur_v<=alpha) {
				*useValue=cur_v;
				return true;
			}
		}
	}
	else if (ttEntry.data_._data.depth == 0) {
		 sd -> cs.tt_d0++;
	}
#endif

	return false;
}

void Brain::updateTT(Tpt *const restrict curtt, const uint64_t hash, const int ttDepth, const int alphaOrig, const int beta, const int bestVal, const Move & m, const bool useNm) const
{
#ifdef USE_TT
	my_assert(bestVal > -infinite && bestVal != infinite, "Brain::updateTT sane values");

	tptEntryFlag newFlags = EXACT;

	if (bestVal <= alphaOrig)
		newFlags = UPPERBOUND;
	else if (bestVal >= beta)
		newFlags = LOWERBOUND;

	curtt -> store(hash, newFlags, ttDepth, bestVal, bestVal <= alphaOrig ? invalidMove.u.m : m.u.m, useNm);
#endif
}

int countPiecesAlive(const ChessPiece *const pieces)
{
	int n = 0;

	for(int i=0; i<16; i++)
		n += !!pieces[i].getAlive();

	return n;
}

Move syzygyMoveToMove(const Scene *const s, const char *const moveStr)
{
	Move cm;

	cm.u.m.fx = uint8_t(moveStr[0] - 'a');
	cm.u.m.fy = uint8_t(moveStr[1] - '1');
	cm.u.m.tx = uint8_t(moveStr[2] - 'a');
	cm.u.m.ty = uint8_t(moveStr[3] - '1');
	cm.u.m.mt = REGULAR;
	cm.u.m.isCatchMove = s -> getBoardConst() -> getAt(cm.u.m.tx, cm.u.m.ty) != NULL; // FIXME en-passant

	if (moveStr[4] == 0x00)
		cm.u.m.promoteTo = NONE;
	else if (moveStr[4] == 'Q')
		cm.u.m.promoteTo = QUEEN;
	else if (moveStr[4] == 'N')
		cm.u.m.promoteTo = KNIGHT;
	else if (moveStr[4] == 'R')
		cm.u.m.promoteTo = ROOK;
	else if (moveStr[4] == 'B')
		cm.u.m.promoteTo = BISHOP;
	else
		log(ll_err, "Unexpected move format %s", moveStr);

	return cm;
}

bool checkSyzygy(const Scene *const s, calc_stats_t *const cs, const PlayerColor c, int *const score, VectorType<Move> *const cm, int *const depth, const bool async, const bool quick)
{
#ifdef W_SYZYGY
	cm -> clear();

	int nWhite = countPiecesAlive(s -> getPieces(WHITE));
	int nBlack = countPiecesAlive(s -> getPieces(BLACK));

	if (nWhite + nBlack <= get_syzygy_etb_piece_count())
	{
		cs -> se_query[quick]++;
		const std::string fen = genFen(s, false, c);

		std::vector<std::string> moveStrings;
		syzygy_result_t res;
		query_syzygy_etb(fen.c_str(), &res, &moveStrings, depth, async, quick);

		if (res != se_fail)
		{
			cs -> se_hit[quick][res]++;

			// log(ll_err, "SYZYGY HIT");
			if (res == se_lost)
				*score = -(value_checkmate_min - *depth);
			else if (res == se_draw)
				*score = 0;
			else if (res == se_win)
				*score = value_checkmate_min - *depth;

			for(std::string m : moveStrings)
				cm -> push_back(syzygyMoveToMove(s, m.c_str()));
		}
	}

	return !cm -> empty();
#else
	return false;
#endif
}

typedef struct
{
	Scene *s;
	calc_stats_t *cs;
	PlayerColor c;
	//
	int score, depth;
	VectorType<Move> moves;
	std::atomic_bool valid;
} syzygyThreadPars_t;

void *syzygyThread(syzygyThreadPars_t *const p)
{
	log(ll_ok, "syzygy etb thread started");

	if (checkSyzygy(p -> s, p -> cs, p -> c, &p -> score, &p -> moves, &p -> depth, false, false))
	{
		log(ll_ok, "syzygy etb thread hit");
		p -> valid = true;
	}

	log(ll_ok, "syzygy etb thread terminating");

	return NULL;
}

void updateBestSiblings(Move *const bestSiblings, const VectorType<PVnode_t> & pv)
{
	if (!pv.empty()) {
		bestSiblings[0] = bestSiblings[1];
		bestSiblings[1] = pv.back().m;
	}
}

int countValidMoves(Scene *const s, const PlayerColor c)
{
	VectorType<Move> *const moves = s -> getMoveList(c);

	int count = 0;

	for(size_t idx=0; idx<moves -> size(); idx++)
		count += s -> validateMoves(c, idx);

	return count;
}

int contemptValue(search_data_t *const restrict sd, const PlayerColor c)
{
#ifdef WITH_CONTEMPT
	const int cont = sd -> s -> isEndGame() ? tune -> contemptEndgame : tune -> contemptMiddlegame;

	return c == sd -> colorPlayedByEngine ? -cont : +cont;
#else
	return 0;
#endif
}

#ifdef QUIESCENCE
int Brain::quiesceSearch(search_data_t *const restrict sd, const int currentSearchDistance, const PlayerColor c, int alpha, int beta, const zobrist_t & zh, const Move siblingPrevBestMove[2], Move *const restrict bestMove)
{
	if (*sd -> time_abort_flag)
		return 0;

	sd -> cs.qcnt++;
	sd -> cs.max_depth = std::max(sd -> cs.max_depth, currentSearchDistance);

	const int alphaOrig = alpha;

	VectorType<Move> *const moves = sd -> s -> getMoveList(c);
	sd -> s -> partialValidateMoves(c);

	if (unlikely(sd -> s -> isInsufficientMaterialDraw()) || unlikely(sd -> s -> isStaleMate(c))) {
		sd -> cs.draw++;
		//printf("stale/insuf 0\n");
		return 0;
	}

	if (sd -> s -> isCheckMate(c)) {
		sd -> cs.mate++;

		return calcMateScore(currentSearchDistance);
	}

	bool useNm = true;

	Move ttMove = invalidMove;
#ifdef USE_TT
	int temp = 0;
	if (evalTT(sd, c, tt, 0, zh.newHash, alpha, beta, &temp, &ttMove, &useNm, currentSearchDistance))
	{
		sd -> cs.qtpt_hit_cnt++;

		*bestMove = ttMove;
		//printf("tt %d\n", temp);

		return temp;
	}
#endif

	int bestVal = -infinite;

	bool maxDepthReached = currentSearchDistance > MAX_QUIESCENCE_DEPTH;

	bool isCheck = sd -> s -> isKingUnderAttack(c);

#ifdef STANDING_PAT
	if (!isCheck || moves -> size() == 0 || maxDepthReached) {
		bestVal = evaluate(zh, sd -> s, &sd -> cs, c, currentSearchDistance, sd -> colorPlayedByEngine);

		if (maxDepthReached) {
			sd -> cs.max_depth_reached++;
			//printf("mdr %d\n", bestVal);
			return bestVal;
		}

		if (bestVal > alpha) {
			if (bestVal >= beta) {
				sd -> cs.qstand_pat_bco++;
			//printf("standpat %d\n", bestVal);
				return bestVal;
			}

			alpha = bestVal;
		}
	}
#endif

	const Board *const b = sd -> s -> getBoardConst();

        if (!isCheck) {
                for(size_t idx=0; idx<moves -> size();) {
                        const Move m = moves -> at(idx);

                        if (m.u.m.isCatchMove || m.u.m.promoteTo != NONE)
                                idx++;
                        else
                                moves -> erase(idx);
                }
        }

	const size_t nMoves = moves -> size();

	VectorType<Move> putInFront;
	if (isValidMove(ttMove))
		putInFront.push_back(ttMove);

#ifdef KILLER_HEURISTIC
	if (isValidMove(siblingPrevBestMove[1])) {
		putInFront.push_back(siblingPrevBestMove[1]);

		if (isValidMove(siblingPrevBestMove[0]) && moveCompareTo(siblingPrevBestMove[0], siblingPrevBestMove[1]) == false)
			putInFront.push_back(siblingPrevBestMove[0]);
	}
#endif

	 sd -> cs.pif_size += putInFront.size();
	 sd -> cs.pif_size_n++;

	reorderAndSortList(moves, putInFront, sd, c);

	const int newSearchDistance = currentSearchDistance + 1;

	const PlayerColor co = opponentColor(c);

	Move curBestSibling[2] = { invalidMove, invalidMove };

	size_t nm = 0;
	BoardState bs;
	for(size_t idx=0; idx<nMoves && !*sd -> time_abort_flag; idx++) {
		if (!sd -> s -> validateMoves(c, idx))
			continue;

		const Move m = moves -> at(idx);

#ifdef BLIND
		if (!isCheck && m.u.m.isCatchMove) {
			 sd -> cs.blind_total++;

			// blind
			const ChessPiece *me = b -> getAt(m.u.m.fx, m.u.m.fy);
			// victim can be NULL when it is only a promotion
			const ChessPiece *victim = b -> findVictim(m);
			if (!sd -> s -> getSEE(victim -> getX(), victim -> getY(), me)) {
				 sd -> cs.blind++;
				continue;
			}
			/// ^
		}
#endif

		nm++;

		const Move epMove = isEpMove(b, m) ? m : invalidMove;

		Move siblingMove = invalidMove;

		int score = 0;

		sd -> s -> doMove(m, &bs);

		zobrist_t newHash = updateZobrist(zh, bs, b, co, epMove);

#ifdef REPETITION
		if (addToHistory(sd -> history, newHash.newHash)) {
			score = contemptValue(sd, co);
			 sd -> cs.history_hits++;
		}
		else
#endif
		{
			sd -> s -> generateMovelists();
			//printf("try %d %s\n", newSearchDistance, moveToString(m).c_str());

			score = -quiesceSearch(sd, newSearchDistance, co, -beta, -alpha, newHash, curBestSibling, &siblingMove);

#ifdef REPETITION
			removeFromToHistory(sd -> history);
#endif
		}

		sd -> s -> undoMove();

		if (score > bestVal)
		{
			bestVal = score;

			if (isValidMove(siblingMove)) {
				curBestSibling[0] = curBestSibling[1];
				curBestSibling[1] = siblingMove;
			}

			if (score > alpha)
			{
				alpha = score;

				*bestMove = m;

				if (score >= beta)
					break;
			}
		}
	}

	 sd -> cs.q_nm += nm;

	// if:
	// - see removed all moves
	// - not check but all non-capture moves were valid but the capture moves were not
	if (nm == 0) {
		if (bestVal == -infinite)
			return evaluate(zh, sd -> s, &sd -> cs, c, currentSearchDistance, sd -> colorPlayedByEngine);

		//printf("nm %d\n", bestVal);

		return bestVal;
	}

	my_assert(abs(bestVal) != infinite, "Qs");

#ifdef QS_TT_UPD
	if (alpha > alphaOrig && !*sd -> time_abort_flag) {
		my_assert(abs(bestVal) != 32767, "sane eval val for tt");

		updateTT(tt, zh.newHash, 0, alphaOrig, beta, bestVal, *bestMove, true);
	}
#endif

	//printf("--- %d\n", bestVal);

	return bestVal;
}
#endif

bool evalResult(const int val, int *const restrict bestVal, int *const restrict alpha, int *const restrict beta, bool *const betaCutOff)
{
	bool useMove = false;

	if (val > *bestVal)
	{
		*bestVal = val;
		useMove = true;
	}

	if (val > *alpha)
	{
		*alpha = val;

		if (val >= *beta)
			*betaCutOff = true;
	}

	return useMove;
}

// cbcm = caused by capture move
int Brain::search(search_data_t *const restrict sd, const bool cbcm, const int depthLeft, const PlayerColor c, int alpha, int beta, const zobrist_t & zh, const Move siblingPrevBestMove[2], const bool isNullMove, VectorType<PVnode_t> *const restrict pv)
{
	my_assert(beta >= alpha, "search");
	my_assert(depthLeft >= 0, "Brain::search depthLeft must be >= 0");
	my_assert(alpha >= -infinite && alpha <= infinite, "search");
	my_assert(beta >= -infinite && beta <= infinite, "search");
#ifdef _DEBUG
	isValidMove(siblingPrevBestMove[0]);
	isValidMove(siblingPrevBestMove[1]);
#endif

	const int currentSearchDistance = sd -> curMaxDepth - depthLeft;

#ifdef QUIESCENCE
	if (depthLeft == 0) {
		Move bestMove = invalidMove;
		int score = quiesceSearch(sd, sd -> curMaxDepth, c, alpha, beta, zh, siblingPrevBestMove, &bestMove);

		pv -> clear();
		if (isValidMove(bestMove))
			push_back(pv, zh.newHash, bestMove, score, MS_QS);

		my_assert(score >= -value_checkmate && score <= value_checkmate, "qs");
		return score;
	}
#else
	if (depthLeft == 0) {
		// FIXME check for draw/checkmate
		int score = evaluate(zh, sd -> s, &sd -> cs, c, currentSearchDistance, sd -> colorPlayedByEngine);
		my_assert(score >= -value_checkmate && score <= value_checkmate, "dl");
		return score;
	}
#endif

	const bool rootOfTree = depthLeft == sd -> curMaxDepth;

	VectorType<Move> *const moves = sd -> s -> getMoveList(c);

	sd -> s -> partialValidateMoves(c);
	if (((sd -> s -> getNHalfMoves() >= 100 || sd -> s -> isInsufficientMaterialDraw() || sd -> s -> isStaleMate(c)) && !rootOfTree) || *sd -> time_abort_flag) {
		 sd -> cs.draw++;
		return 0;
	}

	if (sd -> s -> isCheckMate(c))
		return calcMateScore(currentSearchDistance);

	long unsigned int nMoves = moves -> size();
	bool isCheck = sd -> s -> isKingUnderAttack(c);

	if (nMoves == 0) {
		int score = 0;

		if (isCheck) {
			score = calcMateScore(sd -> curMaxDepth - depthLeft);
			 sd -> cs.mate++;
		}

		my_assert(score >= -value_checkmate && score <= value_checkmate, "nm=0");

		return score;
	}

	 sd -> cs.cnt++;

	const int alphaOrig = alpha;

	const Board *const b = sd -> s -> getBoardConst();

	bool useNm = true;
	Move ttMove = invalidMove;
#ifdef USE_TT
	int temp = 0;
	if (evalTT(sd, c, tt, depthLeft, zh.newHash, alpha, beta, &temp, &ttMove, &useNm, currentSearchDistance))
	{
		 sd -> cs.tpt_hit_cnt++;

		my_assert(temp >= -value_checkmate && temp <= value_checkmate, "tt eval");

		if (isValidMove(ttMove)) {
			pv -> clear();
			push_back(pv, zh.newHash, ttMove, temp, MS_TPT);

			return temp;
		}

		if (!rootOfTree)
			return temp;
	}
#endif

	// http://www.talkchess.com/forum/viewtopic.php?p=656652#656652
#ifdef W_SYZYGY
	Move sepcMove = invalidMove;
	if (sepc && !isValidMove(ttMove) && cbcm && depthLeft >= SE_MIN_PROBE_DEPTH)
	{
		int score = 0;

		VectorType<Move> smoves;
		int seDepth = 0;
		if (checkSyzygy(sd -> s, &sd -> cs, c, &score, &smoves, &seDepth, true, true))
		{
			VectorType<Move> putInFront;
			reorderAndSortList(&smoves, putInFront, sd, c);
			sepcMove = smoves.at(0);

			log(ll_ok, "SYZYGY HIT IN SEARCH");
			push_back(pv, zh.newHash, sepcMove, score, MS_SYZYGY);

			my_assert(score >= -value_checkmate && score <= value_checkmate, "sepc");

			return score;
		}
	}
#endif

	const PlayerColor co = opponentColor(c);

	/********************************** NULL MOVE */
#ifdef NULL_MOVE

        // -1: cannot directly jump into quiesce search: that seems to cause
        // segfaults. not yet understood why.
        int nmDepth = std::min(depthLeft, NULL_MOVE_JUMP_MAX);

#ifdef _DEBUG
	if (isNullMove) {
		my_assert(beta == alpha + 1, "NULL MOVE");
		my_assert(!sd -> s -> isEndGame(), "NULL MOVE");
		my_assert(currentSearchDistance >= 1, "NULL MOVE");
	}
#endif

	if (useNm && !rootOfTree && nmDepth > 0 && !isCheck && !isNullMove && !sd -> s -> isEndGame()) {
		sd -> cs.null_move++;

		Move nmm;
		int nmhm = 0;
		sd->s->setNullMove(&nmm, &nmhm);

		zobrist_t newHash;
#ifdef USE_TT
		newHash = tt -> updateZobristNullMove(zh, b, co);
#endif

		VectorType<PVnode_t> nmTemp;
		int val = 0;

#ifdef REPETITION
		if (addToHistory(sd -> history, newHash.newHash)) {
			sd -> cs.history_hits++;
			val = contemptValue(sd, co);
		}
		else
#endif
		{
			val = -search(sd, false, depthLeft - nmDepth, co, -beta, -beta + 1, newHash, invalidBestSibblings, true, &nmTemp);
#ifdef REPETITION
			removeFromToHistory(sd -> history);
#endif
		}

		sd->s->unsetNullMove(nmm, nmhm);

#ifdef _DEBUG
		if ((val < -value_checkmate || val > value_checkmate) && !*stopFlag) {
			printf("%d %d %d\n", alphaOrig, beta, val);
			printf("%s\n", genFen(sd -> s, false, co).c_str());
			printf("%s\n", pvToString(nmTemp, false, false, false).c_str());
			my_assert(0, "NM");
		}
#endif

		if (val >= beta)
		{
			if (val >= value_checkmate_min) // not proven yet
				val = beta;

#ifdef NM_TT
			if (!*sd -> time_abort_flag) {
				updateTT(tt, zh.newHash, depthLeft, alphaOrig, beta, val, invalidMove, true);
				my_assert(val >= -value_checkmate && val <= value_checkmate, "nm");
			}
#endif

			sd -> cs.null_move_hit++;
			return val;
		}
	}
#endif
	/**********************************************/

	VectorType<Move> putInFront;
	bool doInternalItDeepening = true;

	ssize_t pvEntry = indexOf(*sd -> oldPv, zh.newHash);
	if (pvEntry != -1) {
		putInFront.push_back(sd -> oldPv -> at(pvEntry).m);
		doInternalItDeepening = false;
	}

	if (isValidMove(ttMove)) {
		putInFront.push_back(ttMove);
		doInternalItDeepening = false;
	}

	// https://chessprogramming.wikispaces.com/Killer+Heuristic
#ifdef KILLER_HEURISTIC
	if (isValidMove(siblingPrevBestMove[1])) {
		putInFront.push_back(siblingPrevBestMove[1]);

		if (isValidMove(siblingPrevBestMove[0]) && moveCompareTo(siblingPrevBestMove[0], siblingPrevBestMove[1]) == false)
			putInFront.push_back(siblingPrevBestMove[0]);
	}
#endif

#ifdef IID
	if (doInternalItDeepening && !isNullMove && depthLeft > tune -> iidMinDepth) { // reduce == 2 so skip the compare
		VectorType<PVnode_t> iidPV;
		(void)search(sd, cbcm, depthLeft - tune -> iidReduce, c, alpha, beta, zh, siblingPrevBestMove, false, &iidPV);

		 sd -> cs.iid++;

		// iidPV can be empty when a nmoves==0 or rules50
		if (!iidPV.empty()) {
			 sd -> cs.iid_hit++;
			putInFront.insert(iidPV.back().m);
		}
	}
#endif

#ifdef W_SYZYGY
	if (isValidMove(sepcMove))
		putInFront.push_back(sepcMove);
#endif

	 sd -> cs.pif_size += putInFront.size();
	 sd -> cs.pif_size_n++;

	reorderAndSortList(moves, putInFront, sd, c);

	const int newDepthLMR1 = ((depthLeft - tune -> lmr1sub) * tune -> lmr1mul) / tune -> lmr1div;
	const int newDepthLMR2 = ((depthLeft - tune -> lmr2sub) * tune -> lmr2mul) / tune -> lmr2div;
	const bool allowLMR = !sd -> s -> isKingUnderAttack(c) && depthLeft >= tune -> lmrMinDepth;
// FIXME putInFront.size() is different for semi-valid possibly!
	const size_t lmrStart = allowLMR ? std::max(size_t(tune -> lmrStart), putInFront.size()) : 999/* should not happen */;
	const size_t lmrNext = lmrStart + tune -> lmrNext;

	Move curBestSiblings[2] = { invalidMove, invalidMove };

//printf("%s\n", toString(*moves, false).c_str());

	VectorType<PVnode_t> tempPv;

	BoardState bs;
	int bestVal = -infinite;
	size_t mn = 0;
	for(size_t idx=0; idx<nMoves && !*sd->time_abort_flag; idx++) {
		if (!sd -> s -> validateMoves(c, idx))
			continue;

		mn++;
		const Move m = moves -> at(idx);
		my_assert(isValidMove(m), "search");

		if (rootOfTree)
			io -> emitNodeProgress(m, pv -> empty() ? invalidMove : pv -> back().m);

//printf("%s ", moveToString(m).c_str());

		bool isLMR = false;
		int newDepth = depthLeft - 1;
#ifdef WITH_LMR
		if (mn >= lmrStart && m.u.m.promoteTo == NONE) {
			const ChessPiece *const me = b -> getAt(m.u.m.fx, m.u.m.fy);
			bool isCap = m.u.m.isCatchMove;

			if (isCap) {
				const ChessPiece *const victim = b -> findVictim(m);

				if (!sd -> s -> getSEE(victim -> getX(), victim -> getY(), me))
					isCap = false;
			}

			if (isCap == false) {
				my_assert(mn > 1, "lmr");
				isLMR = true;

				if (mn >= lmrNext)
					newDepth = newDepthLMR2;
				else
					newDepth = newDepthLMR1;
			}
		}
#endif

		{
			const Move epMove = isEpMove(b, m) ? m : invalidMove;

			sd -> s -> doMove(m, &bs);

			int extension = 0;

			// https://chessprogramming.wikispaces.com/Check+Extensions
#ifdef CHECK_EXTENSION
			extension += isCheck * tune -> checkExtension;
#endif

			zobrist_t newHash = updateZobrist(zh, bs, b, co, epMove);

			tempPv.clear();

			int score = 0;
#ifdef REPETITION
			if (addToHistory(sd -> history, newHash.newHash)) {
				score = contemptValue(sd, co);
				 sd -> cs.history_hits++;
			}
			else
#endif
			{
				sd -> s -> generateMovelists();

				my_assert(isLMR || newDepth == depthLeft - 1, "LMR");

				sd -> cs.r_extensions += extension;

#ifdef WITH_LMR
				if (isCheck || sd -> s -> isKingUnderAttack(co))
					goto skip_lmr;
#endif

				// LMR
				score = -search(sd, m.u.m.isCatchMove, newDepth + extension, co, -beta, -alpha, newHash, curBestSiblings, false, &tempPv);

#ifdef WITH_LMR
				if (isLMR) {
					if (score > alpha) {
						updateBestSiblings(curBestSiblings, tempPv);

					skip_lmr:
						score = -search(sd, m.u.m.isCatchMove, depthLeft - 1 + extension, co, -beta, -alpha, newHash, curBestSiblings, false, &tempPv);
						sd -> cs.lmr_fail++;
					}
					else {
						sd -> cs.lmr_ok++;
					}
				}
#endif

#ifdef REPETITION
				removeFromToHistory(sd -> history);
#endif
			}

//fprintf(stderr, "%d %s %s %d\n", depthLeft, genFen(sd -> s, false, c).c_str(), moveToString(m).c_str(), score);
//fprintf(stderr, "%d %s %s\n", depthLeft, genFen(sd -> s, false, c).c_str(), moveToString(m).c_str());

			sd -> s -> undoMove();

                        if (score > bestVal) {
                                bestVal = score;

				updateBestSiblings(curBestSiblings, tempPv);

                                if (score > alpha) {
                                        alpha = score;

					pv -> assign(tempPv, false);
					push_back(pv, zh.newHash, m, bestVal, MS_SEARCH);

                                        if (score >= beta) {
						// https://chessprogramming.wikispaces.com/History+Heuristic
						if (!m.u.m.isCatchMove) {
#ifdef _DEBUG
							int hhtemp =
#endif
#ifdef WITH_HBT
						        sd -> hbt[c][m.u.m.fy * 8 + m.u.m.fx][m.u.m.ty * 8 + m.u.m.tx] += depthLeft * depthLeft;
#endif

							my_assert(hhtemp < 2000000000, "hbt");
							my_assert(hhtemp >= 0, "hbt");
						}

						sd -> cs.cnt_cutoff_index += mn - 1;
						sd -> cs.cnt_cutoff++;

						break;
                                        }
                                }
                        }
		}

	}

	my_assert(mn > 0 || *stopFlag, "mn sanity");

	// in theory stopFlag may get set right after the for-loop completely
	// finished its work and before the if statement checking for its
	// state. that is unfortunate but not a problem.
	if (alpha > alphaOrig && !*sd -> time_abort_flag) {
		my_assert(abs(bestVal) != infinite, "sane eval val for tt");

		updateTT(tt, zh.newHash, depthLeft, alphaOrig, beta, bestVal, pv -> empty() ? invalidMove : pv -> back().m, false);
	}

	my_assert(alpha >= alphaOrig, "search");
	my_assert((bestVal >= -value_checkmate && bestVal <= value_checkmate) || *stopFlag, "search");

	return bestVal;
}

void Brain::askBookQuestions(std::map<int, Book *> *const bookQuestions, Scene *const s, const PlayerColor c)
{
	for(Book *const cur : *books)
	{
		int br_id = cur -> ask(*s, c);

		bookQuestions -> insert(std::pair<int, Book *>(br_id, cur));
	}
}

void Brain::learnBookQuestions(const Board & b, const PlayerColor c, const Move & epMove, const Move & bestMove, const int depth, const winner_t w)
{
	for(Book *const cur : *books)
		cur -> learn(b, c, epMove, bestMove, depth, w);
}

void Brain::stopBookQuestions(std::map<int, Book *> *const bookQuestions)
{
	for(auto it = bookQuestions -> begin(); it != bookQuestions -> end(); it++)
		it -> second -> stopAllAsks();
}

bool Brain::checkBookQuestions(std::map<int, Book *> *const bookQuestions, Scene *const s, const PlayerColor c, VectorType<PVnode_t> **const ml, bool *const isFinalAnswer)
{
	bool found = false;

	for(auto it = bookQuestions -> begin(); it != bookQuestions -> end(); it++)
	{
		VectorType<PVnode_t> *mBook = NULL;
		book_result_t br = it -> second -> getResult(it -> first, s, c, &mBook, isFinalAnswer);

		if (br == BR_OK)
		{
			*ml = mBook;

			std::string pv = pvToString(*mBook, false, false, false);
			log(ll_ok, "Book hit: %s", pv.c_str());

			found = true;
			break;
		}

		if (br == BR_ERR)
		{
			bookQuestions -> erase(it);
			break;
		}
	}

	if (found)
		stopBookQuestions(bookQuestions);

	return found;
}

double calcEBF(const std::vector<uint64_t> & nodeCounts)
{
	size_t nNodeCounts = nodeCounts.size();

	return nNodeCounts >= 3 ? sqrt(double(nodeCounts.at(nNodeCounts - 1)) / double(nodeCounts.at(nNodeCounts - 3))) : -1;
}

#if 0
Move findBestStaticEval(search_data_t *const sd, const PlayerColor c, const VectorType<Move> & ml)
{
	Move m = invalidMove;
	int bestVal = -infinite;
	const size_t n = ml.size();

	BoardState bs;
	for(size_t i=0; i<n; i++) {
		sd -> s -> doMove(ml.at(i), &bs);

		int v = evaluate(sd -> s, sd -> cs, c, 1);

		if (v > bestVal) {
			m = ml.at(i);
			bestVal = v;
		}

		sd -> s -> undoMove();
	}

	return m;
}
#endif

Move findLeastValuableAttacket(const Board & b, const VectorType<Move> & ml, const int tx, const int ty)
{
	int leastEval = infinite;
	Move which = invalidMove;

	for(size_t i=0; i<ml.size(); i++) {
		Move m = ml.at(i);

		if (m.u.m.tx == tx && m.u.m.ty == ty) {
			int curEvalVal = b.getAt(m.u.m.fx, m.u.m.fy) -> getEvalVal();

			if (curEvalVal < leastEval) {
				leastEval = curEvalVal;
				which = m;

				if (curEvalVal == ChessPiece::evalVal[PAWN])
					break;
			}
		}
	}

	return which;
}

void Brain::resetStats()
{
	memset(&gcs, 0x00, sizeof gcs);
}

typedef struct
{
	int depthWork;
	Move m;
	int depth, score;
	calc_stats_t tcs;
} thread_pars_t;

void thread(Brain *const b, const Scene & s, const PlayerColor c, std::atomic_bool *const taf, Tpt *const tt, const Move epMove, const int threadNr, const int laws, const int raws, std::vector<thread_pars_t> *const results, const int n_tp)
{
	log(ll_ok, "threadNr %d aws %d/%d", threadNr, laws, raws);
	setCPUAffinity(threadNr + 1);

	search_data_t sd;
	sd.colorPlayedByEngine = c;
	sd.time_abort_flag = taf;
	sd.s = new Scene(s);
	memset(&sd.cs, 0x00, sizeof sd.cs);
	sd.history = new std::vector<uint64_t>();
	VectorType<PVnode_t> oldPv;
	sd.oldPv = &oldPv;
	memset(sd.hbt, 0x00, sizeof sd.hbt);

	int alpha = -infinite, beta = infinite;

	zobrist_t hash = tt -> calcZobrist(s.getBoardConst(), c, epMove);

	Move mPrevBest[2] = { invalidMove, invalidMove };

	sd.s -> generateMovelists();
	sd.s -> partialValidateMoves(c);

	int score = -infinite, scoreDepth = -infinite;

	int tual = laws, tuar = raws;

	while(!*taf) {
		sd.curMaxDepth = results ->  at(threadNr).depthWork;
		sd.cs.max_depth = 0;

		VectorType<PVnode_t> pv;
		int v = b -> search(&sd, false, results ->  at(threadNr).depthWork, c, alpha, beta, hash, mPrevBest, false, &pv);

		if (*taf)
			break;

		log(ll_ok, "%d|%d/%d/%d] %d %s\n", threadNr, results ->  at(threadNr).depthWork, alpha, beta, v, pvToString(pv, true, false, false).c_str());

		io -> emitThreadStats(threadNr, alpha, beta, v, pv, results ->  at(threadNr).depthWork);

		oldPv.assign(pv, false);
		score = v;
		scoreDepth = results ->  at(threadNr).depthWork;


		if (v <= alpha) {
			beta = (alpha + beta) / 2;

			alpha = v - tual;
			if (alpha < -infinite)
				alpha = -infinite;

			tual += tual / 15 + 1;
		}
		else if (v >= beta) {
			alpha = (alpha + beta) / 2;

			beta = v + tuar;
			if (beta > infinite)
				beta = infinite;

			tuar += tuar / 15 + 1;
		}
		else {
			alpha = v - tual;
			if (alpha < -infinite)
				alpha = -infinite;

			beta = v + tuar;
			if (beta > infinite)
				beta = infinite;

			Move selMove = pv.empty() ? invalidMove : pv.back().m;
			if (pv.empty() || !isValidMove(selMove) || abs(v) == infinite)
				log(ll_err, "L-S: search returned with no results");
			else
				updateBestSiblings(mPrevBest, pv);

			results -> at(threadNr).depthWork++;

			int n = 0;
			for(int i=0; i<n_tp; i++)
				n += results -> at(i).depthWork == results -> at(threadNr).depthWork;

			if (n >= n_tp / 2) {
				log(ll_ok, "%d] extra depth", threadNr);
				results -> at(threadNr).depthWork++;
			}

			tual = laws;
			tuar = raws;
		}
	}

	delete sd.history;
	delete sd.s;

	if (oldPv.empty())
		results -> at(threadNr).m = invalidMove;
	else {
		results -> at(threadNr).m = oldPv.back().m;
		results -> at(threadNr).depth = scoreDepth;
		results -> at(threadNr).score = score;
	}

	log(ll_ok, "thread result, %d| depth %d score %d pv %s", threadNr, scoreDepth, score, pvToString(oldPv, true, false, false).c_str());

	memcpy(&results -> at(threadNr).tcs, &sd.cs, sizeof sd.cs);
}

void Brain::calculateMove(const int moveNr, Scene & s, const PlayerColor c, const int maxDepth, const int durationMin, const int durationMax, const VectorType<Move> *const targetMoves,  std::vector<uint64_t> *const historyIn, const Move epMove, VectorType<PVnode_t> **const result, int *const value, end_indicator_t *const ei, uint64_t *const nodesVisited, const Move ponderHit, const Move lastOppMove)
{
	uint64_t startTs = get_ts_ms();

	setCPUAffinity(0);

	const std::string ps = ponderThreadHandle ? "P " : "  ";

#ifdef USE_TT
	tt -> incAge();
#endif
#ifdef PAWN_HASH
	ptt -> incAge();
#endif
	std::thread *st = NULL;
	if (statsLogging)
		st = new std::thread(statsThread, ei, this);

	search_data_t sd;
	gcs.setStartedAt();
	sd.cs.setStartedAt();

	log(ll_ok, ps + "Start calculating on: " + genFen(&s, false, c));

	log(ll_ok, ps + "time: %d/%d, max depth: %d", durationMin, durationMax, maxDepth);

	s.generateMovelists();
	*result = new VectorType<PVnode_t>();

	// safety check
	if (s.isCheckMate(WHITE) || s.isCheckMate(BLACK)) {
		log(ll_err, ps + " Already checkmate");
		sd.cs.addCounts(&gcs);
		return;
	}

#ifdef USE_TT
	zobrist_t hash = tt -> calcZobrist(s.getBoardConst(), c, epMove);
#else
	zobrist_t hash;
#endif

	VectorType<Move> *ml = s.getMoveList(c);
	Move only = invalidMove;
	int onlyCnt = 0;

	for(size_t idx=0; idx<ml -> size() && onlyCnt < 2; idx++) {
		if (!s.validateMoves(c, idx)) // FIXME erase the invalid ones?
			continue;

		only = ml -> at(idx);
		onlyCnt++;
	}

	if (onlyCnt == 1 && !ponderThreadHandle) {
		io -> emitDebug("Only one move possible");
		
		*value = evaluate(hash, &s, &gcs, c, 0, c);

		if (nodesVisited)
			*nodesVisited = 0;

		push_back(*result, hash.newHash, only, *value, MS_ONE_MOVE_ONLY);

		my_assert(*value >= -value_checkmate && *value <= value_checkmate, "Brain::calculateMove sane eval score");
		sd.cs.addCounts(&gcs);
		return;
	}
#if 0
for(size_t i=0; i<ml -> size(); i++) {
	if (ml -> at(i).score != -INFINITE)
		printf("hier %d\n", ml -> at(i).score);
}
#endif

	sd.s = &s;
	syzygyThreadPars_t stp;
	stp.s = new Scene(s);
	stp.cs = &sd.cs;
	stp.c = c;
	stp.valid = false;
	std::thread *const stpThread = sepc ? new std::thread(syzygyThread, &stp) : NULL;

	std::thread *t = NULL;

	if (durationMax >= 0)
		t = new std::thread(timerThread, startTs + durationMax, ei, io);

	std::vector<std::atomic_bool *> tafs;
	std::vector<thread_pars_t> tresults;
	for(int i=0; i<nThreads - 1; i++) {
		tresults.push_back({ 1 + (i & 1), invalidMove, -infinite, -infinite });
		tafs.push_back(new std::atomic_bool(false));
	}

	std::vector<std::thread *> lsmp;
	int llaws = tune -> aspiration_window_size, rlaws = tune -> aspiration_window_size;
	for(int i=0; i<nThreads - 1; i++) {
		if (i & 1) {
			llaws = 5;
			do {
				llaws = rand() % (2 * tune -> aspiration_window_size);
			}
			while(llaws < 5);

			rlaws = 5;
			do {
				rlaws = rand() % (2 * tune -> aspiration_window_size);
			}
			while(rlaws < 5);
		}

		// scne, color, ei, tt, ep, threadnr, startdepth, aws
		lsmp.push_back(new std::thread(thread, this, s, c, tafs.at(i), tt, epMove, i, llaws, rlaws, &tresults, nThreads - 1));
	}

	int depth = 1, alpha = -infinite, beta = infinite;

	// memset(m, 0x00, sizeof *m);
	*value = INT_MAX;

	// FIXME list of books
	std::map<int, Book *> bookQuestions;
	askBookQuestions(&bookQuestions, &s, c);

	VectorType<PVnode_t> pv, tempPv, oldPv;

	sd.colorPlayedByEngine = c;
	sd.time_abort_flag = &ei -> flag;
	memset(&sd.cs, 0x00, sizeof(sd.cs));
	sd.oldPv = &oldPv;
#ifdef WITH_HBT
	memset(sd.hbt, 0x00, sizeof(sd.hbt));
#endif

	if (historyIn)
		sd.history = new std::vector<uint64_t>(*historyIn);
	else {
		sd.history = new std::vector<uint64_t>();

#ifdef REPETITION
		addToHistory(sd.history, hash.newHash);
#endif
	}

#if 0
	Move recap = invalidMove;
	if (isValidMove(lastOppMove) && lastOppMove.m.isCatchMove) {
		const Board *const b = s.getBoardConst();

		recap = findLeastValuableAttacket(*b, *ml, lastOppMove.m.tx, lastOppMove.m.ty);
	}

	Move mPrevBest[2] = { recap, invalidMove };
#else
	Move mPrevBest[2] = { invalidMove, invalidMove };
#endif

	Move repeatCheck = invalidMove;
	int repeatCheckCount = 0;

	int failLowBonus = 200;

	std::vector<uint64_t> nodeCounts;
	uint64_t prevNodeCount = 0;

	uint64_t deadlineTs = startTs + durationMin;
	int bestDepth = -1;

	while((depth <= maxDepth || maxDepth <= 0) && depth < depth_sanity_limit && !ei -> flag) {
		uint64_t curStartTs = get_ts_ms();
		int timeLeft = int(deadlineTs - curStartTs);

		if (ponderThreadHandle == NULL && durationMin >= 0)
			log(ll_ok, ps + "current depth: %d, time left: %ld", depth, timeLeft);
		else
			log(ll_ok, ps + "current depth: %d", depth);

		if (durationMin > 0 && timeLeft < durationMin / 2) {
			if (ponderThreadHandle)
				log(ll_ok, ps + "abort search with %ld time left", timeLeft);
			else
				io -> emitDebug("abort search with %ld time left", timeLeft);
			break;
		}

		uint64_t cur_n_nodes = gcs.qcnt + gcs.cnt;
		if (stp.valid && stp.moves.empty() == false)
		{
			log(ll_ok, "SYZYGY HIT IN ROOT");

			VectorType<Move> putInFront;
			reorderAndSortList(&stp.moves, putInFront, &sd, c);

			VectorType<PVnode_t> temp;
			push_back(&temp, hash.newHash, stp.moves.at(0), stp.score, MS_SYZYGY);
			(*result) -> assign(temp, false);

			*value = stp.score;

			io -> emitThinking(moveNr, stp.depth, sd.cs.max_depth, *value, (ULI)(curStartTs - startTs), (ULI)cur_n_nodes, temp, false, -1, &sd.cs, &s, c, alpha, beta);
			break;
		}

		VectorType<PVnode_t> *bml = NULL;
		bool isFinalAnswer = false;
		if (checkBookQuestions(&bookQuestions, &s, c, &bml, &isFinalAnswer))
		{
			Move m = bml -> at(0).m;

			// store shallow eval in TT with depth 0
			constexpr int bdepth = 0;
			int score = evaluate(hash, sd.s, &sd.cs, c, bdepth, c);
			updateTT(tt, hash.newHash, bdepth, alpha, beta, score, m, true);

			// inform user
			std::string bpv = pvToString(*bml, false, false, false);
			std::string msg = myformat("Using book-hit %s (%d)", bpv.c_str(), *value);

			if (ponderThreadHandle)
				io -> emitDebug(msg);
			else
				io -> tellAll(msg);

			uint64_t curEndTs = get_ts_ms();
			int64_t tDiff = std::max(uint64_t(1), curEndTs - startTs);

			size_t bDepth = bml -> size();
			io -> emitThinking(moveNr, bDepth, sd.cs.max_depth, *value, (ULI)tDiff, 0, *bml, false, -1, &gcs, &s, c, alpha, beta);

			// store result
			if (ponderThreadHandle || !isFinalAnswer) {
				log(ll_ok, "Using book-hit as suggestion");
				mPrevBest[0] = m;
			}
			else {
				log(ll_ok, "Using book-hit as final result");
				(*result) -> assign(*bml, false);
				delete bml;

				break;
			}
		}

		sd.curMaxDepth = depth;

		sd.cs.max_depth = 0;

		int v = search(&sd, false, depth, c, alpha, beta, hash, mPrevBest, false, &pv);

		sd.cs.addCounts(&gcs);
		memset(&sd.cs, 0x00, sizeof sd.cs);

		// do not use score if it is not completely calculated
		if (ei -> flag)
		{
			log(ll_ok, ps + "stop search due to timeout");
			break;
		}

		if (abs(v) != infinite) // FIXME assert? because stopflag?
			*value = v;

		oldPv.assign(pv, false);

		uint64_t curEndTs = get_ts_ms();
		int64_t tDiff = std::max(uint64_t(1), curEndTs - startTs);

		cur_n_nodes = gcs.qcnt + gcs.cnt;

		Move selMove = pv.empty() ? invalidMove : pv.back().m;
		log(ll_ok, ps + "move: %s (%d), alpha/score/beta: %d/%d/%d, took %.2f, %.2fMnps, pv: %s", moveToString(selMove).c_str(), depth, alpha, v, beta, double(tDiff) / 1000.0, double(cur_n_nodes) / double(tDiff), pvToString(pv, true, true, false).c_str());

		nodeCounts.push_back(cur_n_nodes - prevNodeCount);
		prevNodeCount = cur_n_nodes;

		double ebf = calcEBF(nodeCounts);
		io -> emitThinking(moveNr, depth, gcs.max_depth, v, (ULI)tDiff, (ULI)cur_n_nodes, pv, true, ebf, &gcs, &s, c, alpha, beta);

		unsigned int absScore = abs(v);
		if (absScore >= value_checkmate_min) {
			std::string msg = myformat("checkmate detected %d", value_checkmate - abs(v));

			log(ll_ok, ps + msg);
			io -> tellAll(msg);
		}

#ifdef REPEAT_CHECK
		if (isValidMove(selMove) && ponderThreadHandle == NULL && targetMoves == NULL) {
			if (moveCompareTo(repeatCheck, selMove))
			{
				if (++repeatCheckCount > IT_DEEPENING_MAX_REPEAT_MOVE_SEL)
				{
					log(ll_ok, ps + "search abort by repeat-check-count (move: %s)", moveToString(selMove).c_str());
					break;
				}
			}
			else
			{
				repeatCheck = selMove;
				repeatCheckCount = 0;
			}
		}
#endif

		if (v <= alpha)
			alpha = -infinite;
		else if (v >= beta) {
			log(ll_ok, "Fail low, increasing time");
			deadlineTs = startTs + durationMax;
			beta = infinite;
		}
		else {
			alpha = v - tune -> aspiration_window_size;
			if (alpha < -infinite)
				alpha = -infinite;

			beta = v + tune -> aspiration_window_size;
			if (beta > infinite)
				beta = infinite;

			if (pv.empty() || !isValidMove(selMove) || abs(v) == infinite)
				log(ll_ok, ps + "search returned with no results");
			else
			{
				if (tt)
					log(ll_ok, ps + "Using pv: %s", recreatePvFromHash(tt, &s, c, selMove, true, false).c_str());

				updateBestSiblings(mPrevBest, pv);

				(*result) -> assign(pv, true);
				bestDepth = depth;

				if (targetMoves != NULL && indexOf(*targetMoves, selMove) != -1)
				{
					log(ll_ok, ps + "Target move found, stopping search");
					break;
				}

				if (depth > value_checkmate - abs(v) && !ponderThreadHandle && !targetMoves) {
					log(ll_ok, ps + "Mate detected, stopping search");
					break;
				}
			}

			depth++;
		}
	}

	stopBookQuestions(&bookQuestions);

	for(std::atomic_bool *taf : tafs)
		*taf = true;

	ei -> flag = true;
	ei -> cv.notify_all();

	if (stpThread) {
		stpThread -> join();
		delete stpThread;
	}

	delete stp.s;

	for(std::thread *cls : lsmp) {
		cls -> join();
		delete cls;
	}

	for(std::atomic_bool *taf : tafs)
		delete taf;

	for(int i=0; i<nThreads - 1; i++) {
		if (tresults.at(i).depth >= bestDepth && tresults.at(i).score > *value) {
			std::string betterMove = moveToString(tresults.at(i).m);
			std::string old = !(*result)->empty() ? moveToString((*result)->back().m) : "";
			log(ll_ok, "Better result from thread %d, depth %d (%d), score %d (%d), move %s (%s)", i + 1, tresults.at(i).depth, bestDepth, tresults.at(i).score, *value, betterMove.c_str(), old.c_str());
			(*result)->clear();
			*value = tresults.at(i).score;
			bestDepth = tresults.at(i).depth;
			push_back(*result, hash.newHash, tresults.at(i).m, *value, MS_THREAD);
		}
	}

       if (!(*result)->empty())
               learnBookQuestions(*s.getBoard(), c, epMove, (*result)->at(0).m, (*result) -> size(), WIN_UNKNOWN);

	if ((*result) -> empty()) {
		log(ll_err, ps + "no move selected! picking a random one from list!");

		*value = -666;

		size_t n = ml -> size();
		for(size_t i=0; i<n; i++) {
			if (s.validateMoves(c, i)) {
				push_back(*result, hash.newHash, ml -> at(i), *value, MS_EMERGENCY);
				break;
			}
		}
	}

//	double ebf = calcEBF(nodeCounts);

	for(const thread_pars_t & tp: tresults)
		tp.tcs.addCounts(&gcs);

	uint64_t ts_diff = get_ts_ms() - gcs.getStartedAt();
	double nps = double(gcs.qcnt + gcs.cnt) / double(ts_diff) * 1000.0;
//	log(ll_ok, ps + "nps %f (q: %lu, reg: %lu, q limit: %lu), total thread count: %ld, threads/s: %f, qqtpt: %lu, qtpt: %lu, tpt: %lu, tpt-err: %lu, nodes: %lu, nullmove: %lu, iid: %lu, se_hit: %lu/%f%% (%f%% tpt, %f%% of all threads), ebf: %f, lmr ok/fail: %lu/%lu, max-depth: %d", nps, (ULI)gcs.qcnt, (ULI)gcs.cnt, (ULI)gcs.q_limit, (ULI)gcs.threads_created, double(gcs.threads_created) / double(ts_diff) / 1000.0, (ULI)gcs.qqtpt_hit_cnt, (ULI)gcs.qtpt_hit_cnt, (ULI)gcs.tpt_hit_cnt, (ULI)gcs.tt_err, (ULI)gcs.cnt, (ULI)gcs.null_move_hit, (ULI)gcs.iid, (ULI)gcs.se_hit, double(gcs.se_hit * 100) / double(gcs.se_query), double(gcs.se_hit_move * 100) / double(gcs.se_query), double(gcs.se_query_thread * 100) / double(gcs.threads_created), ebf, (ULI)gcs.lmr_ok, (ULI)gcs.lmr_fail, gcs.max_depth);

	io->emitDebug("NPS: %f %f", nps / double(nThreads), nps);

	log(ll_ok, "cut off avg index: %f, nodes: %lu, cut nodes: %f", gcs.cnt_cutoff_index / double(gcs.cnt_cutoff), (ULI)gcs.cnt, gcs.cnt_cutoff / double(gcs.cnt));
	log(ll_ok, "qs cut off avg index: %f, nodes: %lu", gcs.q_nm / double(gcs.qcnt), (ULI)gcs.qcnt);

#ifdef USE_TT
	log(ll_ok, "TPT full: %d‰", tt -> hashFullPermill());
#endif

	if (!ponderThreadHandle)
		io -> emitDebug("nps: %dk, eval: %d, depth: %d, time: %.3f", int((nps + 1023) / 1024), *value, depth - 1, ts_diff / 1000.0);

	if (t) {
		t -> join();
		delete t;
	}

	delete sd.history;

	if (nodesVisited)
		*nodesVisited = gcs.qcnt + gcs.cnt;

	if ((*result) -> empty())
		log(ll_err, "calculation finished, no results?! ***");
	else
		log(ll_ok, "calculation finished %s %d ***", moveToString((*result) -> back().m).c_str(), *value);
}

// FIXME select move to ponder on
void Brain::startPonder(const int moveNr, const Scene & sIn, const PlayerColor c, const Move epMove, const std::vector<uint64_t> *const historyIn, const Move oppMove)
{
	log(ll_ok, "start ponder thread");

	ponderMoveNr = moveNr;
	ponderScene = new Scene(sIn);
	ponderColor = c;
	ponderEPMove = epMove;
	ponderei.flag = false;
	ponderHistory = std::vector<uint64_t>(*historyIn);
	ponderOppMove = oppMove;

	ponderPVlock.lock();
	ponderPV.clear();
	ponderPVlock.unlock();

	ponderThreadHandle = new std::thread(std::ref(*this));
}

void Brain::operator()()
{
	log(ll_ok, "Ponder thread started");

	int value = 0;
	uint64_t cnts = 0;

	VectorType<PVnode_t> *temp = NULL;
	calculateMove(ponderMoveNr, *ponderScene, ponderColor, -1, -1, -1, NULL, &ponderHistory, ponderEPMove, &temp, &value, &ponderei, &cnts, invalidMove, ponderOppMove);

	if (temp) {
		ponderPVlock.lock();
		ponderPV.assign(*temp, false);
		ponderPVlock.unlock();

		delete temp;
	}

	log(ll_ok, "Ponder thread terminating with pv: %s (%d)", pvToString(ponderPV, false, false, false).c_str(), value);
}

void Brain::stopPonder(VectorType<PVnode_t> *const target)
{
	if (target)
		target -> clear();

	if (ponderThreadHandle)
	{
		ponderei.flag = true;
		ponderei.cv.notify_all();

		ponderThreadHandle -> join();

		if (target) {
			ponderPVlock.lock();
			target -> assign(ponderPV, false);
			ponderPVlock.unlock();
		}

		delete ponderThreadHandle;
		ponderThreadHandle = NULL;

		delete ponderScene;
	}
}
