// (C) Folkert van Heusden. AGPL v3.0.
// $Revision: 5130 $
#ifndef __BRAIN_H__
#define __BRAIN_H__

#include <atomic>
#include <condition_variable>
#include <map>
#include <memory>
#include <mutex>
#include <random>
#include <thread>
#include <unistd.h>
#include <vector>

#include "Exception.h"
#include "config.h"
#include "Gen.h"
#include "Scene.h"
#include "Tpt.h"
#include "Ptt.h"
#include "Semaphore.h"
#include "IO.h"
#include "Book.h"
#include "Rng.h"
#include "PV.h"

extern char REV_BRAIN[];

bool evalResult(const int val, int *const restrict bestVal, int *const restrict alpha, int *const restrict beta, bool *const betaCutOff);

typedef struct
{
	PlayerColor colorPlayedByEngine;
	std::atomic_bool *restrict time_abort_flag;
	Scene *restrict s;
	int curMaxDepth;
	calc_stats_t cs;
	std::vector<uint64_t> *restrict history;
	const VectorType<PVnode_t> *restrict oldPv;
#ifdef WITH_HBT
	unsigned int hbt[2][64][64];
#endif
} search_data_t;

typedef struct
{
	uint64_t captures, ep, castles, promotions, checks, checkmates, stalemates, imd, whiteWin;
} perft_counts_t;

std::string perftCountsToStr(const perft_counts_t *const pc);
void sumPerftCounts(perft_counts_t *const tgt, const perft_counts_t *const src);

int scorePassedPawns(const ChessPiece * const *restrict whiteObjects, const ChessPiece * const *restrict blackObjects);

bool checkSyzygy(const Scene *const sd, calc_stats_t *const cs, const PlayerColor c, int *const score, VectorType<Move> *const cm, int *const depth, const bool async, const bool quick);

class Brain final
{
private:
	Tpt *const restrict tt;
	Ptt *restrict ptt;
	//
	const bool sepc; // syzygy_etb_piece_count > 0
	//
	std::thread *ponderThreadHandle;
	Scene *ponderScene;
	PlayerColor ponderColor;
	Move ponderEPMove, ponderOppMove;
	end_indicator_t ponderei;
	VectorType<PVnode_t> ponderPV;
	std::mutex ponderPVlock;
	std::vector<uint64_t> ponderHistory;
	int ponderMoveNr;

	bool statsLogging;

	IO *io;

	std::vector<Book *> *books;

	int nThreads;

	std::random_device rd;
	MERS_RNG *mt;

	calc_stats_t gcs;

	bool evalTT(search_data_t *const sd, const PlayerColor c, Tpt *const restrict curtt, const int depthLeft, const uint64_t hash, const int alpha, const int beta, int *const restrict useValue, Move *const restrict m, bool *const useNm, const int cursd) const;
	void updateTT(Tpt *const restrict curtt, const uint64_t hash, const int ttDepth, const int alphaOrig, const int beta, const int bestVal, const Move & m, const bool useNm) const;

#ifdef QUIESCENCE
#endif

public:
	Brain(IO *io, Tpt * const tt, Ptt *const ptt, const int nThreads, std::vector<Book *> *const books);
	~Brain();

	void resetStats();

	void registerIO(IO *const ioIn);
	static const std::map<std::string, std::string> & getParameters();
	bool setParam(const std::string & key, const std::string & value);
	void setNTHreads(const int n);
	int getNTHreads() const { return nThreads; }

	void askBookQuestions(std::map<int, Book *> *const bookQuestions, Scene *const s, const PlayerColor c);
	void stopBookQuestions(std::map<int, Book *> *const bookQuestions);
	bool checkBookQuestions(std::map<int, Book *> *const bookQuestions, Scene *const s, const PlayerColor c, VectorType<PVnode_t> **const ml, bool *const isFinalAnswer);
	void learnBookQuestions(const Board & b, const PlayerColor c, const Move & epMove, const Move & bestMove, const int depth, const winner_t w);

	zobrist_t calcZobrist(const Board * const restrict b, const PlayerColor c, const Move epMove);
	zobrist_t updateZobrist(const zobrist_t & in, const BoardState & bs, const Board *const restrict b, const PlayerColor c, const Move epMove);

	int countTotalObjectsValue(const int counts[]);
	int evaluate(const zobrist_t & nh, const Scene *const s, calc_stats_t *const cs, const PlayerColor me, const int currentSearchDistance, const PlayerColor localColor);
	int calcShannon(const zobrist_t & nh, const Scene *const s, calc_stats_t *const cs, const PlayerColor me, const PlayerColor localColor);

	unsigned long int perft(Scene & s, const int depthLeft, const PlayerColor c, perft_counts_t *const pc, const bool isDivide);

	void calculateMove(const int moveNr, Scene & s, const PlayerColor c, const int maxDepth, const int durationMin, const int durationMax, const VectorType<Move> *const targetMoves, std::vector<uint64_t> *const history, const Move epMove, VectorType<PVnode_t> **const result, int *const value, end_indicator_t *const ei, uint64_t *const nodesVisited, const Move ponderHit, const Move lastOppMove);

	int search(search_data_t *const restrict sd, const bool cbcm, const int depthLeft, const PlayerColor c, int alpha, int beta, const zobrist_t & zh, const Move siblingPrevBestMoves[2], const bool isNullMove, VectorType<PVnode_t> *const restrict pv);
	int quiesceSearch(search_data_t *const restrict sd, const int currentSearchDistance, const PlayerColor c, int alpha, int beta, const zobrist_t & zh, const Move prevBestMoves[2], Move *const restrict pv);

	void startPonder(const int moveNr, const Scene & sIn, const PlayerColor c, const Move epMove, const std::vector<uint64_t> *const historyIn, const Move oppMove);
	void operator()();
	void stopPonder(VectorType<PVnode_t> *const target);
};

#endif
