// (C) Folkert van Heusden. AGPL v3.0.
// $Revision: 4815 $
#include "Tweaks.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "Debug.h"
#include "Exception.h"
#include "ChessPiece.h"
#include "Board.h"
#include "Utils.h"
#include "Tweaks.h"
#include "Gen.h"

bool compareChessPieceState(const ChessPieceState & a, const ChessPieceState &b)
{
	return a.x == b.x && a.y == b.y && a.first_move == b.first_move && a.alive == b.alive && a.cpt == b.cpt;
}

const int ChessPiece::evalVal[] = { 100, 325, 325, 500, 975, 10000, 0, 0 };

ChessPiece::ChessPiece(const PlayerColor c, const int xIn, const int yIn, const bool first_move, const ChessPieceType cpt)
{
	my_assert(yIn >= 0 && yIn <= 7, "ChessPiece: y within limits");
	my_assert(xIn >= 0 && xIn <= 7, "ChessPiece: x within limits");

	initialize(c, xIn, yIn, first_move, cpt);
}

void ChessPiece::initialize(const PlayerColor c, const int xIn, const int yIn, const bool first_move, const ChessPieceType cpt)
{
	my_assert(yIn >= 0 && yIn <= 7, "ChessPiece: y within limits");
	my_assert(xIn >= 0 && xIn <= 7, "ChessPiece: x within limits");

	mask = (uint32_t)-1;

	color = c;

	id = -1;

	memset(&state, 0x00, sizeof state);
	state.x = xIn;
	state.y = yIn;

	state.cpt = cpt;

	state.first_move = first_move;
	state.alive = true;
}

ChessPiece::ChessPiece(const ChessPiece & cIn) : id(cIn.getId()), color(cIn.color), state(cIn.state)
{
	mask = (uint32_t)-1;

	const unsigned long int n_cps = cIn.cps.size();

	if (n_cps)
		this -> cps.push_back(cIn.cps.at(n_cps - 1));
}

void ChessPiece::initialize(const ChessPiece & cIn)
{
	mask = (uint32_t)-1;

	color = cIn.color;
	this -> state = cIn.state;

	const unsigned long int n_cps = cIn.cps.size();

	if (n_cps)
		this -> cps.push_back(cIn.cps.at(n_cps - 1));
}

ChessPiece::~ChessPiece()
{
}

void ChessPiece::dump() const
{
	switch(getType())
	{
		case PAWN:
			printf("pawn");
			break;
		case ROOK:
			printf("rook");
			break;
		case BISHOP:
			printf("bishop");
			break;
		case KNIGHT:
			printf("knight");
			break;
		case QUEEN:
			printf("queen");
			break;
		case KING:
			printf("king");
			break;
		case NONE:
			printf("!!");
			break;
		default:
			printf("?!");
			break;
	}

	printf(" alive: %d, first move: %d, x: %d, y: %d, color: %s\n",
		state.alive, state.first_move, state.x, state.y, color == WHITE ? "white" : "black");
}

#ifdef _DEBUG 
void ChessPiece::resetMoveState(ChessPieceState * const restrict pcps)
{
        pcps -> move_related.n_can_be_caught_by = 0;
        pcps -> move_related.valid = false;
}
#endif

void ChessPiece::pushState()
{
	cps.push_back(state);

#ifdef _DEBUG 
	resetMoveState(&state);
#endif
}

void ChessPiece::popState()
{
	my_assert(cps.size() > 0, "ChessPiece: state stack underflow");

	state = cps.back();

#ifdef _DEBUG
	cps.back() = { -1, -1, NONE, true, true, true, -1, { -1, true } };
#endif

	cps.pop_back();
}

ChessPieceState ChessPiece::getState() const
{
	return state;
}

void ChessPiece::doMove(const int toX, const int toY)
{
	my_assert(toY >= 0 && toY <= 7, "ChessPiece::doMove: y within limits");
	my_assert(toX >= 0 && toX <= 7, "ChessPiece::doMove: x within limits");
	my_assert(state.alive, "ChessPiece::doMove must be alive");
	my_assert(state.x != toX || state.y != toY, "ChessPiec::doMove must be different location");

	state.x = toX;
	state.y = toY;

	state.first_move = false;

	state.was_promotion = (toY == 0 || toY == 7) && state.cpt == PAWN;
}

void ChessPiece::doKill()
{
	my_assert(state.alive, "ChessPiece::doMove must be alive");

	state.alive = false;
}

void ChessPiece::setXY(const int newX, const int newY)
{
	my_assert(newY >= 0 && newY <= 7, "ChessPiece::setXY: y within limits");
	my_assert(newX >= 0 && newX <= 7, "ChessPiece::setXY: x within limits");
	my_assert(state.alive, "ChessPiece::setXY must be alive");

	state.x = newX;
	state.y = newY;
}

void ChessPiece::getXY(int *const restrict x, int *const restrict y) const
{
	my_assert(state.alive, "ChessPiece::getXY must be alive");

	*x = state.x;
	*y = state.y;
}

int ChessPiece::getX() const
{
	my_assert(state.alive, "ChessPiece::getX must be alive");

	return state.x;
}

int ChessPiece::getY() const
{
	my_assert(state.alive, "ChessPiece::getY must be alive");

	return state.y;
}

bool ChessPiece::isUnderAttack(const uint32_t * const restrict bitmap) const
{
	return *bitmap & mask;
}

bool ChessPiece::getAlive() const
{
	return state.alive;
}

void ChessPiece::setAlive(const bool alive)
{
	state.alive = alive;
}

PlayerColor ChessPiece::getColor() const
{
	return color;
}

void ChessPiece::promote(const ChessPieceType newType)
{
	my_assert(state.alive, "ChessPiece::promote must be alive");
	my_assert(state.cpt == PAWN, "ChessPiece::promote current type must be pawn");
	my_assert(newType != PAWN, "ChessPiece::promote new type must not be pawn");
	my_assert((state.y == 0 && color == BLACK) || (state.y == 7 && color == WHITE), "ChessPiece::promote: promote on correct line");

	state.cpt = newType;
}

ChessPieceType ChessPiece::getType() const
{
	return state.cpt;
}

void ChessPiece::setType(const ChessPieceType cpt)
{
	state.cpt = cpt;
}

int ChessPiece::getEvalVal() const
{
	my_assert(state.cpt >=0 && state.cpt < 6, "ChessPiece::getEvalVal: cpt valid value");

	return evalVal[state.cpt];
}

bool ChessPiece::getFirstMove() const
{
	return state.first_move;
}

void ChessPiece::clearFirstMove()
{
	my_assert(state.first_move, "ChessPiece::clearFirstMove tried to clear first-move state while not set");

	state.first_move = false;
}

void ChessPiece::setFirstMove()
{
	state.first_move = true;
}

#ifdef _DEBUG
bool ChessPiece::getMoveRelatedValid() const
{
	return state.move_related.valid;
}
#endif

void ChessPiece::setKingPinned(uint32_t *const restrict bitmap) const
{
	*bitmap |= mask;
}

bool ChessPiece::getIsKingPinned(const uint32_t * const restrict bitmap) const
{
	return *bitmap & mask;
}

int ChessPiece::generateMoveList(gml_pars_t *const restrict pars)
{
#ifdef _DEBUG
	my_assert(!state.move_related.valid, "ChessPiece::generateMoveList: list of moves should not already be known");

	uint32_t isKingPinned = *pars -> isKingPinned;
#endif

	size_t before_n = pars -> ml -> size();
	state.move_count = 0;

	// replace by a std::function implementation FIXME

	switch(getType())
	{
		case PAWN:
			generateMoveListPawn(pars);
			break;

		case KNIGHT:
			generateMoveListKnight(pars);
			break;

		case BISHOP:
			generateMoveListBishop(pars);
			break;

		case ROOK:
			generateMoveListRook(pars);
			break;

		case QUEEN:
			generateMoveListQueen(pars);
			break;

		case KING:
			generateMoveListKing(pars);
			break;

		case NONE:
		default:
			my_assert(false, "ChessPiece::generateMoveList: invalid type");
	}

#ifdef _DEBUG
	int nikp = countBits1(*pars -> isKingPinned) - countBits1(isKingPinned);
	my_assert(nikp >= 0 && nikp <= 1, "isKingPinned");


	for(size_t i=0; i < pars -> ml -> size(); i++) {
		my_assert(isValidMove(pars -> ml -> at(0)), "addMove");
	}
#endif

	size_t after_n = pars -> ml -> size();
	state.move_count = after_n - before_n;

#ifdef _DEBUG
	state.move_related.valid = true;
#endif

	return state.move_count;
}

void ChessPiece::registerPiece(uint32_t *const tcs, const int x, const int y) const
{
	tcs[(y * 8) + x] |= mask;
}

void ChessPiece::addMove(gml_pars_t *const restrict pars, ChessPiece * const restrict pa, const int tx, const int ty, const MoveType moveType) const
{
	pars -> ml -> push_back({ { uint8_t(state.x), uint8_t(state.y), uint8_t(tx), uint8_t(ty), moveType, NONE, pa }, -MS_INF });

	registerPiece(pars -> tcs, tx, ty);
}

void ChessPiece::addMove(gml_pars_t *const restrict pars, const int tx, const int ty) const
{
	pars -> ml -> push_back({ { uint8_t(state.x), uint8_t(state.y), uint8_t(tx), uint8_t(ty), REGULAR, NONE, false }, -MS_INF });

	registerPiece(pars -> tcs, tx, ty);
}

void ChessPiece::addMovePromoteTo(gml_pars_t *const restrict pars, Move m, const ChessPieceType promoteTo) const
{
	m.u.m.promoteTo = promoteTo;

	my_assert(m.u.m.promoteTo == ROOK || m.u.m.promoteTo == BISHOP || m.u.m.promoteTo == KNIGHT || m.u.m.promoteTo == QUEEN, "sane ChessPiece::addMovePromoteTo");

	pars -> ml -> push_back(m);
}

void ChessPiece::addEPMove(gml_pars_t *const restrict pars, const int toX, const int toY, ChessPiece * const restrict victim) const
{
	pars -> ml -> push_back({ { uint8_t(state.x), uint8_t(state.y), uint8_t(toX), uint8_t(toY), EN_PASSANT, NONE, true }, -MS_INF });

	registerPiece(pars -> tcs, toX, toY);
}

void ChessPiece::addPromotionMove(gml_pars_t *const restrict pars, ChessPiece * const restrict pa, const int toX, const int toY) const
{
	if (unlikely(toY == 7 || toY == 0)) {
		registerPiece(pars -> tcs, toX, toY);

		Move m = { { uint8_t(state.x), uint8_t(state.y), uint8_t(toX), uint8_t(toY), REGULAR, NONE, pa }, -MS_INF };

		addMovePromoteTo(pars, m, QUEEN);
#ifdef REQUIRE_ALL_POSSIBLE_MOVES
		addMovePromoteTo(pars, m, BISHOP);
		addMovePromoteTo(pars, m, ROOK);
#endif
		addMovePromoteTo(pars, m, KNIGHT);
	}
	else {
		addMove(pars, pa, toX, toY, REGULAR);
	}
}

void ChessPiece::generateMoveListPawn(gml_pars_t *const restrict pars)
{
	/* en passant */
	/* is that last moved object a pawn which did it's first move? */
	/* initial move 2 in one? destination is empty? */
	if (unlikely(pars -> lastOpponentMove.u.m.ty == state.y && abs(pars -> lastOpponentMove.u.m.tx - state.x) == 1))
	{
		const int destX = pars -> lastOpponentMove.u.m.tx;
		const int destY = color == WHITE ? 5 : 2;

		addEPMove(pars, destX, destY, pars -> b -> getAt(pars -> lastOpponentMove.u.m.tx, pars -> lastOpponentMove.u.m.ty));
	}

	const int yPlus1 = color == WHITE ? state.y + 1 : state.y - 1;

	if (pars -> b -> getAt(state.x, yPlus1) == NULL)
	{
		addPromotionMove(pars, NULL, state.x, yPlus1);

		if (getFirstMove()) {
			const int yPlus2 = color == WHITE ? state.y + 2 : state.y - 2;

			if (pars -> b -> getAt(state.x, yPlus2) == NULL)
				addMove(pars, state.x, yPlus2);
		}
	}

	if (likely(state.x))
	{
		int leftX = state.x - 1;
		ChessPiece *restrict obj = pars -> b -> getAt(leftX, yPlus1);

		if (obj) {
			if (obj -> getColor() != color)
				addPromotionMove(pars, obj, leftX, yPlus1);
			else
				registerPiece(pars -> tcs, leftX, yPlus1);
		}
	}

	if (likely(state.x < 7))
	{
		int rightX = state.x + 1;
		ChessPiece *restrict obj = pars -> b -> getAt(rightX, yPlus1);

		if (obj) {
			if (obj -> getColor() != color)
				addPromotionMove(pars, obj, rightX, yPlus1);
			else
				registerPiece(pars -> tcs, rightX, yPlus1);
		}
	}
}

void ChessPiece::getMovesPath(gml_pars_t *const restrict pars, const int deltaX, const int deltaY) const
{
	ChessPiece *pinnedPiece = NULL;

	// NOTE: using unsigned variablexs for x/y, that way
	// if it beomes negative, it will be a really large unsigned value > 7
	// that way we only have to check if the value became > 7
	// deltaX/Y will never be so large that their underflow will end up
	// in the range of 0...7
	unsigned int cx = state.x + deltaX, cy = state.y + deltaY;

	const Board *const b = pars -> b;

	// http://stackoverflow.com/questions/18994563/is-unsigned-underflow-and-signed-overflow-defined-behaviour
	// for unsigned integers, overflow is defined behaviour
	while(likely(cx < 8 && cy < 8))
	{
		ChessPiece *const destField = b -> getAt(cx, cy);

		if (destField)
		{
			my_assert(destField -> getAlive(), "ChessPiece::getMovesPath: target piece must be alive");

			if (likely(pinnedPiece == NULL))
			{
				// if this piece is of the opponent, remember this
				// move and continue searching for another of the
				// opponent. then if another is found, this piece
				// (destField) is pinned.
				if (destField -> getColor() != color)
				{
					addMove(pars, destField, cx, cy, REGULAR);

					pinnedPiece = destField;
				}
				// if the piece we encounter is one of mine, then
				// stop searching
				else {
					registerPiece(pars -> tcs, cx, cy);
					break;
				}
			}
			else
			{
				if (unlikely(destField == pars -> oking))
					pinnedPiece -> setKingPinned(pars -> isKingPinned);

				break;
			}
		}

		if (pinnedPiece == NULL)
			addMove(pars, cx, cy);

		cx += deltaX;
		cy += deltaY;
	}
}

void ChessPiece::getMovesPathNoPin(gml_pars_t *const restrict pars, const int deltaX, const int deltaY) const
{
	unsigned int cx = state.x + deltaX, cy = state.y + deltaY;

	const Board *const b = pars -> b;

	while(likely(cx < 8 && cy < 8))
	{
		ChessPiece *const destField = b -> getAt(cx, cy);

		if (destField)
		{
			my_assert(destField -> getAlive(), "ChessPiece::getMovesPath: target piece must be alive");

			if (destField -> getColor() != color)
				addMove(pars, destField, cx, cy, REGULAR);
			else
				registerPiece(pars -> tcs, cx, cy);

			break;
		}

		addMove(pars, cx, cy);

		cx += deltaX;
		cy += deltaY;
	}
}

void ChessPiece::getMovesPaths(gml_pars_t *const restrict pars, const int paths[][2], const int nMaps, const bool scanPinned) const
{
	if (scanPinned) {
		for(int i=0; i<nMaps; i++)
			getMovesPath(pars, paths[i][0], paths[i][1]);
	}
	else {
		for(int i=0; i<nMaps; i++)
			getMovesPathNoPin(pars, paths[i][0], paths[i][1]);
	}
}

void ChessPiece::generateMoveListBishop(gml_pars_t *const restrict pars) const
{
	const ChessPiece *const king = pars -> oking;
	int xdiff = abs(king -> getX() - state.x);
	int ydiff = abs(king -> getY() - state.y);

	bool scanPinned = xdiff == ydiff;

	constexpr int dirs[][2] = {
		 { -1, -1 },
		 {  1, -1 },
		 { -1,  1 },
		 {  1,  1 }
		};

	getMovesPaths(pars, dirs, 4, scanPinned);
}

void ChessPiece::addFromDirsList(gml_pars_t *const restrict pars, const int dirs[][2]) const
{
	for(int i=0; i<8; i++)
	{
		const unsigned int tx = state.x + dirs[i][0];
		const unsigned int ty = state.y + dirs[i][1];

		if (likely(tx < 8 && ty < 8)) {
			ChessPiece *destField = pars -> b -> getAt(tx, ty);

			if (destField == NULL)
				addMove(pars, tx, ty);
			else if (destField -> getColor() != color)
				addMove(pars, destField, tx, ty, REGULAR);
			else
				registerPiece(pars -> tcs, tx, ty);
		}
	}
}

void ChessPiece::generateMoveListKnight(gml_pars_t *const restrict pars) const
{
	constexpr int dirs[][2] = {
		{ -1, -2 },
		{  1, -2 },
		{ -2, -1 },
		{  2, -1 },
		{ -2,  1 },
		{  2,  1 },
		{ -1,  2 },
		{  1,  2 }
		};

	addFromDirsList(pars, dirs);
}

void ChessPiece::generateMoveListKing(gml_pars_t *const restrict pars)
{
	constexpr int dirs[][2] = {
		{ -1, -1 },
		{  0, -1 },
		{  1, -1 },
		{ -1,  0 },
		{  1,  0 },
		{ -1,  1 },
		{  0,  1 },
		{  1,  1 }
		};

	addFromDirsList(pars, dirs);
}

void ChessPiece::generateMoveListQueen(gml_pars_t *const restrict pars) const
{
	const ChessPiece *const king = pars -> oking;
	int xdiff = abs(king -> getX() - state.x);
	int ydiff = abs(king -> getY() - state.y);

	bool scanPinned = xdiff == ydiff || ydiff == 0 || xdiff == 0;

	constexpr int dirs[][2] = {
		{ -1, -1 },
		{  0, -1 },
		{  1, -1 },
		{ -1,  0 },
		{  1,  0 },
		{ -1,  1 },
		{  0,  1 },
		{  1,  1 }
		};

	getMovesPaths(pars, dirs, 8, scanPinned);
}

void ChessPiece::addMoveCastling(gml_pars_t *const restrict pars, const int fx, const int fy, const int tx) const
{
	const int ty = fy;

	pars -> ml -> push_back({ { uint8_t(fx), uint8_t(fy), uint8_t(tx), uint8_t(ty), CASTLING, NONE, false }, -MS_INF });
}

void ChessPiece::generateMoveListRook(gml_pars_t *const restrict pars)
{
	const ChessPiece *const kingOp = pars -> oking;

	bool scanPinned = kingOp -> getX() == state.x || kingOp -> getY() == state.y;

	constexpr int dirs[][2] = {
		{  0, -1 },
		{ -1,  0 },
		{  1,  0 },
		{  0,  1 }
		};

	getMovesPaths(pars, dirs, 4, scanPinned);

	const ChessPiece *const king = pars -> mking;
	my_assert(king -> getColor() == getColor(), "ChessPiece::generateMoveListRook: king color must be mine");

	/* CASTLING */
	if (getFirstMove() && king -> getFirstMove())
	{
		my_assert(king -> getX() == 4 && ((king -> getY() == 0 && king -> getColor() == WHITE) || (king -> getY() == 7 && king -> getColor() == BLACK)), "ChessPiece::generateMoveListRook king at correct location in first move");
		my_assert(state.y == 0 || state.y == 7, "ChessPiece::generateMoveListRook must be on home-row");

		///////////////////////////////////////////////////////
		// IN VALIDATEMOVES FIELD 3,Y / 5,Y MUST BE CHECKED! //
		///////////////////////////////////////////////////////

		/* short/long castling? */
		if (state.x == 0)   /* long */
		{
			if (pars -> b -> getAt(1, state.y) == NULL && pars -> b -> getAt(2, state.y) == NULL && pars -> b -> getAt(3, state.y) == NULL) {
				addMoveCastling(pars, 4, state.y, 2);
				registerPiece(pars -> tcs, 2, state.y);
				registerPiece(pars -> tcs, 3, state.y);
			}
		}
		else  /* short */
		{
			if (pars -> b -> getAt(5, state.y) == NULL && pars -> b -> getAt(6, state.y) == NULL) {
				addMoveCastling(pars, 4, state.y, 6);
				registerPiece(pars -> tcs, 5, state.y);
				registerPiece(pars -> tcs, 6, state.y);
			}
		}
	}
}

void ChessPiece::setId(const int newId)
{
	my_assert(newId >= 0 && newId < 32, "ChessPiece::setId: must fit in uint32_t");

	id = newId;
	mask = 1 << newId;
}

#ifdef _DEBUG
void ChessPiece::prepareForGenerateMoveList()
{
	state.move_related.n_can_be_caught_by = 0;
	state.move_related.valid = false;
}
#endif

char ChessPiece::getTypeChar(const bool whiteUppercase) const
{
	static constexpr char names[] = "pnbrqk??";

	if (whiteUppercase && color == WHITE)
		return toupper(names[state.cpt]);

	return names[state.cpt];
}

bool ChessPiece::isUnderAttackByPawn(const isUnderAttackBy_pars_t *const restrict ibp) const
{
	const int yPlus1 = color == WHITE ? state.y + 1 : state.y - 1;

	if (yPlus1 == ibp -> ty)
		return state.x - 1 == ibp -> tx || state.x + 1 == ibp -> tx;

	return false;
}

bool ChessPiece::isUnderAttackScan(const isUnderAttackBy_pars_t *const restrict ibp, const int dx, const int dy) const
{
	my_assert(dx >= -2 && dx < 8, "ChessPiece::isUnderAttackScan");
	my_assert(dy >= -2 && dy < 8, "ChessPiece::isUnderAttackScan");

	unsigned x = state.x;
	unsigned y = state.y;

	for(;;) {
		x += dx;
		y += dy;

		if (x == (unsigned)ibp -> tx && y == (unsigned)ibp -> ty)
			return true;

		if (x >= 8 || y >= 8)
			break;

		if (ibp -> b -> getAt(x, y))
			break;
	}

	return false;
}

inline int vecToD(const int v)
{
	if (v >= 1)
		return 1;

	if (v <= -1)
		return -1;

	return 0;
}

bool ChessPiece::isUnderAttackByRook(const isUnderAttackBy_pars_t *const restrict ibp) const
{
	int dx = ibp -> tx - state.x;
	int dy = ibp -> ty - state.y;

	if (dx && dy)
		return false;

	return isUnderAttackScan(ibp, vecToD(dx), vecToD(dy));
}

bool ChessPiece::isUnderAttackByBishop(const isUnderAttackBy_pars_t *const restrict ibp) const
{
	int dx = ibp -> tx - state.x;
	int dy = ibp -> ty - state.y;

	if (dx == 0 || dy == 0)
		return false;

	return isUnderAttackScan(ibp, vecToD(dx), vecToD(dy));
}

bool ChessPiece::isUnderAttackByKnight(const isUnderAttackBy_pars_t *const restrict ibp) const
{
	constexpr int dirs[][2] = {
		{ -1, -2 },
		{  1, -2 },
		{ -2, -1 },
		{  2, -1 },
		{ -2,  1 },
		{  2,  1 },
		{ -1,  2 },
		{  1,  2 } 
		};

	for(int i=0; i<8; i++) {
		if (state.x + dirs[i][0] == ibp -> tx && state.y + dirs[i][1] == ibp -> ty)
			return true;
	}

	return false;
}

bool ChessPiece::isUnderAttackByQueen(const isUnderAttackBy_pars_t *const restrict ibp) const
{
	int dx = ibp -> tx - state.x;
	int dy = ibp -> ty - state.y;

	return isUnderAttackScan(ibp, vecToD(dx), vecToD(dy));
}

bool ChessPiece::isUnderAttackByKing(const isUnderAttackBy_pars_t *const restrict ibp) const
{
	constexpr int dirs[][2] = {
		 { -1, -1 },
		 {  0, -1 },
		 {  1, -1 },
		 { -1,  0 },
		 {  1,  0 },
		 { -1,  1 },
		 {  0,  1 },
		 {  1,  1 } 
		};

	for(int i=0; i<8; i++) {
		if (state.x + dirs[i][0] == ibp -> tx && state.y + dirs[i][1] == ibp -> ty)
			return true;
	}

	return false;
}

bool ChessPiece::isUnderAttackByThis(const isUnderAttackBy_pars_t *const restrict ibp) const
{
	my_assert(ibp -> tx >= 0 && ibp -> tx < 8, "ChessPiece::isUnderAttackScan");
	my_assert(ibp -> ty >= 0 && ibp -> ty < 8, "ChessPiece::isUnderAttackScan");
#ifdef _DEBUG
	ibp -> b -> verifyBoard();
#endif

	switch(state.cpt) {
		case PAWN:
			return isUnderAttackByPawn(ibp);
		case ROOK:
			return isUnderAttackByRook(ibp);
		case BISHOP:
			return isUnderAttackByBishop(ibp);
		case KNIGHT:
			return isUnderAttackByKnight(ibp);
		case QUEEN:
			return isUnderAttackByQueen(ibp);
		case KING:
			return isUnderAttackByKing(ibp);
		case NONE:
		default:
			return true;
	}

	return false;
}

void verifyCPS(const ChessPieceState & cps)
{
#ifdef _DEBUG
	my_assert(cps.x >= 0 && cps.x < 8, "verifyCPS");
	my_assert(cps.y >= 0 && cps.y < 8, "verifyCPS");
	my_assert(cps.cpt < 6 || cps.cpt == 7, "verifyCPS");
	my_assert((cps.first_move && !cps.was_promotion) || !cps.first_move, "verifyCPS");
	my_assert(cps.move_related.n_can_be_caught_by < 16 && cps.move_related.n_can_be_caught_by >= 0, "verifyCPS");
	//my_assert(cps.move_related.valid || !cps.alive, "verifyCPS");
#endif
}

void ChessPiece::validate()
{
#ifdef _DEBUG
	// not always possible to check at this point (see Scene putPieces) my_assert(id >=0 && id < 32, "ChessPiece::verifyPiece");
	my_assert(color == WHITE || color == BLACK, "ChessPiece::verifyPiece");

	verifyCPS(state);

	for(ChessPieceState cur : cps)
		verifyCPS(cur);
#endif
}
