// (C) Folkert van Heusden. AGPL v3.0.
// $Revision: 4673 $
#ifndef __CHESSPIECE_H__
#define __CHESSPIECE_H__

#include <stdint.h>
#include <vector>

#include "PlayerColor.h"
#include "ChessPieceType.h"
#include "MoveList.h"
#include "Tweaks.h"

class Board;

#include "Move.h"

class ChessPiece;
typedef struct
{
	VectorType<Move> *restrict ml;
	const Board *restrict b;
	Move lastOpponentMove;
	ChessPiece *restrict oking, *restrict mking;
	uint32_t *restrict isKingPinned;
	uint32_t *tcs;
} gml_pars_t;

typedef struct _cps_
{
	int x, y;
	ChessPieceType cpt; // in state because of promotion
	bool first_move, alive, was_promotion;
	int move_count;

#ifdef _DEBUG
	struct
	{
		int n_can_be_caught_by;
		bool valid;
	} move_related;
#endif
} ChessPieceState;

typedef struct {
	const Board *const restrict b;
	const int tx, ty;
} isUnderAttackBy_pars_t;

bool compareChessPieceState(const ChessPieceState & a, const ChessPieceState &b);

class ChessPiece final
{
private:
	int id;
	uint32_t mask; // is basically an index in the pinned/king-pinned/under-attack bitmap

protected:
	std::vector<ChessPieceState> cps;
	//
	PlayerColor color;
	ChessPieceState state;

	void resetMoveState(ChessPieceState * const restrict pcps);

	void registerPiece(uint32_t *const tcs, const int x, const int y) const;

	void addMove(gml_pars_t *const restrict pars, const int tx, const int ty) const;
	void addMove(gml_pars_t *const restrict pars, ChessPiece * const restrict pa, const int tx, const int ty, const MoveType moveType) const;

	// pawn
	void addEPMove(gml_pars_t *const restrict pars, const int toX, const int toY, ChessPiece * const restrict victim) const;
	void addMovePromoteTo(gml_pars_t *const restrict pars, Move m, const ChessPieceType promoteTo) const;
	void addPromotionMove(gml_pars_t *const restrict pars, ChessPiece *const restrict pa, const int toX, const int toY) const;
	void generateMoveListPawn(gml_pars_t *const restrict pars);

	// rook/bishop/queen
	bool conditionalAddMove(gml_pars_t *const restrict pars, const int tx, const int ty) const;
	void addFromDirsList(gml_pars_t *const restrict pars, const int dirs[][2]) const;
	void getMovesPaths(gml_pars_t *const restrict pars, const int paths[][2], const int nMaps, const bool scanPinned) const;
	void getMovesPath(gml_pars_t *const restrict pars, const int deltaX, const int deltaY) const;
	void getMovesPathNoPin(gml_pars_t *const restrict pars, const int deltaX, const int deltaY) const;

	// bishop
	void generateMoveListBishop(gml_pars_t *const restrict pars) const;

	// knight
	void generateMoveListKnight(gml_pars_t *const restrict pars) const;

	// king
	void generateMoveListKing(gml_pars_t *const restrict pars);

	// queen
	void generateMoveListQueen(gml_pars_t *const restrict pars) const;

	// rook
	void addMoveCastling(gml_pars_t *const restrict pars, const int fx, const int fy, const int tx) const;
	void generateMoveListRook(gml_pars_t *const restrict pars);

	void setKingPinned(uint32_t *const restrict bitmap) const;

	bool isUnderAttackByPawn(const isUnderAttackBy_pars_t *const restrict ibp) const;
	bool isUnderAttackScan(const isUnderAttackBy_pars_t *const restrict ibp, const int dx, const int dy) const;
	bool isUnderAttackByRook(const isUnderAttackBy_pars_t *const restrict ibp) const;
	bool isUnderAttackByBishop(const isUnderAttackBy_pars_t *const restrict ibp) const;
	bool isUnderAttackByKnight(const isUnderAttackBy_pars_t *const restrict ibp) const;
	bool isUnderAttackByQueen(const isUnderAttackBy_pars_t *const restrict ibp) const;
	bool isUnderAttackByKing(const isUnderAttackBy_pars_t *const restrict ibp) const;

public:
	ChessPiece(const PlayerColor c, const int xIn, const int yIn, const bool first_move, const ChessPieceType cpt);
	ChessPiece(const ChessPiece & cIn);
	ChessPiece() { }
	~ChessPiece();

	void initialize(const PlayerColor c, const int xIn, const int yIn, const bool first_move, const ChessPieceType cpt);
	void initialize(const ChessPiece & cIn);

	int getId() const { return id; }
	void setId(const int newId);

	void dump() const;

	int getX() const;
	int getY() const;
	void getXY(int *const restrict x, int *const restrict y) const;
	void setXY(const int new_x, const int new_y);
	bool getAlive() const;
	void setAlive(const bool alive);
	PlayerColor getColor() const;
	void promote(const ChessPieceType newType);
	ChessPieceType getType() const;
	void setType(const ChessPieceType cpt);
	char getTypeChar(const bool whiteUppercase) const;
	int getEvalVal() const;
	bool getFirstMove() const;
	void clearFirstMove();
	void setFirstMove();
	void setFirstMove(const bool s) { state.first_move = s; }
	bool isWasPromotion() const { return state.was_promotion; }

#ifdef _DEBUG
	bool getMoveRelatedValid() const;
	int getNCanBeCaughtBy() const;
#endif
	bool isUnderAttack(const uint32_t * const restrict bitmap) const;
	bool getIsPinned(const uint32_t * const restrict bitmap) const;
	bool getIsKingPinned(const uint32_t * const restrict bitmap) const;
	bool getIsProtected(const uint32_t * const restrict bitmap) const;
#ifdef _DEBUG
	void prepareForGenerateMoveList();
#endif
	int generateMoveList(gml_pars_t *const restrict pars);
	int getMoveCount() const { return state.move_count; }

	bool isUnderAttackByThis(const isUnderAttackBy_pars_t *const restrict ibp) const;

	void doMove(const int toX, const int toY);
	void doKill();

	ChessPieceState getState() const;
	void pushState(); // should be private, but for testcases
	void popState();

	static const int evalVal[];

	void validate();
};

#endif
