// (C) Folkert van Heusden. AGPL v3.0.
// $Revision$
#include "Tweaks.h"
#include "config.h"
#ifdef W_SQLITE3
#include <limits.h>
#include "DBBookEvalSqlite3.h"
#include "Utils.h"
#include "Move.h"
#include "MoveType.h"
#include "error.h"

DBBookEvalSqlite3::DBBookEvalSqlite3(IO *const ioIn, const std::string & bookFile) : DBBook(ioIn)
{
	log(ll_ok, "DBBookEvalSqlite3 started");

	if (sqlite3_open(bookFile.c_str(), &handle))
		error_exit(false, "Cannot use book file %s: %s", bookFile.c_str(), sqlite3_errmsg(handle));

	th = new std::thread(std::ref(*this));
}

DBBookEvalSqlite3::~DBBookEvalSqlite3()
{
	sqlite3_close(handle);
}

int sl_callback(void *cb_data, int argc, char **argv, char **column_names)
{
	std::vector<std::string> row;

	for(int index=0; index<argc; index++)
	{
		if (argv[index])
			row.push_back(argv[index]);
		else
			row.push_back("");
	}

	std::vector<std::vector<std::string> > *result_set = (std::vector<std::vector<std::string> > *)cb_data;

	result_set -> push_back(row);

	return 0;
}

static int do_query(sqlite3 *const handle, const std::string & query, std::vector<std::vector<std::string> > *const results)
{
	char *error = NULL;
	if (sqlite3_exec(handle, query.c_str(), sl_callback, results, &error))
		error_exit("DB error: %s for query %s: %s", error, query.c_str(), sqlite3_errmsg(handle));

	return 1;
}

void DBBookEvalSqlite3::operator()()
{
	log(ll_ok, "DBBookEvalSqlite3::operator() starting");

	std::mutex mtx;
	std::unique_lock<std::mutex> lk(mtx);

	bool first = true;

	for(;!terminate;)
	{
		bool wasFirst = first;

		if (first)
			first = false;
		else
		{
			workAdded.wait(lk);

			if (terminate)
				break;
		}

		int cur_id = -1;
		PlayerColor wfColor;
		std::string workFen;

		w_lock(rm_lock);
		auto iq_it = in_queue.begin();

		if (iq_it != in_queue.end())
		{
			cur_id = iq_it -> first;
			workFen = iq_it -> second.fen;
			wfColor = iq_it -> second.c;
		}

		w_ulock(rm_lock);

		if (cur_id == -1)
		{
			if (!wasFirst)
				log(ll_err, "DBBookEvalSqlite3: nothing in queue?");
			continue;
		}

		log(ll_ok, "DBBookEvalSqlite3: searching for %s", workFen.c_str());

		std::string query = myformat("SELECT pv FROM entries WHERE fen='%s'", workFen.c_str());
		//printf("%s\n", query.c_str());

		std::vector<std::vector<std::string> > query_results;
		if (!do_query(handle, query, &query_results))
		{
			log(ll_err, "DBBookEvalSqlite3: database access failed!!");
			putResult(cur_id, invalidMove, INT_MIN, false);
			continue;
		}

		long unsigned int nResults = query_results.size();
		log(ll_ok, "DBBookEvalSqlite3: %d results from database", nResults);

		if (nResults == 0) {
			putResult(cur_id, invalidMove, INT_MIN, false);
			continue;
		}

		std::uniform_int_distribution<int> dist(0, nResults - 1);
		long unsigned int useEntry = dist(*mt);

		std::vector<std::string> *moves = split(query_results.at(useEntry).at(0), " ");

		if (!moves -> empty()) {
			const std::string move = moves -> at(0);

			Move m = invalidMove;

			Scene s(workFen);
			s.generateMovelists();
// FIXME remove invalid moves for SANToMove to work
// FIXME remove invalid moves for SANToMove to work

			if (SANToMove(move, s, wfColor, s.getMoveList(wfColor), &m)) {
				putResult(cur_id, m, INT_MIN, true);

				log(ll_ok, "DBBookEvalSqlite3: result stored");
			}
			else {
				putResult(cur_id, invalidMove, INT_MIN, false);
			}
		}
		else {
			putResult(cur_id, invalidMove, INT_MIN, false);

			log(ll_ok, "DBBookEvalSqlite3: no result found");
		}

		delete moves;
	}

	log(ll_ok, "DBBookEvalSqlite3::operator(): terminating");
}
#endif
