// (C) Folkert van Heusden. AGPL v3.0.
// $Revision: 4222 $
// WDL book
#include "Tweaks.h"
#include "config.h"
#include "DBBookLearn.h"
#include "Utils.h"
#include "Move.h"
#include "MoveType.h"
#include "Tpt.h"

DBBookLearn::DBBookLearn(IO *const ioIn, const std::string & host, const std::string & db, const std::string & user, const std::string & password, const std::string & db_tableIn) : DBBook(ioIn), db_table(db_tableIn)
{
#ifdef WITH_MYSQL_BOOKS
	log(ll_ok, "DBBookLearn constructor");
	handle = connect_to_database(host.c_str(), user.c_str(), password.c_str(), db.c_str());

	th = new std::thread(std::ref(*this));
#endif
}

DBBookLearn::DBBookLearn(IO *const ioIn) : DBBook(ioIn), handle(NULL)
{
#ifdef WITH_MYSQL_BOOKS
	log(ll_ok, "DBBookLearn constructor");
	th = new std::thread(std::ref(*this));
#endif
}

DBBookLearn::~DBBookLearn()
{
#ifdef WITH_MYSQL_BOOKS
	log(ll_ok, "DBBookLearn destructor");
	if (handle)
		mysql_close(handle);
#endif
}

void DBBookLearn::learn(const Board & b, const PlayerColor c, const Move & epMove, const Move & bestMove, const int depth, const winner_t win)
{
#ifdef WITH_MYSQL_BOOKS
	log(ll_ok, "DBBookLearn learn");
	zobrist_t zh = Tpt::calcZobrist(&b, c, epMove);

	std::string m = moveToUciMove(bestMove);

	int ww = win == WIN_WHITE, wb = win == WIN_BLACK, wd = win == WIN_DRAW, wu = win == WIN_UNKNOWN;

	std::string query = myformat("INSERT INTO bl(hash, move, depth, ww, wb, wd, wu) VALUES(%zu, '%s', %d, %d, %d, %d, %d) ON DUPLICATE KEY UPDATE ww=ww+%d, wb=wb+%d, wd=wd+%d, wu=wu+%d",
			zh.newHash, m.c_str(), depth, ww, wb, wd, wu, ww, wb, wd, wu);

	log(ll_ok, "%s", query.c_str());

	do_query(handle, query);
	do_query(handle, "COMMIT");
#endif
}

void DBBookLearn::operator()()
{
#ifdef WITH_MYSQL_BOOKS
	log(ll_ok, "DBBookLearn::operator() starting");

	std::mutex mtx;
	std::unique_lock<std::mutex> lk(mtx);

	bool first = true;

	for(;!terminate;)
	{
		bool wasFirst = first;

		if (first)
			first = false;
		else
		{
			workAdded.wait(lk);

			if (terminate)
				break;
		}

		int cur_id = -1;
		std::string workFen;

		w_lock(rm_lock);

		auto iq_it = in_queue.begin();

		if (iq_it != in_queue.end())
		{
			cur_id = iq_it -> first;
			workFen = iq_it -> second.fen;
		}

		w_ulock(rm_lock);

		if (cur_id == -1)
		{
			if (!wasFirst)
				log(ll_err, "DBBookLearn: nothing in queue?");
			continue;
		}

		std::vector<std::string> *fen_parts = split(workFen, " ");
		const PlayerColor c = fen_parts -> at(1) == "b" ? BLACK : WHITE;
		delete fen_parts;

		Scene s(workFen);
		const Move *const em = s.getLastOpponentMove();
		zobrist_t hash = Tpt::calcZobrist(s.getBoardConst(), c, em ? *em : invalidMove);

		log(ll_ok, "DBBookLearn: searching for %s (%ld)", workFen.c_str(), hash.newHash);

		// FIXME ook de andere velden meenemen in query
		std::string query = myformat("SELECT move FROM bl WHERE hash=%ld ORDER BY depth DESC LIMIT 1", hash.newHash);

		std::vector<std::vector<std::string> > query_results;
		if (!do_query(handle, query, &query_results))
		{
			log(ll_err, "DBBookLearn: database access failed!!");
			putResult(cur_id, invalidMove, INT_MIN, false, false);
			continue;
		}

		long unsigned int nResults = query_results.size();
		log(ll_ok, "DBBookLearn: %d results from database", nResults);

		if (nResults == 1)
		{
			std::string move = query_results.at(0).at(0);
			log(ll_ok, "DBBookLearn: Move returned: %s", move.c_str());

			Move m = moveStringToMove(move, c);

			putResult(cur_id, m, INT_MIN, false, true);

			log(ll_ok, "DBBookLearn: result stored");
		}
		else
		{
			putResult(cur_id, invalidMove, INT_MIN, false, false);

			log(ll_ok, "DBBookLearn: no result found");
		}
	}

	log(ll_ok, "DBBookLearn::operator() terminating");
#endif
}
