// (C) Folkert van Heusden. AGPL v3.0.
// $Revision: 5130 $
#include "Tweaks.h"
#include "config.h"
#include <sys/time.h>
#ifdef WSJS
#include <jansson.h>
#endif

#include "Exception.h"
#include "IO.h"
#include "Brain.h"
#include "error.h"
#include "ws.h"

std::string moveToUciMove(const Move & m)
{
	std::string bm = myformat("%c%c%c%c", m.u.m.fx + 'a', m.u.m.fy + '1', m.u.m.tx + 'a', m.u.m.ty + '1');

	if (m.u.m.promoteTo != NONE)
		bm = myformat("%c%c%c%c%c", m.u.m.fx + 'a', m.u.m.fy + '1', m.u.m.tx + 'a', m.u.m.ty + '1', tolower(typeToChar(m.u.m.promoteTo)));

	return bm;
}

ChessPieceType charToChessObjectType(const char cIn)
{
	int c = toupper(cIn);

	if (c == 'P')
		return PAWN;

	if (c == 'B')
		return BISHOP;

	if (c == 'R')
		return ROOK;

	if (c == 'N')
		return KNIGHT;

	if (c == 'Q')
		return QUEEN;

	if (c == 'K')
		return KING;

	my_assert(false, "charToChessObjectType: unknown promote-to");

	return (ChessPieceType)6;
}

char typeToChar(const ChessPieceType t)
{
	char types[6] = { 'P', 'N', 'B', 'R', 'Q', 'K' };

	return types[t];
}

std::string logfile = "trace.log";

void setLogfile(const std::string & file)
{
	logfile = file;
}

void log(const loglevel_t ll, const std::string what, ...)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);

	FILE *fh = fopen(logfile.c_str(), "a+");

	if (fh)
	{
		time_t tv_temp = tv.tv_sec;
		struct tm tm;
		localtime_r(&tv_temp, &tm);

		char *msg = NULL;
		va_list ap;
		va_start(ap, what);
		(void)vasprintf(&msg, what.c_str(), ap);
		va_end(ap);

		fprintf(fh, "%04d-%02d-%02d %02d:%02d:%02d.%06ld %3s %s\n", 
			tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday,
			tm.tm_hour, tm.tm_min, tm.tm_sec, tv.tv_usec,
			ll == ll_err ? "ERR" : "OK",
			msg);

		free(msg);

		fclose(fh);
	}
}

static bool findKV(const std::map<std::string, std::string> & parameters, const std::string & key, std::string *const value)
{
	std::map<std::string, std::string>::const_iterator it = parameters.find(key);

	if (it != parameters.end()) {
		*value = it -> second;
		return true;
	}

	return false;
}

bool cfgGetBool(const std::map<std::string, std::string> & parameters, const std::string & key, const bool def)
{
	std::string val;

	if (!findKV(parameters, key, &val))
		return def;

	return yesOrNo(val);
}

int cfgGetInt(const std::map<std::string, std::string> & parameters, const std::string & key, const int def)
{
	std::string val;

	if (!findKV(parameters, key, &val))
		return def;

	return atoi(val.c_str());
}

std::string cfgGetString(const std::map<std::string, std::string> & parameters, const std::string & key, const std::string & def)
{
	std::string val;

	if (!findKV(parameters, key, &val))
		return def;

	return val;
}

IO::IO(const std::map<std::string, std::string> & parametersIn, Tpt *const ttIn, Ptt *const pttIn) : parameters(parametersIn), b(NULL), stopFlag(false), verbose(false), kibitz(false), tt(ttIn), ptt(pttIn)
{
#ifdef ALWAYS_PONDER
	bool pDefault = true;
#else
	bool pDefault = false;
#endif

	ei.flag = false;

	alwaysPonder = cfgGetBool(parametersIn, "always_ponder", pDefault);

	resetStats();

#ifdef WSJS
	ws_thread_handle = NULL;

	screen_refresh_trigger = 0;

	ws = NULL;
#endif
}

IO::~IO()
{
	stopFlag = true;
}

void IO::setBrain(Brain *const bIn)
{
	b = bIn;
}

void IO::setBooks(std::vector<Book *> *const booksIn)
{
	books = booksIn;
}

void IO::emitNodeProgress(const Move & m, const Move & best) const
{
#ifdef WSJS
	if (ws) {
		json_t *j = json_object();

		add_json_key_value(j, "type", "node-progress");

		add_json_key_value(j, "fx", (ULI)m.u.m.fx);
		add_json_key_value(j, "fy", (ULI)m.u.m.fy);
		add_json_key_value(j, "tx", (ULI)m.u.m.tx);
		add_json_key_value(j, "ty", (ULI)m.u.m.ty);
		add_json_key_value(j, "currentmove", moveToString(m));

		if (isValidMove(best)) {
			add_json_key_value(j, "bfx", (ULI)best.u.m.fx);
			add_json_key_value(j, "bfy", (ULI)best.u.m.fy);
			add_json_key_value(j, "btx", (ULI)best.u.m.tx);
			add_json_key_value(j, "bty", (ULI)best.u.m.ty);
			add_json_key_value(j, "bestmove", moveToString(best));
		}

		char *temp = json_dumps(j, 0);
		ws -> update(temp, false);
		free(temp);

		json_decref(j);
	}
#endif
}

void IO::resetStats()
{
	if (b) {
		log(ll_ok, "average time_used: %.2f%%", thinking.thinkTimeUsedms * 100.0 / thinking.thinkTimeGivenms);
		b -> resetStats();
	}

	memset(&thinking.cs, 0x00, sizeof thinking.cs);
	memset(&thinking.s, 0x00, sizeof thinking.s);

	thinking.thinkTimeGivenms = thinking.thinkTimeUsedms = thinking.thinkTimeTotalCnt = 0;
}

std::string underAttack(const Scene *const s, const PlayerColor c)
{
	const ChessPiece *const pieces = s -> getPieces(c);
	std::string out;

	for(int i=0; i<16; i++) {
		if (pieces[i].getAlive() == false)
			continue;

		if (s -> getIsUnderAttack(&pieces[i])) {
			int ch = typeToChar(pieces[i].getType());
			if (pieces[i].getColor() == BLACK)
				ch = tolower(c);

			out += myformat("%c%c%c ", ch, 'a' + pieces[i].getX(), '1' + pieces[i].getY());
		}
	}

	return out;
}

void evalPieces(const Scene *const s, const PlayerColor c, int *const nFirstMove, int *const nPromoted)
{
	const ChessPiece *const pieces = s -> getPieces(c);
	std::string out;

	for(int i=0; i<16; i++) {
		if (pieces[i].getAlive() == false)
			continue;

		*nFirstMove += pieces[i] .getFirstMove();
		*nPromoted += pieces[i].isWasPromotion();
	}
}

#ifdef WSJS
void IO::setWebSockets(wsserver *const wsIn)
{
	ws = wsIn;

	if (ws) {
		json_t *j = json_object();
		add_json_key_value(j, "type", "meta");
		add_json_key_value(j, "description", cfgGetString(parameters, "description", "Embla v " VERSION));
		char *temp = json_dumps(j, 0);
		ws -> setDefault(temp);
		free(temp);
		json_decref(j);
	}
}
#endif

void IO::emitThreadStats(const int threadNr, const int alpha, const int beta, const int v, const VectorType<PVnode_t> & pv, const int depth)
{
#ifdef WSJS
	if (!ws)
		return;

	json_t *j = json_object();


// FIXME


	char *temp = json_dumps(j, 0);
	ws -> update(temp, false);
	free(temp);

	json_decref(j);
#endif
}

void IO::emitThinking(const int moveNr, const unsigned depth, const unsigned seldepth, const int score, const unsigned time, const unsigned long nodes, VectorType<PVnode_t> & pv, const bool revPv, const double ebf, const calc_stats_t *const cs, const Scene *const s, const PlayerColor c, const int alpha, const int beta)
{
#ifdef WSJS
	thinking.lock.lock();
	memcpy(&thinking.cs, cs, sizeof(thinking.cs));

	thinking.s.depth = depth;
	thinking.s.seldepth = seldepth;
	thinking.s.score = score;
	thinking.s.time = time;
	thinking.s.ebf = ebf;
	thinking.s.c = c;

	thinking.pv.assign(pv, true);
	thinking.lock.unlock();

	if (ws) {
		json_t *j = json_object();

		add_json_key_value(j, "type", "data");

		thinking_t *const t = this -> getThinking();
		t -> lock.lock();

		uint64_t total_nodes = t -> cs.cnt + t -> cs.qcnt;
		uint64_t ts_diff = get_ts_ms() - t -> cs.getStartedAt();
		double nps = double(t -> cs.qcnt + t -> cs.cnt) / double(ts_diff) / 1000.0;

		add_json_key_value(j, "wua", underAttack(s, WHITE));
		add_json_key_value(j, "bua", underAttack(s, BLACK));

		add_json_key_value(j, "M_NPS", nps / 1000);
		add_json_key_value(j, "reg_cnt", (ULI)t -> cs.cnt);
		add_json_key_value(j, "qs_cnt", (ULI)t -> cs.qcnt);
		add_json_key_value(j, "TT_queries", (ULI)(t -> cs.cnt + t -> cs.qcnt));
		add_json_key_value(j, "TT_hits", (t -> cs.tpt_hit_cnt * 100.0) / t -> cs.cnt);
		add_json_key_value(j, "QS_TT_hits", (t -> cs.qtpt_hit_cnt * 100.0) / t -> cs.qcnt);
		add_json_key_value(j, "TT_err", (ULI)t -> cs.tt_err);
		add_json_key_value(j, "TT_d0", (ULI)t -> cs.tt_d0);
		add_json_key_value(j, "TT_fill", tt -> hashFullPermill() / 10.0);
		add_json_key_value(j, "tt_hit_invalid", (ULI)t -> cs.tt_hit_type[0]);
		add_json_key_value(j, "tt_hit_exact", (ULI)t -> cs.tt_hit_type[1]);
		add_json_key_value(j, "tt_hit_lower", (ULI)t -> cs.tt_hit_type[2]);
		add_json_key_value(j, "tt_hit_upper", (ULI)t -> cs.tt_hit_type[3]);
		add_json_key_value(j, "tt_depth_ok", (ULI)t -> cs.tt_depth_ok);
		add_json_key_value(j, "PTT_hits", (t -> cs.ptt_hit * 100.0) / (t -> cs.ptt_hit + t -> cs.ptt_mis));
		add_json_key_value(j, "PTT_fill", ptt -> hashFullPermill() / 10.0);

		add_json_key_value(j, "depth", (ULI)t -> s.depth);
		add_json_key_value(j, "sel_depth", (ULI)t -> s.seldepth);
		add_json_key_value(j, "max_depth_reached", (ULI)t -> cs.max_depth_reached);
		add_json_key_value(j, "score", (long int)t -> s.score);
		add_json_key_value(j, "alpha", (long int)alpha);
		add_json_key_value(j, "beta", (long int)beta);
		add_json_key_value(j, "ebf", t -> s.ebf);
		add_json_key_value(j, "avg_bco_idx", t -> cs.cnt_cutoff_index / double(t -> cs.cnt_cutoff));
		add_json_key_value(j, "IID", (ULI)t -> cs.iid);
		add_json_key_value(j, "IID_hit", t -> cs.iid_hit * 100.0 / t -> cs.iid);
		add_json_key_value(j, "NM_hit", t -> cs.null_move_hit * 100.0 / t -> cs.null_move);
		add_json_key_value(j, "evals", t -> cs.evals);

		add_json_key_value(j, "PIFsize", double(t -> cs.pif_size) / total_nodes);
		add_json_key_value(j, "LMR_OK", (t -> cs.lmr_ok * 100.0) / (t -> cs.lmr_ok + t -> cs.lmr_fail));
		add_json_key_value(j, "R_extensions", (ULI)t -> cs.r_extensions);
		add_json_key_value(j, "standpt", t -> cs.qstand_pat_bco * 100.0 / t -> cs.qcnt);
		add_json_key_value(j, "draw", t -> cs.draw * 100.0 / total_nodes); 
		add_json_key_value(j, "mate", t -> cs.mate * 100.0 / total_nodes); 
		add_json_key_value(j, "qmoves", t -> cs.q_nm * 100.0 / t -> cs.qcnt); 
		add_json_key_value(j, "blind", t -> cs.blind * 100.0 / t -> cs.blind_total);

		if (t -> thinkTimeTotalCnt)
			add_json_key_value(j, "time_given", (ULI)(t -> thinkTimeGivenms / t -> thinkTimeTotalCnt));

		add_json_key_value(j, "time_used", t -> thinkTimeUsedms * 100.0 / t -> thinkTimeGivenms);
		add_json_key_value(j, "history", t -> cs.history_hits * 100.0 / (total_nodes + t -> cs.null_move));

		add_json_key_value(j, "w_prot", (ULI)s -> countProtected(WHITE));
		add_json_key_value(j, "b_prot", (ULI)s -> countProtected(BLACK));
		add_json_key_value(j, "w_uatt", (ULI)s -> countUnderAttack(WHITE));
		add_json_key_value(j, "b_uatt", (ULI)s -> countUnderAttack(BLACK));
		add_json_key_value(j, "n_white", (ULI)s -> getMoveList(WHITE) -> size());
		add_json_key_value(j, "n_black", (ULI)s -> getMoveList(BLACK) -> size());
		int wNFM = 0, wNP = 0;
		evalPieces(s, WHITE, &wNFM, &wNP);
		int bNFM = 0, bNP = 0;
		evalPieces(s, WHITE, &bNFM, &bNP);
		add_json_key_value(j, "w_nfm", (ULI)wNFM);
		add_json_key_value(j, "b_nfm", (ULI)bNFM);
		add_json_key_value(j, "w_np", (ULI)wNP);
		add_json_key_value(j, "b_np", (ULI)bNP);

		int whitePinnedN = 0, blackPinnedN = 0;
		s -> getKingPinnedCount(&whitePinnedN, &blackPinnedN);
		add_json_key_value(j, "w_k_pinned", (ULI)whitePinnedN);
		add_json_key_value(j, "b_k_pinned", (ULI)blackPinnedN);

		add_json_key_value(j, "syzygy_query_normal", (ULI)t -> cs.se_query[0]);
		add_json_key_value(j, "syzygy_query_quick", (ULI)t -> cs.se_query[1]);
		add_json_key_value(j, "syzygy_hit_normal_fail", (ULI)t -> cs.se_hit[0][0]);
		add_json_key_value(j, "syzygy_hit_normal_lost", (ULI)t -> cs.se_hit[0][1]);
		add_json_key_value(j, "syzygy_hit_normal_draw", (ULI)t -> cs.se_hit[0][2]);
		add_json_key_value(j, "syzygy_hit_normal_win", (ULI)t -> cs.se_hit[0][3]);
		add_json_key_value(j, "syzygy_hit_quick_fail", (ULI)t -> cs.se_hit[1][0]);
		add_json_key_value(j, "syzygy_hit_quick_lost", (ULI)t -> cs.se_hit[1][1]);
		add_json_key_value(j, "syzygy_hit_quick_draw", (ULI)t -> cs.se_hit[1][2]);
		add_json_key_value(j, "syzygy_hit_quick_win", (ULI)t -> cs.se_hit[1][3]);

		std::string fen = genFen(s, false, t -> s.c);
		add_json_key_value(j, "FEN", fen.c_str());
		add_json_key_value(j, "color", c == WHITE ? "w" : "b");
		add_json_key_value(j, "movenr", (ULI)moveNr);

		add_json_key_value(j, "cPV", pvToString(t -> pv, false, false, false).c_str());

		bool wl, Ws, bl, bs;
		s -> getCastlingRights(&wl, &Ws, &bl, &bs);
		std::string cr = std::string(wl ? "Q" : "") + (Ws ? "K" : "") + (bl ? "q" : "") + (bs ? "k" : "");
		add_json_key_value(j, "cr", cr);

		std::string material;
		constexpr char names[] = { 'p', 'n', 'b', 'r', 'q', 'k' };
		const int *wc = s -> getMaterialCounts(WHITE);
		for(int i=0; i<6; i++)
			material += myformat("%c%d", names[i], wc[i]) + " ";
		material += "/ ";
		const int *bc = s -> getMaterialCounts(BLACK);
		for(int i=0; i<6; i++)
			material += myformat("%c%d", names[i], bc[i]) + " ";
		add_json_key_value(j, "material", material);

		add_json_key_value(j, "ismiddlegame", s -> isMiddleGame() ? "true" : "false");
		add_json_key_value(j, "isendgame", s -> isEndGame() ? "true" : "false");

		t -> lock.unlock();

		if (!pv.empty()) {
			const Move best = pv.at(0).m;

			add_json_key_value(j, "bfx", (ULI)best.u.m.fx);
			add_json_key_value(j, "bfy", (ULI)best.u.m.fy);
			add_json_key_value(j, "btx", (ULI)best.u.m.tx);
			add_json_key_value(j, "bty", (ULI)best.u.m.ty);
		}

		char *temp = json_dumps(j, 0);
		ws -> update(temp, true);
		free(temp);

		json_decref(j);
	}

	screen_refresh_trigger++;
#endif
}

void IO::emitPvDetails(const Scene *const sIn, const PlayerColor cIn, Move m)
{
	if (!tt)
		return;

	std::string out = recreatePvFromHash(tt, sIn, cIn, m, false, true);

	if (!out.empty())
		emitDebug(out);
}

void IO::dumpBoard(const zobrist_t & nh, Scene *const s, const DmpBrd_t type, const PlayerColor c)
{
	const Board *const lb = s -> getBoardConst();

	if (type == DB_P) {
		std::vector<std::string> out;

		for(int y=7; y>=0; y--)
		{
			std::string line;

			line += myformat("%d ", y + 1);

			for(int x=0; x<8; x++)
			{
				ChessPiece *p = lb -> getAt(x, y);

				if (!p)
					line += ".";
				else
				{
					bool upper = p -> getColor() == WHITE;
					char pt = typeToChar(p -> getType());

					line += myformat("%c", upper ? toupper(pt) : tolower(pt));
				}
			}

			emitCmd(line);
		}

		std::string line = "  ";
		for(int x=0; x<8; x++)
			line += myformat("%c", 'A' + x);

		emitCmd(line);
	}
	else if (type == DB_EVAL) {
		int evals[8][8];
		for(int y=0; y<8; y++) {
			for(int x=0; x<8; x++)
				evals[y][x] = INT_MIN;
		}

		bool hit[8][8];
		memset(hit, 0x00, sizeof hit);

		s -> generateMovelists();
		const VectorType<Move> *const moves = s -> getMoveList(c);

		calc_stats_t cs;
		memset(&cs, 0x00, sizeof cs);

		for(size_t i=0; i<moves -> size(); i++) {
			if (!s -> validateMoves(c, i))
				continue;

			const Move m = moves -> at(i);

			BoardState bs;
			s -> doMove(m, &bs);

			evals[m.u.m.ty][m.u.m.tx] = std::max(evals[m.u.m.ty][m.u.m.tx], b -> evaluate(nh, s, &cs, c, 0, c));
			hit[m.u.m.ty][m.u.m.tx] = true;

			s -> undoMove();
		}

		std::vector<std::string> out;

		for(int y=7; y>=0; y--)
		{
			std::string line;

			line += myformat("%d ", y + 1);

			for(int x=0; x<8; x++)
			{
				if (hit[y][x])
					line += myformat("%6d", evals[y][x]);
				else
					line += "   *  ";
			}

			emitCmd(line);
		}

		std::string line = "  ";
		for(int x=0; x<8; x++)
			line += myformat("%c     ", 'A' + x);

		emitCmd(line);
	}
	else if (type == DB_MP) {
		bool types[8][8][6];
		memset(types, 0x00, sizeof types);

		s -> generateMovelists();
		const VectorType<Move> *const moves = s -> getMoveList(c);

		for(size_t i=0; i<moves -> size(); i++) {
			if (!s -> validateMoves(c, i))
				continue;

			const Move m = moves -> at(i);

			types[m.u.m.ty][m.u.m.tx][lb -> getAt(m.u.m.fx, m.u.m.fy) -> getType()] = true;
		}

		std::vector<std::string> out;

		for(int y=7; y>=0; y--)
		{
			std::string line;

			line += myformat("%d ", y + 1);

			for(int x=0; x<8; x++)
			{
				std::string pieces;
				if (types[y][x][0])
					pieces += "P";
				if (types[y][x][1])
					pieces += "N";
				if (types[y][x][2])
					pieces += "B";
				if (types[y][x][3])
					pieces += "R";
				if (types[y][x][4])
					pieces += "Q";
				if (types[y][x][5])
					pieces += "K";

				line += myformat("%-6s", pieces.c_str());
			}

			emitCmd(line);
		}

		std::string line = "  ";
		for(int x=0; x<8; x++)
			line += myformat("%c     ", 'A' + x);

		emitCmd(line);
	}
	else if (type == DB_PROT) {
		std::vector<std::string> out;

		for(int y=7; y>=0; y--)
		{
			std::string line;

			line += myformat("%d ", y + 1);

			for(int x=0; x<8; x++)
			{
				ChessPiece *p = lb -> getAt(x, y);

				if (!p) {
					line += ".";
					continue;
				}

				if (!s -> getIsProtected(p))
					line += "!";
				else
				{
					bool upper = p -> getColor() == WHITE;
					char pt = typeToChar(p -> getType());

					line += myformat("%c", upper ? toupper(pt) : pt);
				}
			}

			emitCmd(line);
		}

		std::string line = "  ";
		for(int x=0; x<8; x++)
			line += myformat("%c", 'A' + x);

		emitCmd(line);
	}
}

const std::map<std::string, std::string> & IO::getParameters()
{
	static const std::map<std::string, std::string> pars = { { "always_ponder", "Always think in opponent time (true/false)" }, { "initial_fen", "Set the initial position to work with" }, { "name_extra", "Extra text to add to the UCI identification string" }, { "max_depth", "Maximum search depth" }, { "max_time", "Maximum duration to think (in ms)" } };

	return pars;
}

void IO::loop()
{
}

void IO::emitJSWSLog(const std::string & msg) const 
{
#ifdef WSJS
	if (ws) {
		json_t *j = json_object();
		add_json_key_value(j, "type", "log");
		add_json_key_value(j, "msg", msg);
		char *temp = json_dumps(j, 0);
		ws -> update(temp, false);
		free(temp);
		json_decref(j);
	}
#endif
}

void IO::emitCmd(const std::string what, ...) const
{
	char *msg = NULL;

	va_list ap;
	va_start(ap, what);
	(void)vasprintf(&msg, what.c_str(), ap);
	va_end(ap);

	log(ll_ok, "CMD: " + std::string(msg));

	free(msg);
}

void IO::emitDebug(const std::string what, ...) const
{
	char *msg = NULL;

	va_list ap;
	va_start(ap, what);
	(void)vasprintf(&msg, what.c_str(), ap);
	va_end(ap);

	log(ll_ok, "DEBUG: " + std::string(msg));
	emitJSWSLog(myformat("DEBUG: %s", msg));

	free(msg);
}

void IO::tellAll(const std::string what, ...) const
{
	char *msg = NULL;

	va_list ap;
	va_start(ap, what);
	(void)vasprintf(&msg, what.c_str(), ap);
	va_end(ap);

	log(ll_ok, "TELL: " + std::string(msg));
	emitJSWSLog(myformat("TELL: %s", msg));

	free(msg);
}

void IO::terminate()
{
	close(0); // FIXME uggly hack because poll() does not work with polyglot

	ei.flag = true;
	ei.cv.notify_all();

	stopFlag = true;
}

void IO::startPonder(const int moveNr, Scene & s, const PlayerColor c, const Move epMove, const std::vector<uint64_t> *const historyIn, const Move oppMove)
{
	s.generateMovelists();
	s.partialValidateMoves(c);

	if (!s.isCheckMate(c) && !s.isStaleMate(c))
	{
		emitDebug("Pondering...");

		b -> startPonder(moveNr, s, c, epMove, historyIn, oppMove);
	}
}

void IO::new_Scene(Scene **s, PlayerColor *const c)
{
	std::string initialFen = cfgGetString(parameters, "initial_fen", "");

	if (initialFen.empty()) {
		*s = new Scene();
		*c = WHITE;
	}
	else {
		log(ll_ok, "Using FEN %s", initialFen.c_str());
		*s = new Scene(initialFen);

		std::vector<std::string> *const parts = split(initialFen, " ");
		std::string color = parts -> at(1);
		delete parts;

		*c = color == "w" ? WHITE : BLACK;
	}
}

std::string emitMoves(Scene *const s, const PlayerColor c, const bool all)
{
	const Board *const b = s -> getBoardConst();
	VectorType<Move> *const moves = s -> getMoveList(c);

	// FIXME board/syzygy annotations
	for(size_t i=0; i<moves -> size(); i++) {
		const char *n = !s -> validateMoves(c, i) ? "not" : "";
		printf("%2zu] %s%s\n", i + 1, moveToString(moves -> at(i), true, b, s -> isKingUnderAttack(c)).c_str(), n);
	}

	std::string out;

	for(size_t i=0; i<moves -> size(); i++) {
		if (!all && !s -> validateMoves(c, i))
			continue;

		if (!s -> validateMoves(c, i))
			out += " (" + moveToString(moves -> at(i), true, b, s -> isKingUnderAttack(c)) + ")";
		else
			out += " " + moveToString(moves -> at(i), true, b, s -> isKingUnderAttack(c));
	}

	return out;
}

bool IO::genericCommands(const zobrist_t & nh, const std::vector<std::string> *const parts, const std::vector<pos_t> *const h, Scene *const s, const PlayerColor c)
{
	if (parts -> at(0) == "help") {
		emitCmd("# board / mp / eval / history / fen / moves / moves_eval / book / color / prot");
		emitCmd("# divide / perft");
		emitCmd("# search");
	}
	else if (parts -> at(0) == "color")
		emitCmd("# Current color is %s", c == WHITE ? "white" : "black");
	else if (parts -> at(0) == "board" && parts -> size() == 1) {
		dumpBoard(nh, s, DB_P, c);
		emitCmd("# color: %s", c == WHITE ? "white" : "black");
	}
	else if (parts -> at(0) == "prot" && parts -> size() == 1) {
		s -> generateMovelists();
		dumpBoard(nh, s, DB_PROT, c);
	}
	else if (parts -> at(0) == "book") {
		std::map<int, Book *> bookQuestions;

		s -> generateMovelists();

		b ->  askBookQuestions(&bookQuestions, s, c);

		bool hit = false;
		for(int i=0; i<10; i++) {
			usleep(99000);

			VectorType<PVnode_t> *ml = NULL;
			bool isFinalAnswer = false;
			if (b -> checkBookQuestions(&bookQuestions, s, c, &ml, &isFinalAnswer)) {
				emitCmd(myformat("# Book: %s (%s)", pvToString(*ml, false, false, false).c_str(), isFinalAnswer ? "final" : "suggestion"));
				hit = true;
				delete ml;
				break;
			}
		}

		b -> stopBookQuestions(&bookQuestions);

		if (!hit)
			emitCmd("# no book hit");

		// FIXME syzygy
	}
	else if (parts -> at(0) == "moves") {
		bool all = parts -> size() == 2 && parts -> at(1) == "all";

		s -> generateMovelists();
		emitMoves(s, WHITE, all);
		printf("\n");
		emitMoves(s, BLACK, all);
#if 0

		std::string outw = "# White moves:" + emitMoves(s, WHITE, all);
		emitCmd(outw);
		emitCmd(myformat("# number of white moves: %d", s -> getMoveList(WHITE) -> size()));

		std::string outb = "# Black moves:" + emitMoves(s, BLACK, all);
		emitCmd(outb);
		emitCmd(myformat("# number of black moves: %d", s -> getMoveList(BLACK) -> size()));
#endif
	}
	else if (parts -> at(0) == "moves_eval") {
		const PlayerColor c = (parts -> size() == 1 || parts -> at(1) == "w") ? WHITE : BLACK;

		s -> generateMovelists();
		VectorType<Move> *moves = s -> getMoveList(c);
		const Board *const brd = s -> getBoardConst();

		if (s -> isKingUnderAttack(c))
			emitCmd("# CHECK\n");

		BoardState bs;

		calc_stats_t cs;
		memset(&cs, 0x00, sizeof cs);

		for(size_t i=0; i<moves->size(); i++) {
			if (!s -> validateMoves(c, i))
				continue;

			const Move m = moves -> at(i);
			const Move epMove = isEpMove(brd, m) ? m : invalidMove;

			s -> doMove(m, &bs);

			zobrist_t nh = tt -> calcZobrist(brd, c, epMove);

			emitCmd("Move %s: %d", moveToString(m).c_str(), b -> evaluate(nh, s, &cs, c, 0, c));

			s -> undoMove();
		}
	}
	else if (parts -> at(0) == "mp") {
		dumpBoard(nh, s, DB_MP, c);
		emitCmd("# color: %s", c == WHITE ? "white" : "black");
	}
	else if (parts -> at(0) == "board_eval" || parts -> at(0) == "boardeval" || (parts -> at(0) == "board" && parts -> at(1) == "eval") || parts -> at(0) == "eval") {
		s -> generateMovelists();

		// dumpBoard(nh, s, DB_EVAL, c);
		emitCmd("# color: %s", c == WHITE ? "white" : "black");
#if 0 // FIXME
		emitCmd("# current position: %d", evaluate(s, c, 0));
#endif
		//		emitCmd("# king under attack: %s", s -> isKingUnderAttack(c) ? "true" : "false");

		calc_stats_t cs;
		memset(&cs, 0x00, sizeof cs);
		emitCmd("Total Evaluation: %f (white side)", b -> evaluate(nh, s, &cs, WHITE, 0, WHITE) / 100.0);
		//emitCmd("Total Evaluation: %f (white side)", b -> calcShannon(nh, s, NULL, WHITE) / 100.0);
	}
	else if (parts -> at(0) == "history" && h) {
		for(pos_t p : *h)
			emitCmd("# %s %s %d", p.fen.c_str(), p.move.c_str(), p.eval);
	}
	else if (parts -> at(0) == "fen") {
		std::string fen = genFen(s, false, c);

		emitCmd("# current fen: " + fen);
	}
	else if (parts -> at(0) == "divide" || parts -> at(0) == "perft") {
		if (parts -> size() != 2)
			emitCmd("# Usage: %s depth", parts -> at(0).c_str());
		else {
			uint64_t glob_start_ts = get_ts_ms();

			bool isPerft = parts -> at(0) == "perft";

			int depth = atoi(parts -> at(1).c_str());
			emitCmd("# %s depth: %d", parts -> at(0).c_str(), depth);

			s -> generateMovelists();
			VectorType<Move> *const moves = s -> getMoveList(c);

			PlayerColor co = opponentColor(c);

			perft_counts_t total;
			memset(&total, 0x00, sizeof total);

			unsigned long int cnt = b -> perft(*s, depth, c, &total, !isPerft);

			uint64_t glob_end_ts = get_ts_ms();

			emitCmd("# %lu %s", cnt, perftCountsToStr(&total).c_str());

			double seconds = (glob_end_ts - glob_start_ts) / 1000.0;
			emitCmd("# took %.2fs (%unps)", seconds, unsigned(cnt / seconds));
		}
	}
	else if (parts -> at(0) == "ds" || parts -> at(0) == "divide-search" || parts -> at(0) == "qds") {
		ei.flag = false;

		if (parts -> size() != 2)
			emitCmd("# Usage: ds depth");
		else {
			int depth = atoi(parts -> at(1).c_str());
			emitCmd("# divide-search depth: %d", depth);

			uint64_t glob_start_ts = get_ts_ms();
			uint64_t total_node_cnt = 0;

			s -> generateMovelists();
			VectorType<Move> *const moves = s -> getMoveList(c);

			PlayerColor co = opponentColor(c);

			VectorType<PVnode_t> oldPv;

			Move im[] = { invalidMove, invalidMove };

			BoardState bs;
			for(size_t idx=0; idx<moves -> size(); idx++) {
				if (!s -> validateMoves(c, idx))
					continue;

				std::string moveStr = moveToString(moves -> at(idx), true, s -> getBoardConst(), s -> isKingUnderAttack(co));

				s -> doMove(moves -> at(idx), &bs);
				zobrist_t hash = b -> calcZobrist(s -> getBoardConst(), co, invalidMove);

				s -> generateMovelists();

				search_data_t sd;
				sd.time_abort_flag = &ei.flag;
				sd.s = s;
				memset(&sd.cs, 0x00, sizeof sd.cs);
				sd.history = new std::vector<uint64_t>();
				sd.oldPv = &oldPv;

				tt -> reset();

				Move pvm;
				VectorType<PVnode_t> pv;
				int v = 0;
				if (parts -> at(0) == "qds")
					v = b -> quiesceSearch(&sd, 0, co, -32768, 32768, hash, im, &pvm);
				else
					v = b -> search(&sd, false, depth, co, -32768, 32768, hash, im, false, &pv);

				delete sd.history;

				emitCmd("# %s: %d[%lu@%d] %s", moveToString(moves -> at(idx)).c_str(), v, (unsigned long int)(sd.cs.qcnt + sd.cs.cnt), sd.cs.max_depth, genFen(s, false, c).c_str());
				//if (abs(v) == 67)
				//	break;
				total_node_cnt += sd.cs.qcnt + sd.cs.cnt;

				s -> undoMove();
			}

			uint64_t glob_end_ts = get_ts_ms();
			emitCmd("# took %.3f, %lu nodes in total", (glob_end_ts - glob_start_ts) / 1000.0, (unsigned long int)total_node_cnt);
		}
	}
	else if (parts -> at(0) == "zobrist") {
		zobrist_t zh = tt -> calcZobrist(s -> getBoardConst(), c, s -> getLastOpponentMove() ? *s -> getLastOpponentMove() : invalidMove);
		// start moet zijn 5060803636482931868
		emitCmd("# zobrist hash value %llu / 0x%llx", zh.newHash, zh.newHash);
	}
	else if (parts -> at(0) == "search") {
		Move sm = uciMoveStrToMove(s, parts -> at(1));
		VectorType<Move> tm;
		tm.push_back(sm);

		ei.flag = false;

		uint64_t nv = 0;
		int score = 0;
		VectorType<PVnode_t> *result = NULL;
		b -> calculateMove(0, *s, c, -1, -1, -1, &tm, NULL, invalidMove, &result, &score, &ei, &nv, invalidMove, invalidMove);
		delete result;

		emitDebug("Search finished");
	}
	else {
		return false;
	}

	return true;
}

void * calculateThread(void *p)
{
	uint64_t cnts = 0;

	uint64_t start = get_ts_ms();

	ct_t *const ct = (ct_t *const)p;
	ct -> b -> calculateMove(ct -> moveNr, *ct -> s, ct -> c, ct -> maxDepth, ct -> durationMin, ct -> durationMax, ct -> targetMoves, ct -> history, ct -> epMove, ct -> result, ct -> value, ct -> ei, &cnts, ct -> ponderHit, ct -> lastOppMove);

	uint64_t took = get_ts_ms() - start;

	ct -> th -> lock.lock();
	if (ct -> durationMax > 0) {
		ct -> th -> thinkTimeGivenms += ct -> durationMax;
		ct -> th -> thinkTimeUsedms += took;
		ct -> th -> thinkTimeTotalCnt++;
	}
	ct -> th -> lock.unlock();

	char e = 'E';
	if (write(ct -> p[1], &e, 1) != 1)
		log(ll_err, "info string pipe write error");

	return NULL;
}

Move uciMoveStrToMove(const Scene *const s, const std::string & moveStr)
{
	Move mc;
	mc.u.m.fx = uint8_t(moveStr.at(0) - 'a');
	mc.u.m.fy = uint8_t(moveStr.at(1) - '1');
	mc.u.m.tx = uint8_t(moveStr.at(2) - 'a');
	mc.u.m.ty = uint8_t(moveStr.at(3) - '1');
	mc.u.m.mt = REGULAR;
	mc.u.m.promoteTo = NONE;

	const ChessPiece *fromP = s -> getBoardConst() -> getAt(mc.u.m.fx, mc.u.m.fy);
	const ChessPiece *toP = s -> getBoardConst() -> getAt(mc.u.m.tx, mc.u.m.ty);

	if (fromP)
	{
		mc.u.m.isCatchMove = toP != NULL;

		if (fromP -> getType() == KING && fromP -> getFirstMove() && (mc.u.m.tx == 6 || mc.u.m.tx == 2))
			mc.u.m.mt = CASTLING;
		else if (fromP -> getType() == PAWN)
		{
			if (mc.u.m.fx != mc.u.m.tx && toP == NULL)
				mc.u.m.mt = EN_PASSANT;
			else if (mc.u.m.ty == 0 || mc.u.m.ty == 7)
			{
				if (moveStr.size() == 5)
					mc.u.m.promoteTo = charToChessObjectType(moveStr.at(4));
				else
					mc.u.m.promoteTo = QUEEN;
			}
		}
	}
	else
	{
		mc = invalidMove;
	}

	return mc;
}
