// (C) Folkert van Heusden. AGPL v3.0.
// $Revision: 4815 $
#ifndef __IO__
#define __IO__

#include <atomic>
#include <condition_variable>
#include <map>
#include <mutex>
#include <stdarg.h>
#include <string>
#include <thread>
#include <vector>

#include "Gen.h"
#include "Board.h"
#include "Scene.h"
#include "PV.h"
#include "config.h"
#include "Ptt.h"

class Book;
class Brain;
class wsserver;

std::string moveToUciMove(const Move & m);
ChessPieceType charToChessObjectType(const char cIn);
char typeToChar(const ChessPieceType t);

typedef enum { ll_ok, ll_err, ll_debug } loglevel_t;

typedef enum { DB_P, DB_EVAL, DB_MP, DB_PROT } DmpBrd_t;

void log(const loglevel_t ll, const std::string what, ...);
void setLogfile(const std::string & file);

bool cfgGetBool(const std::map<std::string, std::string> & parameters, const std::string & key, const bool def);
int cfgGetInt(const std::map<std::string, std::string> & parameters, const std::string & key, const int def);
std::string cfgGetString(const std::map<std::string, std::string> & parameters, const std::string & key, const std::string & def);

typedef struct
{
	std::atomic_bool flag;
	std::condition_variable cv;
}
end_indicator_t;

void * calculateThread(void *p);

typedef struct
{
	std::string fen, move;
	int eval;
} pos_t;

typedef struct {
	calc_stats_t cs;
	struct {
		unsigned depth, seldepth, time;
		int score;
		double ebf;
		PlayerColor c;
	} s;
	VectorType<PVnode_t> pv; // lock

	uint64_t thinkTimeGivenms, thinkTimeUsedms, thinkTimeTotalCnt;
	std::mutex lock;
} thinking_t;

typedef struct
{
	Brain *const b;
	Scene *const s;
	const PlayerColor c;
	const int maxDepth;
	const int durationMin, durationMax;
	const VectorType<Move> *const targetMoves;
	std::vector<uint64_t> *const history;
	const Move epMove;
	VectorType<PVnode_t> **result;
	int *const value;
	end_indicator_t *const ei;
	int p[2];
	Move ponderHit, lastOppMove;
	thinking_t *th;
	int moveNr;
} ct_t;

class Tpt;
class IO
{
protected:
#ifdef WSJS
	std::thread *ws_thread_handle;
	std::mutex tt_nc_lock; // tcp thread ncurses lock
	std::atomic_uint screen_refresh_trigger;
	wsserver *ws;
#endif

	thinking_t thinking;

	const std::map<std::string, std::string> & parameters;
	Brain *b;
	std::vector<Book *> *books;
	std::atomic_bool stopFlag;
	bool alwaysPonder;
	end_indicator_t ei;
	bool verbose, kibitz;
	Tpt *tt;
	Ptt *const ptt;

	void new_Scene(Scene **s, PlayerColor *const c);

	void resetStats();

public:
	IO(const std::map<std::string, std::string> & parametersIn, Tpt *const tt, Ptt *const ptt);
	virtual ~IO();

	static const std::map<std::string, std::string> & getParameters();

	void setVerbose(const bool v) { verbose = v; };
	bool getVerbose() { return verbose; }

	void setKibitz(const bool k) { kibitz = k; };

#ifdef WSJS
	void setWebSockets(wsserver *const wsIn);
	thinking_t *getThinking() { return &thinking; }
#endif

	void setBrain(Brain *const b);
	void setBooks(std::vector<Book *> *const books);
	void terminate();

	virtual void emitCmd(const std::string what, ...) const;
	virtual void emitDebug(const std::string what, ...) const;
	virtual void tellAll(const std::string what, ...) const;
	virtual void emitJSWSLog(const std::string & msg) const;

	virtual void emitNodeProgress(const Move & m, const Move & best) const;
	virtual void emitThinking(const int moveNr, const unsigned depth, const unsigned seldepth, const int score, const unsigned time, const unsigned long nodes, VectorType<PVnode_t> & pv, const bool revPv, const double ebf, const calc_stats_t *const cs, const Scene *const s, const PlayerColor c, const int alpha, const int beta);
	void emitThreadStats(const int threadNr, const int alpha, const int beta, const int v, const VectorType<PVnode_t> & pv, const int depth);

	void emitPvDetails(const Scene *const sIn, const PlayerColor cIn, Move m);
	void dumpBoard(const zobrist_t & nh, Scene *const s, const DmpBrd_t type, const PlayerColor c);
	bool genericCommands(const zobrist_t & nh, const std::vector<std::string> *const parts, const std::vector<pos_t> *const h, Scene *const s, const PlayerColor c);

	virtual void loop();
	void startPonder(const int moveNr, Scene & s, const PlayerColor c, const Move epMove, const std::vector<uint64_t> *const historyIn, const Move oppMove);
};

Move uciMoveStrToMove(const Scene *const s, const std::string & moveStr);

#endif
