// (C) Folkert van Heusden. AGPL v3.0.
// $Revision: 5129 $
char REV_IO_UCI[] = "IO_UCI $Revision: 5129 $";

#include "Tweaks.h"
#include <ctype.h>
#include <errno.h>
#include <poll.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/utsname.h>

#include "config.h"
#include "Exception.h"
#include "Utils.h"
#include "Move.h"
#include "IO_UCI.h"
#include "Brain.h"
#include "History.h"
#include "Utils.h"
#include "Main.h"
#include "Tune.h"
#ifdef W_SYZYGY
#include "syzygy_etb/fathom.h"
#endif
#ifdef WSJS
#include "ws.h"
#endif

const int defaultNMovesPerGame = 40;

static const char shell[] = "/bin/sh";

const char *getHardware()
{
	FILE *fh = fopen("/sys/firmware/devicetree/base/model", "r");
	if (fh) {
		char buffer[4096] = { 0 };
		fgets(buffer, sizeof buffer, fh);

		fclose(fh);

		if (buffer[0])
			return strdup(buffer);

		return NULL;
	}

	struct utsname name;
	if (uname(&name) == -1) {
		log(ll_err, "uname failed: %s", strerror(errno));
		return NULL;
	}

	return strdup(myformat("%s %s %s %s", name.sysname, name.release, name.version, name.machine).c_str());
}

IO_UCI::IO_UCI(const std::map<std::string, std::string> & parametersIn, Tpt *const ttIn, Ptt *const pttIn) : IO(parametersIn, ttIn, pttIn)
{
	name_extra = cfgGetString(parametersIn, "name_extra", "");

	if (!name_extra.empty())
		name_extra = "-" + name_extra;
}

IO_UCI::~IO_UCI()
{
}

void IO_UCI::forkProcess(const std::string & what)
{
	// using vfork: with large hash-tables a fork() (copying all data-pages) can be too costly
	pid_t pid = vfork();

	if (pid == -1)
		log(ll_err, "Failed to fork: %s", strerror(errno));
	else if (pid == 0)
	{
		if (execlp(shell, shell, "-c", what.c_str(), NULL) == -1)
			log(ll_err, "Failed to invoke \"%s\" from \"%s\"", what.c_str(), shell);

		exit(1);
	}
	else
	{
		log(ll_ok, "Invoking %s", what.c_str());
	}
}

void IO_UCI::emitCmd(const std::string what, ...) const
{
	char *msg = NULL;

       	va_list ap;
       	va_start(ap, what);
	(void)vasprintf(&msg, what.c_str(), ap);
	va_end(ap);

	printf("%s\n", msg);
	fflush(stdout);

	log(ll_ok, "CMD: " + std::string(msg));

	free(msg);
}

void IO_UCI::emitDebug(const std::string what, ...) const
{
	char *msg = NULL;

       	va_list ap;
       	va_start(ap, what);
	(void)vasprintf(&msg, what.c_str(), ap);
	va_end(ap);

	if (verbose)
		emitCmd("info string %s", msg);

	emitJSWSLog(myformat("DEBUG: %s", msg));

	free(msg);
}

void IO_UCI::tellAll(const std::string what, ...) const
{
	char *msg = NULL;

       	va_list ap;
       	va_start(ap, what);
	(void)vasprintf(&msg, what.c_str(), ap);
	va_end(ap);

	// http://www.talkchess.com/forum/viewtopic.php?t=54167
	// https://www.gnu.org/software/xboard/engine-intf.html#9
	if (kibitz)
		emitCmd("info string xboard tellall %s", msg);
	else
		emitDebug(msg);

	emitJSWSLog(myformat("TELL: %s", msg));

	free(msg);
}

void IO_UCI::emitThinking(const int moveNr, const unsigned depth, const unsigned seldepth, const int score, const unsigned time, const unsigned long nodes, VectorType<PVnode_t> & pv, const bool revPv, const double ebf, const calc_stats_t *const cs, const Scene *const s, const PlayerColor c, const int alpha, const int beta)
{
	IO::emitThinking(moveNr, depth, seldepth, score, time, nodes, pv, revPv, ebf, cs, s, c, alpha, beta);

	std::string pvStr = recreatePvFromHash(tt, s, c, revPv ? pv.back().m : pv.at(0).m, false, true);

	std::string base = myformat("info depth %lu seldepth %d score cp %d", depth, seldepth, score), tail;

	if (time >= 1)
		tail = myformat(" time %ld nodes %ld nps %d pv %s", time, nodes, int(nodes / (time / 1000.0)), pvStr.c_str());
	else
		tail = myformat(" time %ld nodes %ld pv %s", time, nodes, pvStr.c_str());

	std::string ebfstr;
	if (ebf >= 0.0)
		ebfstr = myformat(" ebf %.2f", ebf);

	std::string hf;
	uint64_t nowTs = get_ts_ms();
	if (nowTs - prevHashFull >= 1000) {
		hf = myformat(" hashfull %u", tt -> hashFullPermill());
		prevHashFull = nowTs;
	}

	emitCmd(base + ebfstr + hf + tail);
}

void IO_UCI::loop()
{
	bool firstMove = true;
	uint64_t initialKibitzDelay = get_ts_ms();

	int wTime = 0, bTime = 0;
	int wInc = 0, bInc = 0;
	int maxNMoves = defaultNMovesPerGame, moveNr = 1;
	int maxIt = 0;

	Move epMove = invalidMove, lastOppMove = invalidMove;

	PlayerColor c = WHITE;
	Scene *s = NULL;
	new_Scene(&s, &c);

	std::vector<uint64_t> history;

	//std::vector<std::string> ungetBuffer;
	for(;!stopFlag;)
	{
		std::string temp;

		std::string *line = getLineFromSTDIN();
		if (!line) {
			if (errno == EINTR)
				continue;

			break;
		}

		temp = *line;
		delete line;

		if (temp.empty()) {
			log(ll_err, "nothing?");
			continue;
		}

		log(ll_ok, "IN: " + temp);

		emitJSWSLog(myformat("IN: %s", temp.c_str()));

		std::vector<std::string> *parts = split(temp, " ");

		if (parts -> at(0) == "UCI" || parts -> at(0) == "uci")
		{
			emitCmd(myformat("id name Embla" NAME_EXTRA "%s", name_extra.c_str()));
			emitCmd("id author Folkert van Heusden <mail@vanheusden.com>");
			emitCmd("option name UCI_Opponent type string default <empty>");
			emitCmd("option name Ponder type check default false");
			emitCmd("option name Threads type spin default %d min 1 max 4096", b -> getNTHreads());
#ifdef WITH_CONTEMPT
			emitCmd("option name Contempt type spin default %d min -100 max 100", tune -> contemptMiddlegame);
#endif
			emitCmd("option name Minimum Thinking Time type spin default %d min 0 max 5000", tune -> minThinkTime);
			emitCmd("option name Hash type spin default %d min 17 max 1048576", DEFAULT_HASHTABLE_SIZE);
#ifdef W_SYZYGY
			emitCmd("option name SyzygyPath type string default");
#endif
			emitCmd("uciok");
		}
		else if (parts -> at(0) == "quit" || parts -> at(0) == "exit" || parts -> at(0) == "end")
		{
			emitDebug("TERMINATE");
			delete parts;
			b -> stopPonder(NULL);
			break;
		}
		else if (parts -> at(0) == "setparam" || parts -> at(0) == "setvalue") {
			if (parts -> size() == 3) {
				if (!b -> setParam(parts -> at(1), parts -> at(2)))
					emitDebug("setparam: invalid key and/or value");
			}
			else {
				emitDebug("unexpected size of setparam");
			}
		}
		else if (parts -> at(0) == "setoption")
		{
			// setoption name Ponder value true
			if (parts -> at(2) == "SyzygyPath" && parts -> size() >= 5)
			{
#ifdef W_SYZYGY
				std::string path = parts -> at(4);

				struct stat st;
				if (stat(path.c_str(), &st) != 0)
				{
					log(ll_err, "stat on %s failed: %s", path.c_str(), strerror(errno));
					emitDebug("Path %s cannot be accessed: %s", path.c_str(), strerror(errno));
				}
				else
				{
					if (!init_syzygy_etb(path.c_str()))
					{
						std::string msg = myformat("Failed loading %s", path.c_str(), strerror(errno));
						log(ll_err, msg);
						emitDebug(msg);
					}
					else
					{
						std::string msg = myformat("Successfully loaded %s, %d pieces", path.c_str(), get_syzygy_etb_piece_count());
						log(ll_ok, msg);
						emitDebug(msg);
					}
				}
#endif
			}
			else if (parts -> at(2) == "Hash" && parts -> size() >= 5) {
				tt -> setSize(atoll(parts -> at(4).c_str()) * 1024ll * 1024ll - PTT_SIZE * 1024ll * 1024ll, 1);
				emitDebug("TT size set");
			}
#ifdef WITH_CONTEMPT
			else if (parts -> at(2) == "Contempt" && parts -> size() >= 5) {
				tune -> setParam("contemptMiddlegame", atoi(parts -> at(4).c_str()));
				emitDebug("Contempt set");
			}
			else if (parts -> at(2) == "Threads" && parts -> size() >= 5) {
				b -> setNTHreads(atoi(parts -> at(4).c_str()));
				emitDebug("Thread count set");
			}
#endif
			else if (parts -> at(2) == "Minimum Thinking Time" && parts -> size() >= 5) {
				tune -> setParam("minimumThinkTime", atoi(parts -> at(4).c_str()));
				emitDebug("Min think time set");
			}
			else if (parts -> at(2) == "UCI_Opponent" && parts -> size() >= 5) {
				std::string opponent;

				for(size_t i=4; i<parts -> size(); i++) {
					if (!opponent.empty())
						opponent += " ";

					opponent += parts -> at(i);
				}

				log(ll_ok, "UCI opponent: %s", opponent.c_str());

#ifdef WSJS
				if (ws) {
					json_t *j = json_object();
					add_json_key_value(j, "type", "opponent");
					add_json_key_value(j, "opponent", opponent.c_str());
					ws -> update(json_dumps(j, 0), true);
					json_decref(j);
				}
#endif
			}
			else
			{
				std::string msg = "Ignoring option " + parts -> at(2);
				log(ll_err, msg);
				emitDebug(msg);
			}
		}
		else if (parts -> at(0) == "debug")
			emitCmd("info string debug");
		else if (parts -> at(0) == "isready")
			emitCmd("readyok");
		else if (parts -> at(0) == "ucinewgame")
		{
			b -> stopPonder(NULL);
			history.clear();
			lastOppMove = invalidMove;
			emitCmd("isready");

			initialKibitzDelay = get_ts_ms();
			firstMove = true;

			resetStats();
		}
		else if (parts -> at(0) == "gameover" && parts -> size() == 2) {
		}
		else if (parts -> at(0) == "position")
		{
			bool moves = false;

			lastOppMove = invalidMove;

			for(size_t i=1; i<parts -> size(); i++)
			{
				if (parts -> at(i) == "startpos")
				{
					delete s;
					s = new Scene();

					c = WHITE;
					moveNr = 1;
					epMove = invalidMove;

					history.clear();
					zobrist_t hash = b -> calcZobrist(s -> getBoardConst(), c, invalidMove);
					addToHistory(&history, hash.newHash);
				}
				else if (parts -> at(i) == "fen")
				{
					moveNr = 1;
					epMove = invalidMove;

					// 1: board
					// 2: w/b
					// 3: castling availability
					// 4: last move (en-passant)
					// 5: half moves
					// 6: whole moves
					delete s;

					std::string f5 = i + 5 < parts -> size() ? parts -> at(i + 5) : "";
					std::string f6 = i + 6 < parts -> size() ? parts -> at(i + 6) : "";

					s = new Scene(parts -> at(i + 1) + " " + parts -> at(i + 2) + " " + parts -> at(i + 3) + " " + parts -> at(i + 4) + " " + f5 + " " + f6);

					c = parts -> at(i + 2) == "w" ? WHITE : BLACK;

					const Move *const em = s -> getLastOpponentMove();

					zobrist_t hash = b -> calcZobrist(s -> getBoardConst(), c, em ? *em : invalidMove);
					history.clear();
					addToHistory(&history, hash.newHash);

					i += 4;
				}
				else if (parts -> at(i) == "moves")
					moves = true;
				else if (moves)
				{
					std::string moveStr = parts -> at(i);

					// work around for double space by Chess for Android
					if (moveStr.empty())
						continue;

					Move mc = uciMoveStrToMove(s, moveStr);

					if (!isValidMove(mc))
						emitDebug("'%s' is an invalid move", moveStr.c_str());

					lastOppMove = mc;

					epMove = invalidMove;
					if (isEpMove(s -> getBoardConst(), mc))
						epMove = mc;

					BoardState bs;
					s -> doMove(mc, &bs);

					if (c == BLACK)
						moveNr++;

					c = opponentColor(c);

					zobrist_t hash = b -> calcZobrist(s -> getBoardConst(), c, epMove);
					addToHistory(&history, hash.newHash);
				}
				/*else
				{
					emit("info string: '%s' is not understood in the current \"position\" context", parts -> at(i).c_str());
				}*/
			}
		}
		else if (parts -> at(0) == "go")
		{
			std::string calculating = cfgGetString(parameters, "calc_start", "");

			if (!calculating.empty())
				forkProcess(calculating);

			int cmaxIt = maxIt;
			bool timeSet = false;

			bool ponder = false;
			Move ponderHit = invalidMove;
			VectorType<PVnode_t> ponderPV;
			b -> stopPonder(&ponderPV);

			if (ponderPV.size() >= 2) {
				if (moveCompareTo(ponderPV.at(0).m, lastOppMove)) {
					ponderHit = ponderPV.at(1).m;
					emitDebug("ponder hit %s", pvToString(ponderPV, false, false, false).c_str());
				}
				else {
					tellAll("I would have chosen %s (%d)", moveToString(ponderPV.at(0).m).c_str(), ponderPV.at(0).eval);
				}
			}

			bool isAsIsThinkTime = true;
			for(size_t i=1; i<parts -> size(); i++)
			{
				if (parts -> at(i) == "wtime")
				{
					wTime = atoi(parts -> at(++i).c_str());
					isAsIsThinkTime = false;
					timeSet = true;
				}
				else if (parts -> at(i) == "btime")
				{
					bTime = atoi(parts -> at(++i).c_str());
					isAsIsThinkTime = false;
					timeSet = true;
				}
				else if (parts -> at(i) == "winc")
				{
					wInc = atoi(parts -> at(++i).c_str());
					timeSet = true;
				}
				else if (parts -> at(i) == "binc")
				{
					bInc = atoi(parts -> at(++i).c_str());
					timeSet = true;
				}
				else if (parts -> at(i) == "infinite") {
					timeSet = false;
					cmaxIt = -1;
				}
				else if (parts -> at(i) == "movestogo")
				{
					maxNMoves = atoi(parts -> at(++i).c_str());
					isAsIsThinkTime = false;
				}
				else if (parts -> at(i) == "movetime")
				{
					wTime = bTime = atoi(parts -> at(++i).c_str());
					isAsIsThinkTime = true;
					timeSet = true;
				}
				else if (parts -> at(i) == "ponder")
				{
					ponder = true;
				}
				else if (parts -> at(i) == "depth")
				{
					cmaxIt = atoi(parts -> at(++i).c_str());
				}
				else
				{
					emitDebug("ignoring unknown " + parts -> at(i));
				}
			}

			if (ponder) {
				startPonder(moveNr, *s, c, epMove, &history, lastOppMove);
			}
			else {
				// polyglot waits 5 seconds by default before it will do kibitz
				if (firstMove && get_ts_ms() - initialKibitzDelay >= 5500) {
					firstMove = false;

					tellAll("You're playing against %s", versionStr);

					tellAll("Build on: " __DATE__ " " __TIME__);

					const char *hw = getHardware();
					if (hw) {
						tellAll("Running on: %s", hw);
						free((void *)hw);
					}
				}

				emitDebug("color %s", c == WHITE ? "white" : "black");
				int ms = (c == WHITE) ? wTime : bTime;
				double currentDurationMin = ms / 2;
				double currentDurationMax = ms;

				if (isAsIsThinkTime == false)
				{
					int movesLeft = maxNMoves - moveNr;
					int curNMoves = movesLeft <= 0 ? defaultNMovesPerGame : movesLeft;

					int timeInc = (c == WHITE) ? wInc : bInc;

					currentDurationMin = (ms + (curNMoves - 1) * timeInc) / double(curNMoves + 7);
					currentDurationMax = (ms + (curNMoves - 1) * timeInc) / (curNMoves * 3.0 / 4.0);

					int limitDurationMin = ms / 15;
					if (currentDurationMin > limitDurationMin) {
						log(ll_ok, "limit min from %f to %d (%f)", currentDurationMin, limitDurationMin, -(currentDurationMin - limitDurationMin));
						currentDurationMin = limitDurationMin;
					}
					int limitDurationMax = ms / 10;
					if (currentDurationMax > limitDurationMax) {
						log(ll_ok, "limit max from %f to %d (%f)", currentDurationMax, limitDurationMax, -(currentDurationMax - limitDurationMax));
						currentDurationMax = limitDurationMax;
					}
				}

				currentDurationMin = std::max(currentDurationMin, double(tune -> minThinkTime));
				currentDurationMax = std::max(currentDurationMax, double(tune -> minThinkTime));

				log(ll_ok, "Total time left: %d, selected: %f/%f, moveNr: %d", ms, currentDurationMin, currentDurationMax, moveNr);

				if (!timeSet)
					currentDurationMin = currentDurationMax = -10.0;

				VectorType<PVnode_t> *ml = NULL;
				int value = INT_MIN;

				if (stopFlag)
					break;

				uint64_t startTs = get_ts_ms();
				prevHashFull = startTs;

				ei.flag = false;
				ct_t ct = { b, s, c, cmaxIt, int(currentDurationMin), int(currentDurationMax), NULL, &history, epMove, &ml, &value, &ei, {-1, -1}, ponderHit, lastOppMove, &thinking, moveNr };
				pipe(ct.p);

				struct pollfd fds[2] = { { fileno(stdin), POLLIN, 0 }, { ct.p[0], POLLIN, 0 } };

				std::thread *calcThread = new std::thread(calculateThread, &ct);

				for(;!stopFlag && !ei.flag;) {
					uint64_t nowTs = get_ts_ms();
					int64_t leftTs = currentDurationMax * 1000.0 - int64_t(nowTs - startTs);
					if (leftTs < 1)
						leftTs = 1;
					int rc = poll(fds, 2, leftTs);

					if (rc == -1)
						break;

					if (fds[0].revents == POLLIN) {
						std::string *stop = getLineFromSTDIN();
						if (!stop)
							break;

						if (*stop == "stop" || *stop == "quit") {
							stopFlag = *stop == "quit";
							ei.flag = true;
							delete stop;
							break;
						}

						//ungetBuffer.push_back(*stop);

						delete stop;
					}

					if (fds[1].revents == POLLIN) {
						char buffer[4096];
						read(ct.p[0], buffer, sizeof(buffer));
						break;
					}
				}

				close(ct.p[1]);
				close(ct.p[0]);

//			uint64_t nowTs = get_ts_ms();
//			printf("%ld %f %f\n", nowTs - startTs, currentDurationMin, currentDurationMax);

				calcThread -> join();
				delete calcThread;

				if (stopFlag) {
					emitDebug("string terminate");
					delete parts;
					break;
				}

				emitCmd("info score cp %d", value);

				if (ml -> empty())
					emitCmd("bestmove a1a1");
				else
				{
					Move bm = ml -> at(0).m;

					std::string out = "bestmove " + moveToUciMove(bm);

// FIXME need to understand "ponderhit"					if (ml -> size() >= 2)
// FIXME need to understand "ponderhit"						out += " ponder " + moveToUciMove(ml -> at(1));

					emitCmd(out);

					if (verbose) {
						emitDebug("Considered moves:");

						VectorType<Move> *const cml = s -> getMoveList(c);

						for(size_t i=0; i<cml -> size(); i++) {
							if (moveCompareTo(cml -> at(i), bm) == false)
								emitPvDetails(s, c, cml -> at(i));
						}
					}

					if (isEpMove(s -> getBoardConst(), bm))
						epMove = bm;
					else
						epMove = invalidMove;

					lastOppMove = bm;

					BoardState bs;
					s -> doMove(bm, &bs);
				}

				c = opponentColor(c);

				if (alwaysPonder)
					startPonder(moveNr, *s, c, epMove, &history, lastOppMove);

				std::string finished = cfgGetString(parameters, "calc_finished", "");

				if (!finished.empty())
					forkProcess(finished);

				delete ml;

				s -> generateMovelists();
			}
		}
		else if (!genericCommands(b -> calcZobrist(s -> getBoardConst(), c, invalidMove), parts, NULL, s, c)) {
			emitCmd("Unknown command: %s", temp.c_str());
		}

		delete parts;
	}

	log(ll_ok, "Terminating");

	delete s;
}
