#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <map>
#include <poll.h>
#include <string>
#include <thread>
#include <unistd.h>

#include "IO_Xboard.h"
#include "Utils.h"
#include "Brain.h"
#include "Main.h"
#include "Tune.h"

IO_Xboard::IO_Xboard(const std::map<std::string, std::string> & pin, Tpt *const ttIn, Ptt *const pttIn) : IO(pin, ttIn, pttIn)
{
	name_extra = cfgGetString(pin, "name_extra", "");

	if (!name_extra.empty())
		name_extra = "-" + name_extra;
}

IO_Xboard::~IO_Xboard()
{
}

void IO_Xboard::emitCmd(const std::string what, ...) const
{
	char *msg = NULL;

       	va_list ap;
       	va_start(ap, what);
	(void)vasprintf(&msg, what.c_str(), ap);
	va_end(ap);

	printf("%s\n", msg);
	fflush(stdout);

	log(ll_ok, "CMD: " + std::string(msg));

	free(msg);
}

void IO_Xboard::loop()
{
	bool usePonder = alwaysPonder;

	int cmaxIt = -1;
	bool force = false;
	PlayerColor c = WHITE;
	Scene *s = NULL;
	new_Scene(&s, &c);

	std::vector<uint64_t> history;

	Move lastOppMove = invalidMove, epMove = invalidMove;

	std::vector<pos_t> h;

	int nMovesPlayed = 1;
	int nMoves = 40, base = 5, inc = 0;
	int time = 1000;

	for(;;) {
		std::string *const line = getLineFromSTDIN();
		if (!line)
			break;

		if (line -> empty()) {
			delete line;
			continue;
		}

		log(ll_ok, "IN: " + *line);

		std::vector<std::string> *const parts = split(*line, " ");

		std::string p1 = parts -> at(0);

		std::string moveStr;
		if (p1.size() >= 4 && isalpha(p1.at(0)) && isdigit(p1.at(1)) && isalpha(p1.at(2)) && isdigit(p1.at(3)))
			moveStr = p1;
		else if (p1 == "usermove" && parts -> size() == 2)
			moveStr = parts -> at(1);

		bool hint = parts -> at(0) == "hint";

		if (parts -> at(0) == "xboard")
			emitCmd("");
		else if (parts -> at(0) == "level") {
			nMoves = atoi(parts -> at(1).c_str());
			base = atoi(parts -> at(2).c_str());
			inc = atoi(parts -> at(3).c_str());
		}
		else if (parts -> at(0) == "easy") {
			usePonder = false;
			b -> stopPonder(NULL);
		}
		else if (parts -> at(0) == "setboard") {
			delete s;
			s = new Scene(line -> substr(9));
			c = parts -> at(2) == "w" ? WHITE : BLACK;
		}
		else if (parts -> at(0) == "hard")
			usePonder = true;
		else if (parts -> at(0) == "time") {
			// +5: on average xboard will have 5ms less than uci because of centiseconds versus milliseconds
			time = atoi(parts -> at(1).c_str()) * 10 + 5;
		}
		else if (parts -> at(0) == "protover") {
			emitCmd("feature done=0 ping=0 usermove=1 time=1 draw=1 sigint=0 myname=\"Embla " NAME_EXTRA "%s\" colors=1 ics=1 name=1 done=1", name_extra.c_str());
		}
		else if (parts -> at(0) == "new") {
			delete s;

			usePonder = alwaysPonder;

			force = false;
			nMovesPlayed = 1;

			c = WHITE;
			new_Scene(&s, &c);

			history.clear();
			h.clear();

			b -> stopPonder(NULL);
			resetStats();
		}
		else if (parts -> at(0) == "quit")
			break;
		else if (parts -> at(0) == "random") {
		}
		else if (parts -> at(0) == "force") {
			force = true;
		}
		else if (parts -> at(0) == "go" || !moveStr.empty() || hint) {
			if (parts -> at(0) == "go") {
				log(ll_ok, "reset force flag");
				force = false;
			}

			Move mc = invalidMove;
			if (!moveStr.empty()) {
				h.push_back({ genFen(s, false, c), moveStr, 0/* FIXME evaluate(s, c, 0)*/ });

				mc = uciMoveStrToMove(s, moveStr.c_str());
				if (!isValidMove(mc))
					emitDebug("'%s' is an invalid move", moveStr.c_str());

				lastOppMove = mc;

				epMove = invalidMove;
				if (isEpMove(s -> getBoardConst(), mc))
					epMove = mc;

				BoardState bs;
				s -> doMove(mc, &bs);

				if (c == BLACK) {
					log(ll_ok, "movecount inc 1");
					nMovesPlayed++;
				}

				c = opponentColor(c);
			}

			if (!force || hint) {
				///
				VectorType<PVnode_t> ponderPV;
				b -> stopPonder(&ponderPV);

				cmaxIt = -1;
//				if (ponderPV.size() >= 1 && moveCompareTo(ponderPV.at(0).m, lastOppMove)) {
//					cmaxIt = 3;
//					emitDebug("ponder hit %s", pvToString(ponderPV, false, false, false).c_str());
//				}

				uint64_t start_ts = get_ts_ms();

				VectorType<PVnode_t> *ml = NULL;
				int value = INT_MIN;

				///
				int movesLeft = 40 - nMovesPlayed;
				int curNMoves = movesLeft <= 0 ? 40 : movesLeft;

				int ms = time;
				int currentDurationMin = (ms + (curNMoves - 1) * inc) / double(curNMoves + 7);
				int currentDurationMax = (ms + (curNMoves - 1) * inc) / (curNMoves * 3.0 / 4.0);

				int limitDurationMin = ms / 15;
				if (currentDurationMin > limitDurationMin) {
					log(ll_ok, "limit min from %d to %d", currentDurationMin, limitDurationMin);
					currentDurationMin = limitDurationMin;
				}
				int limitDurationMax = ms / 10;
				if (currentDurationMax > limitDurationMax) {
					log(ll_ok, "limit max from %d to %d", currentDurationMax, limitDurationMax);
					currentDurationMax = limitDurationMax;
				}

				currentDurationMin = std::max(currentDurationMin, tune -> minThinkTime);
				currentDurationMax = std::max(currentDurationMax, tune -> minThinkTime);

				log(ll_ok, "Total time left: %d, selected: %f/%f, moveNr: %d", ms, currentDurationMin / 1000.0, currentDurationMax / 1000.0, nMovesPlayed);
				///

				ei.flag = false;
				ct_t ct = { b, s, c, cmaxIt, currentDurationMin, currentDurationMax, NULL, &history, epMove, &ml, &value, &ei, {-1, -1}, invalidMove, lastOppMove, &thinking, nMovesPlayed };
				pipe(ct.p);

				struct pollfd fds[2] = { { fileno(stdin), POLLIN, 0 }, { ct.p[0], POLLIN, 0 } };

				std::thread *calcThread = new std::thread(calculateThread, &ct);

				for(;!stopFlag && !ei.flag;) {
					uint64_t now_ts = get_ts_ms();
					int64_t left_ts = currentDurationMax - int64_t(now_ts - start_ts);
					if (left_ts < 1)
						left_ts = 1;
					int rc = poll(fds, 2, left_ts);

					if (rc == -1)
						break;

					if (fds[0].revents == POLLIN) {
						std::string *stop = getLineFromSTDIN();
						if (!stop)
							break;

						if (*stop == "stop" || *stop == "quit") {
							stopFlag = *stop == "quit";
							ei.flag = true;
							break;
						}

						if (*stop == "?") {
							ei.flag = true;
							break;
						}

						//ungetBuffer.push_back(*stop);

						delete stop;
					}

					if (fds[1].revents == POLLIN) {
						char buffer[4096];
						read(ct.p[0], buffer, sizeof(buffer));
						break;
					}
				}

				close(ct.p[1]);
				close(ct.p[0]);

				calcThread -> join();
				delete calcThread;

				if (stopFlag) {
					emitDebug("string terminate");
					delete parts;
					break;
				}

				if (ml -> empty()) {
					if (hint)
						emitCmd("Hint: ?");
					else
						emitCmd("move a1a1");
				}
				else {
					Move bm = ml -> at(0).m;
					lastOppMove = bm;

					std::string resultMove = moveToUciMove(bm);
					if (hint)
						emitCmd("Hint: %s", resultMove.c_str());
					else {
						emitCmd("move %s", resultMove.c_str());

						h.push_back({ genFen(s, false, c), resultMove, /* FIXME evaluate(s, c, 0)*/ 0 });

						epMove = invalidMove;
						if (isEpMove(s -> getBoardConst(), bm))
							epMove = bm;

						BoardState bs;
						s -> doMove(bm, &bs);

						if (c == BLACK) {
							log(ll_ok, "movecount inc 2");
							nMovesPlayed++;
						}

						c = opponentColor(c);

						if (usePonder)
							startPonder(nMovesPlayed, *s, c, epMove, &history, lastOppMove);
					}
				}
			}
		}
		else if (parts -> at(0) == "white") {
			c = WHITE;
		}
		else if (parts -> at(0) == "black") {
			c = BLACK;
		}
		else if (parts -> at(0) == "undo") {
			if (!h.empty()) {
				std::string fen = h.back().fen;
				h.pop_back();

				delete s;
				c = opponentColor(c);
				s = new Scene(fen);
			}
		}
		else if (parts -> at(0) == "result") {
			// FIXME
		}
		else if (!genericCommands(b -> calcZobrist(s -> getBoardConst(), c, invalidMove), parts, &h, s, c)) {
			// FIXME error
		}

		delete parts;
		delete line;
	}

	delete s;
}
