// (C) Folkert van Heusden. AGPL v3.0.
// $Revision: 5115 $
#include "Tweaks.h"
#include <ctype.h>
#if !defined(__CYGWIN__) && !defined(ANDROID)
#include <execinfo.h>
#endif
#include <signal.h>
#include <stdio.h>
#include <string.h>

#include "config.h"
#include "error.h"
#include "IO_UCI.h"
#include "Tune.h"
#include "Scene.h"
#include "Brain.h"
#include "Utils.h"
#include "Tpt.h"
#include "IO_stderr.h"
#include "IO_AutoPlay.h"
#include "IO_EPD.h"
#include "IO_Console.h"
#include "IO_perft.h"
#include "IO_Xboard.h"
#include "Book.h"
#include "BookPolyglot.h"
#include "DBBook.h"
#include "DBBookEval.h"
#include "DBBookLearn.h"
#include "DBBookWDL.h"
#include "DetectBookFile.h"
#include "DBBookEvalSqlite3.h"
#include "ws.h"
#include "http_server.h"

#ifdef W_SYZYGY
#include "syzygy_etb/fathom.h"
#endif

char versionStr[] = "Embla v" VERSION " " NAME_EXTRA ", (C) 1995-2019 by folkert@vanheusden.com";

[[ noreturn ]] void sigh(int sig)
{
	log(ll_err, "Caught signal %d", sig);

#if !defined(__CYGWIN__) && !defined(ANDROID)
#define BT_SIZE 128
	void *buffer[BT_SIZE];
	int nptrs = backtrace(buffer, BT_SIZE);
	log(ll_err, "backtrace() returned %d addresses\n", nptrs);

	char **strings = backtrace_symbols(buffer, nptrs);
	if (!strings)
		log(ll_err, "backtrace_symbols failed: %s", strerror(errno));
	else
	{
		for(int i = 0; i < nptrs; i++)
			log(ll_err, "%3d %s\n", i, strings[i]);
	}
#endif
	exit(1);
}

IO *io = new IO_stderr();

void terminate(int sig)
{
	printf("Caught signal; terminating!\n");

	io -> terminate();
}

void version()
{
	printf("%s\n\n", versionStr);
}

void help()
{
	version();

	printf("-H x   hash size in MB, default %ldMB\n", DEFAULT_HASHTABLE_SIZE);
#ifdef THREADING
	printf("-c x   number of threads, defaults to number of cpu cores\n");
#endif
	printf("-p     parameters, comma seperated key=value pairs\n");
	printf("-m x   mode: uci (default), console (run in a terminal, no extra gui required), auto(-play), epd\n");
	printf("-l x   logfile\n");
#ifdef W_SYZYGY
	printf("-s x   path to syzygy-etb database - can also be set via UCI interface\n");
#endif
	printf("-l x   logfile\n");
	printf("-b x   select a (polyglot) book\n");
#ifdef WITH_MYSQL_BOOKS
	printf("-B     use no books\n");
#endif
	printf("-S     set signal handlers\n");
	printf("-t x   tune settings\n");
	printf("-K     in UCI mode, do Kibitz\n");
	printf("-V     show version and exit\n");
	printf("-h     this help\n");
	printf("\n");
	printf("Parameters (-p):\n");
	for(const std::pair<std::string, std::string> & cur : IO::getParameters())
		printf("\t%s = %s\n", cur.first.c_str(), cur.second.c_str());
	printf("\n");
	for(const std::pair<std::string, std::string> & cur : Brain::getParameters())
		printf("\t%s = %s\n", cur.first.c_str(), cur.second.c_str());
	printf("\n");
	printf("websockets_port	use together with http_port\n");
	printf("http_port	internal webserver showing progress\n");
}

typedef enum { M_UCI, M_XBOARD, M_CONSOLE, M_AUTO, M_EPD, M_PERFT } use_mode_t;

int main(int argc, char *argv[])
{
	std::string tptPath = "./";
	std::string tuneFile;

	uint64_t hashSize = DEFAULT_HASHTABLE_SIZE * 1024L * 1024L;
	use_mode_t m = M_UCI;

	bool withBooks = true, verbose = false, kibitz = false;

#ifdef W_SYZYGY
	const char *syzygy_etb_path = NULL;
#endif

#ifdef WITH_MYSQL_BOOKS
	init_db();
#endif

	// thread count
#ifdef THREADING
	int nCores = std::thread::hardware_concurrency();
	if (nCores == 0)
		nCores = (int)sysconf(_SC_NPROCESSORS_ONLN);
#else
	int nCores = 1;
#endif
	log(ll_ok, "Number of cores/threads detected: %d", nCores);

	std::map<std::string, std::string> parameters;

	std::vector<Book *> books;

	log(ll_ok, "Start this program as follows to see a list of options: %s -h", argv[0]);

	signal(SIGCHLD, SIG_IGN);
	signal(SIGWINCH, SIG_IGN);

	int c = -1;
	while((c = getopt(argc, argv, "Kt:vSs:l:Bb:Mp:m:H:c:Vh")) != -1)
	{
		switch(c)
		{
			case 'K':
				kibitz = true;
				break;

			case 't':
				tuneFile = optarg;
				break;

			case 'S':
				signal(SIGABRT, sigh);
				signal(SIGILL,  sigh);
				signal(SIGSEGV, sigh);
				break;

			case 'l':
				setLogfile(optarg);
				break;

#ifdef W_SYZYGY
			case 's':
				syzygy_etb_path = optarg;
				break;
#endif

			case 'B':
				withBooks = false;
				break;

			case 'b': {
					  book_fileformat_t bf = detectBookType(optarg);

					  if (bf == bt_polyglot) {
						  log(ll_ok, "%s recognized as a polyglot book", optarg);
						  books.push_back(new BookPolyglot(io, optarg));
					  }
#ifdef W_SQLITE3
					  else if (bf == bt_sqlite3) {
						  log(ll_ok, "%s recognized as an sqlite3 book", optarg);
						  books.push_back(new DBBookEvalSqlite3(io, optarg));
					  }
#endif
					  else {
						  error_exit(false, "Book fileformat for %s not recognized", optarg);
					  }
				  }
				  break;

			case 'm':
				if (strcasecmp(optarg, "uci") == 0)
					m = M_UCI;
				else if (strcasecmp(optarg, "xboard") == 0)
					m = M_XBOARD;
				else if (strcasecmp(optarg, "console") == 0)
					m = M_CONSOLE;
				else if (strcasecmp(optarg, "auto") == 0)
					m = M_AUTO;
				else if (strcasecmp(optarg, "epd") == 0)
					m = M_EPD;
				else if (strcasecmp(optarg, "perft") == 0)
					m = M_PERFT;
				else
					error_exit(false, "Mode %s not recognized", optarg);

				break;

			case 'p':
				{
					std::vector<std::string> *parts = split(optarg, ",");

					for(size_t i=0; i<parts -> size(); i++)
					{
						std::string cur = parts -> at(i);
						ssize_t is = cur.find("=");
						if (is == -1)
							error_exit(false, "Key-value pair missing '='");

						std::string key = cur.substr(0, is), value = cur.substr(is + 1);

						log(ll_ok, "Adding key/value parameter pair: %s=%s", key.c_str(), value.c_str());

						parameters.insert(std::pair<std::string, std::string>(key, value));
					}

					delete parts;
				}
				break;

			case 'H':
				hashSize = atoll(optarg) * 1024ll * 1024ll;
				break;

			case 'c':
#ifdef THREADING
				nCores = atoi(optarg);
#endif
				break;

			case 'v':
				verbose = true;
				break;

			case 'V':
				version();
				return 0;

			case 'h':
				help();
				return 0;

			default:
				fprintf(stderr, "%c is not understood\n\n", c);
				help();
				return 1;
		}
	}

	log(ll_ok, versionStr);

	emitConfig();

	// 31 is average number of elements in MoveList
	unsigned int imls = int(upper_power_of_two(sizeof(Move) * AVERAGE_MOVELIST_SIZE) / sizeof(Move));
	log(ll_ok, "Initial move list size: %d, size of Move-structure: %d, total: %d", imls, sizeof(Move), imls * sizeof(Move));
	setMoveListInitialSize(imls);

	log(ll_ok, "Hash size: %luMB", hashSize / 1024 / 1024);

#ifdef W_SYZYGY
	if (syzygy_etb_path) {
		init_syzygy_etb(syzygy_etb_path);

		log(ll_ok, "syzygy piece count: %d", get_syzygy_etb_piece_count());
	}
#endif

#if 0
	Tpt *ctt = new Tpt("tt", io, 128, 1);
	uint64_t cnt = 0, strt = get_ts_ms();
	constexpr uint64_t d = 6100;
	constexpr int in = 10000;
	do {
		for(int i=0; i<in; i++)
			ctt -> store(cnt++, EXACT, 0, 0, invalidMove.u.m, false);
	}
	while(get_ts_ms() - strt < d);
	printf("%f\n", cnt / double(d));
	return 1;
#endif
#if 0
	uint64_t cnt = 0, strt = get_ts_ms();
	constexpr uint64_t d = 61000;
	constexpr int in = 10000;
	tptEntry te;
	do {
		for(int i=0; i<in; i++)
			tt -> lookup(0, &te);

		cnt += in;
	}
	while(get_ts_ms() - strt < d);
	printf("%f\n", cnt / double(d));
	return 1;
#endif

#if 0
	//Scene stest("8/1k6/2P5/8/8/8/6K1/8 w - -");
	Scene stest("2r1b3/1k6/r1P3r1/8/b1r1b3/8/6K1/8 w - -");
	stest.generateMovelists();
	printf("%d\n", stest.getIsUnderAttackByPawn(stest.getKing(BLACK)));
	printf("%d\n", stest.getIsUnderAttackByPawn(stest.getKing(WHITE)));

	Scene stest2("8/1k6/ppp5/2P5/4p3/5P1P/6K1/8 w - -");
	stest2.generateMovelists();
	int kingPawnShield(const Scene *const restrict s, const PlayerColor me);
	printf("%d\n", kingPawnShield(&stest2, WHITE));
	printf("%d\n", kingPawnShield(&stest2, BLACK));
	return 1;
#endif

#ifdef USE_TT
	Tpt *tt = new Tpt("tt", io, hashSize - PTT_SIZE * 1024 * 1024, nCores);
#else
	Tpt *tt = NULL;
#endif

#ifdef PAWN_HASH
	Ptt *ptt = new Ptt("", PTT_SIZE * 1024 * 1024, nCores, "");
#else
	Ptt *ptt = NULL;
#endif

#if 0
	Scene s("K7/8/k2n1n2/8/2q1P3/8/3N1N2/8 w - -");
	s.generateMovelists();
	printf("-550: %d\n", s.getSEE(4, 3, s.getBoardConst() -> getAt(2, 3)));
	printf(" 100: %d\n", s.getSEE(4, 3, s.getBoardConst() -> getAt(3, 5)));
	printf(" 100: %d\n", s.getSEE(4, 3, s.getBoardConst() -> getAt(5, 5)));
	Scene s2("8/2q2k2/4rb2/4pR2/5P2/3N4/Q7/4K3 w - - 0 1");
	s2.generateMovelists();
	printf("   0: %d\n", s2.getSEE(4, 4, s2.getBoardConst() -> getAt(5, 3)));
	return 1;
#endif

	// not thread-safe
	IO *old = io;
	if (m == M_UCI)
		io = new IO_UCI(parameters, tt, ptt);
	else if (m == M_XBOARD)
		io = new IO_Xboard(parameters, tt, ptt);
#ifdef W_READLINE
	else if (m == M_CONSOLE)
		io = new IO_Console(parameters, tt, ptt);
#endif
	else if (m == M_AUTO)
		io = new IO_AutoPlay(parameters, tt, ptt);
	else if (m == M_EPD)
		io = new IO_EPD(parameters, tt, ptt);
	else if (m == M_PERFT)
		io = new IO_perft(parameters, tt, ptt);

	delete old;

#ifdef WSJS
	wsserver *ws = NULL;
	int ws_port = cfgGetInt(parameters, "websockets_port", -1);
	if (ws_port != -1) {
		log(ll_ok, "Starting websockets on port %d", ws_port);
		ws = new wsserver(ws_port);
		io -> setWebSockets(ws);
	}
	else {
		log(ll_ok, "Not starting websockets");
	}

	http_server *hs = NULL;
	int http_port = cfgGetInt(parameters, "http_port", -1);
	if (http_port != -1) {
		log(ll_ok, "Starting http server on port %d", http_port);
		hs = new http_server(http_port, ws_port);
	}
	else {
		log(ll_ok, "Not starting http server");
	}
#endif

	io -> setVerbose(verbose);
	io -> setKibitz(kibitz);

	initTune(tuneFile);

	if (withBooks)
	{
#ifdef WITH_MYSQL_BOOKS
//		books.push_back(new DBBookEval(io, "192.168.64.100", "PMbook", "dbp", "readbook", "entries"));
//		books.push_back(new DBBookLearn(io, "192.168.64.1", "emblabook", "embla", "book", ""));
#endif
	}

	Brain *b = new Brain(io, tt, ptt, nCores, &books);

	io -> setBrain(b);
	io -> setBooks(&books);

	for(const std::pair<std::string, std::string> & cur : b -> getParameters()) {
		std::map<std::string, std::string>::const_iterator it = parameters.find(cur.first);

		if (it != parameters.end())
			b -> setParam(cur.first, it -> second);
	}

	io -> emitDebug("Embla v" VERSION " " NAME_EXTRA);
	io -> emitDebug("Using %d threads", nCores);

	uint64_t t = get_ts_ms();
	srand(t ^ (t >> 32));

	signal(SIGINT, terminate);
	signal(SIGTERM, terminate);
	signal(SIGPIPE, SIG_IGN);

	log(ll_ok, "initial move list size: %d", imls);

#ifdef USE_TT
	tt -> registerIO(io);
#endif

	io -> loop();

	delete b;

	signal(SIGINT, SIG_IGN);
	signal(SIGTERM, SIG_IGN);
	delete ptt;
	delete tt;

	for(Book *const cur : books)
		delete cur;

	freeTune();

	io -> emitDebug("Embla terminating");

	delete io;

#ifdef WSJS
	delete hs;
	delete ws;
#endif

#ifdef WITH_MYSQL_BOOKS
	mysql_library_end();
#endif

	return 0;
}
