// (C) Folkert van Heusden. AGPL v3.0.
// $Revision: 4914 $
#include "Tweaks.h"
#include <stdlib.h>

#include "config.h"
#include "Exception.h"
#include "PV.h"
#include "IO.h"

void assign(VectorType<PVnode_t> *const pv, const VectorType<Move> & source, const bool reverse)
{
	const size_t n = source.size();

	pv -> setSize(n);

	if (reverse) {
		for(size_t i=0; i<n; i++) {
			PVnode_t cur = { 0, source.at(n - i - 1), 0, MS_ML_CONV };

			pv -> set(i, cur);
		}
	}
	else {
		for(size_t i=0; i<n; i++) {
			PVnode_t cur = { 0, source.at(i), 0, MS_ML_CONV };

			pv -> set(i, cur);
		}
	}
}

ssize_t indexOf(const VectorType<PVnode_t> & pv, const Move & which)
{
	const size_t nIn = pv.size();

	for(size_t i=0; i<nIn; i++) {
		if (moveCompareTo(pv.at(i).m, which))
			return i;
	}

	return -1;
}

ssize_t indexOf(const VectorType<PVnode_t> & pv, const uint64_t hash)
{
	const PVnode_t *const p = pv.raw();
	const size_t nIn = pv.size();

	for(size_t i=0; i<nIn; i++) {
		if (p[i].hash == hash)
			return i;
	}

	return -1;
}

// FIXME add flag for score output
std::string pvToString(const VectorType<PVnode_t> & pv, const bool reverse, const bool debug, const bool uciMode)
{
	const size_t nIn = pv.size();

	std::string out;

	for(size_t i=0; i<nIn; i++)
	{
		Move m = pv.at(i).m;

		std::string cur = moveToUciMove(m);

		if (uciMode && m.u.m.mt == CASTLING)
			cur = myformat("%c%c%c%c", m.u.m.fx + 'a', m.u.m.fy + '1', m.u.m.tx + 'a', m.u.m.ty + '1');

		if (debug) {
			const char *src_str[] = { "search", "qs", "tpt", "syzyg", "book", "dummy", "only1move", "nomovesel", "mlconv" };

			const PVnode_t p = pv.at(i);
			cur += myformat("/%d/%s", p.eval, src_str[p.src]);
		}

		if (out.empty())
			out = cur;
		else if (reverse)
			out = cur + " " + out;
		else
			out += " " + cur;
	}

	return out;
}

void push_back(VectorType<PVnode_t> *const pv, const uint64_t hash, const Move & m, const int score, const move_source_t src)
{
	const PVnode_t n = { hash, m, score, src };

	pv -> push_back(n);
}

#define MAX_PV_DETAILS_DEPTH 500 // pv may be max 500 moves

std::string recreatePvFromHash(Tpt *const tt, const Scene *const sIn, const PlayerColor cIn, Move m, const bool debug, const bool uciMode)
{
	Scene s(*sIn);
	s.generateMovelists();

	PlayerColor c = cIn;
	const Board *const bCur = s.getBoardConst();

	Move epMove = isEpMove(bCur, m) ? m : invalidMove;
	zobrist_t zh = tt -> calcZobrist(bCur, c, epMove);

	std::string out = uciMode ? moveToUciMove(m) : moveToString(m, true, bCur, s.isKingUnderAttack(c));

	int sanityLimit = 0;

	for(; sanityLimit < MAX_PV_DETAILS_DEPTH;) {
		epMove = isEpMove(bCur, m) ? m : invalidMove;

		BoardState bs;
		s.doMove(m, &bs);

		PlayerColor co = opponentColor(c);

		zh = tt -> updateZobrist(zh, bs, bCur, co, epMove);

		tptEntry ttEntry;
		if (!tt -> lookup(zh.newHash, &ttEntry))
			break;

		//const tptEntryFlag flags = (tptEntryFlag)ttEntry.flags;

		s.generateMovelists();
		VectorType<Move> *const moves = s.getMoveList(co);
		m.u.m = ttEntry.data_._data.m;
// FIXME check if valid Move validateMoves, see evalTT
		if (!isValidMove(m) || indexOf(*moves, m) == -1)
			break;

		if (debug) {
			out += myformat("[%d@%d]", ttEntry.data_._data.value, ttEntry.data_._data.depth);

			if (ttEntry.data_._data.flags == EXACT)
				out += "E";
			else if (ttEntry.data_._data.flags == LOWERBOUND)
				out += "L";
			else if (ttEntry.data_._data.flags == UPPERBOUND)
				out += "U";
			else
				out += "?";
		}

		std::string currentMoveString = uciMode ? moveToUciMove(m) : moveToString(m, true, bCur, s.isKingUnderAttack(co));
		if (out.find(currentMoveString) != std::string::npos) // repeating moves
			break;

		out += " " + currentMoveString;

		c = co;

		sanityLimit++;
	}

	return out;
}
