// (C) Folkert van Heusden. Not to be distributed nor open-source.
// $Revision: 4080 $
#include "Tweaks.h"
#include "config.h"
#include <cstdint>
#include <stdlib.h>
#include <string.h>
#ifdef USE_MMAP
#include <sys/mman.h>
#endif

#include "Exception.h"
#include "Ptt.h"
#include "Utils.h"
#include "IO.h"
#include "error.h"

extern uint64_t Random64[];

#define GB (1024 * 1024 * 1024)

#ifdef PAWN_HASH
Ptt::Ptt(const std::string & nameIn, const uint64_t nBytes, const int nThreads, const std::string & pathIn) : name(nameIn), path(pathIn), age(0)
{
	nEntries = nBytes / sizeof(pttHashGroup);
	nBytesTruncated = nEntries * sizeof(pttHashGroup);

	isMmap = false;
#ifdef USE_MMAP
#ifdef __ia64__
	void *addr = 0x8000000000000000UL;
#else
	void *addr = NULL;
#endif
	// http://stackoverflow.com/questions/2688466/why-mallocmemset-is-slower-than-calloc says the zeroing only takes place when a page is first written to
	int mmap_flags = MAP_PRIVATE | MAP_ANONYMOUS;
	int extra_mmap_flags = 0;

#ifdef MAP_HUGE_SHIFT
	mmap_flags |= MAP_HUGETLB;

#if !defined(MAP_HUGE_1GB) || !defined(MAP_HUGE_2MB)
	#define MAP_HUGE_2MB    (21 << MAP_HUGE_SHIFT)
	#define MAP_HUGE_1GB    (30 << MAP_HUGE_SHIFT)
#endif
	log(ll_ok, "%s size in bytes: %ld", name.c_str(), nBytesTruncated);
	int remainder = nBytesTruncated % GB;
	if (remainder >= MMAP_1GB_THRESHOLD || remainder == 0)
	{
		log(ll_ok, "Attempting 1GB memory pages for %s", name.c_str());
		extra_mmap_flags |= MAP_HUGE_1GB;
	}
	else
	{
		log(ll_ok, "Attempting 2MB memory pages for %s", name.c_str());
		extra_mmap_flags |= MAP_HUGE_2MB;
	}
#endif

	uint8_t *temp = (uint8_t *)mmap(addr, nBytesTruncated, PROT_READ | PROT_WRITE, mmap_flags | extra_mmap_flags, 0, 0);
	if (temp != (uint8_t *)-1)
		isMmap = true;
	else
	{
		temp = NULL;

		if (extra_mmap_flags)
		{
			log(ll_err, "Cannot allocate TPT space using MMAP non default settings for %s; trying default instead", name.c_str());

			temp = (uint8_t *)mmap(addr, nBytesTruncated, PROT_READ | PROT_WRITE, mmap_flags, 0, 0);
			if (temp != (uint8_t *)-1)
				isMmap = true;
			else
			{
				log(ll_err, "Mmap with default settings for %s failed: %s", name.c_str(), strerror(errno));
				temp = NULL;
			}
		}

		if (!temp)
		{
			log(ll_err, "Cannot allocate TPT space using MMAP for %s: %s (using malloc instead, linux users check /proc/sys/vm/nr_hugepages)", name.c_str(), strerror(errno));
			temp = (uint8_t *)malloc(nBytesTruncated);
		}
	}
#else
	uint8_t *temp = (uint8_t *)malloc(nBytesTruncated);
#endif

	if (!temp)
		error_exit(true, "Cannot allocate memory for %s", name.c_str());

	entries = (pttHashGroup *)temp;

	void *curAffinity = getCPUAffinity();

	// http://talkchess.com/forum/viewtopic.php?t=58830
	// memset the ptt in nThreads blocks while setting cpu affinity so that
	// each thread gets its own block of memory (think numa here)
	size_t bytesPerThread = nBytesTruncated / std::max(1, nThreads);
	for(int i=0; i<std::max(1, nThreads); i++)
	{
		setCPUAffinity(i);

		size_t offset = i * bytesPerThread;
		size_t curNBytes = (i == nThreads - 1) ? nBytesTruncated - offset : bytesPerThread;
		memset(&temp[offset], 0x00, curNBytes);
	}

	resetCPUAffinity(curAffinity);
}

Ptt::~Ptt()
{
#ifdef USE_MMAP
	if (isMmap)
		munmap(entries, nBytesTruncated);
	else
		free(entries);
#else
	free(entries);
#endif
}

bool Ptt::lookup(const uint64_t hash, pttEntry *const te)
{
	unsigned long int index = hash % nEntries;

	const pttEntry *const e = entries[index].entries;

	for(int i=0; i<N_TE_PER_HASH_GROUP; i++)
	{
		const pttEntry *const cur = &e[i];

#ifdef THREADING
		if ((cur -> hash ^ cur -> data_.data) == hash) { 
#else
		if (cur -> hash == hash) { 
#endif
			*te = *cur;
			return true;
		}
	}

	return false;
}

void Ptt::store(const uint64_t hash, const int d, const int value)
{
	unsigned long int index = hash % nEntries; // FIXME is biased at the bottom of the list

	pttEntry *const e = entries[index].entries;

	int useSubIndex = 1;

        for(int i=0; i<N_TE_PER_HASH_GROUP; i++)
        {
#ifdef THREADING
		if ((e[i].hash ^ e[i].data_.data) == hash) { 
#else
		if (e[i].hash == hash) { 
#endif
			useSubIndex = i;
                        break;
                }

                if (e[i].data_._data.age != age)
                        useSubIndex = i;
        }

#ifdef THREADING
	pttEntry pe;
	pe.data_._data.value = int16_t(value);
	pe.data_._data.age = age;

	pttEntry *const cur = &e[useSubIndex];
	cur -> hash = hash ^ pe.data_.data;
	cur -> data_.data = pe.data_.data;
#else
	pttEntry *const cur = &e[useSubIndex];
	cur -> hash = hash;
	cur -> data_._data.value = int16_t(value);
	cur -> data_._data.age = age;
#endif
}

int Ptt::hashFullPermill()
{
	int p = 0;

	for(int idx=0; idx<1000/N_TE_PER_HASH_GROUP; idx++) {
		for(int i=0; i<N_TE_PER_HASH_GROUP; i++) {
			if (entries[idx].entries[i].data_._data.age)
				p++;
		}
	}

	return p;
}
#endif
