// (C) Folkert van Heusden. AGPL v3.0.
// $Revision: 5116 $
#include "Tweaks.h"
#include <ctype.h>
#include <stdexcept>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "error.h"
#include "Debug.h"
#include "Exception.h"
#include "Utils.h"
#include "Scene.h"
#include "ChessPiece.h"
#include "Move.h"
#include "Board.h"
#include "IO_Console.h"

char REV_SCENE[] = "Scene $Revision: 5116 $";

constexpr uint32_t idmask[] = { 0xffff, 0xffff0000 };
constexpr uint32_t idmaskpawn[] = { 0xff, 0xff0000 };

void Scene::genBoard()
{
	b.clearBoard();

	int nr = 0;
	for(int c=0; c<2; c++) {
		for(int idx=0; idx<16; idx++)
		{
			if (pieces[c][idx].getAlive())
				b.put(&pieces[c][idx]);

			pieces[c][idx].setId(nr++);
		}
	}
}

void Scene::updateBoard()
{
	genBoard();
}

void Scene::setupPieces()
{
	int white_idx = 0, black_idx = 0;

	// 132465
	for(int loop=0; loop<8; loop++)
	{
		pieces[WHITE][white_idx++].initialize(WHITE, loop, 1, true, PAWN);
		pieces[BLACK][black_idx++].initialize(BLACK, loop, 6, true, PAWN);
	}

	pieces[WHITE][white_idx++].initialize(WHITE, 1, 0, true, KNIGHT);
	pieces[BLACK][black_idx++].initialize(BLACK, 1, 7, true, KNIGHT);
	pieces[WHITE][white_idx++].initialize(WHITE, 6, 0, true, KNIGHT);
	pieces[BLACK][black_idx++].initialize(BLACK, 6, 7, true, KNIGHT);

	pieces[WHITE][white_idx++].initialize(WHITE, 0, 0, true, ROOK);
	pieces[BLACK][black_idx++].initialize(BLACK, 0, 7, true, ROOK);
	pieces[WHITE][white_idx++].initialize(WHITE, 7, 0, true, ROOK);
	pieces[BLACK][black_idx++].initialize(BLACK, 7, 7, true, ROOK);

	pieces[WHITE][white_idx++].initialize(WHITE, 2, 0, true, BISHOP);
	pieces[BLACK][black_idx++].initialize(BLACK, 2, 7, true, BISHOP);
	pieces[WHITE][white_idx++].initialize(WHITE, 5, 0, true, BISHOP);
	pieces[BLACK][black_idx++].initialize(BLACK, 5, 7, true, BISHOP);

	pieces[WHITE][white_idx++].initialize(WHITE, 4, 0, true, KING);
	pieces[BLACK][black_idx++].initialize(BLACK, 4, 7, true, KING);

	pieces[WHITE][white_idx++].initialize(WHITE, 3, 0, true, QUEEN);
	pieces[BLACK][black_idx++].initialize(BLACK, 3, 7, true, QUEEN);

	my_assert(white_idx == 16, "Scene::setupPieces: not 16 pieces for white");
	my_assert(black_idx == 16, "Scene::setupPieces: not 16 pieces for black");
}

void Scene::findKings()
{
	for(int loop=0; loop<16; loop++)
	{
		for(int c=0; c<2; c++) {
			if (pieces[c][loop].getType() == KING)
				kings[c] = &pieces[c][loop];
		}
	}

	my_assert(kings[WHITE] != NULL, "Scene::findKings: white king missing");
	my_assert(kings[WHITE] -> getAlive(), "Scene::findKings: white king dead");
	my_assert(kings[BLACK] != NULL, "Scene::findKings: black king missing");
	my_assert(kings[BLACK] -> getAlive(), "Scene::findKings: black king dead");
}

void Scene::allocateMoveListPair()
{
	moveLists.push_back(new MoveLists);
}

void Scene::allocateMoveLists()
{
	allocateMoveListPair();

	mlStateDepth = 0;
}

void Scene::initializeMembers()
{
	SceneState state;
	memset(&state, 0x00, sizeof state);
	state.lastMove.lastOpponentMove = invalidMove;

	ss.push_back(state);
	css = &ss.back();

	allocateMoveLists();
}

void Scene::countMaterial()
{
	memset(css -> materialCounts, 0x00, sizeof css -> materialCounts);

	countMaterial(getPieces(WHITE), css -> materialCounts[WHITE]);
	countMaterial(getPieces(BLACK), css -> materialCounts[BLACK]);
}

Scene::Scene()
{
	initializeMembers();

	setupPieces();

	genBoard();

	findKings();

	countMaterial();
}

Scene::Scene(const ChessPiece *const restrict wIn, const ChessPiece * const restrict bIn)
{
	initializeMembers();

	for(int loop=0; loop<16; loop++)
	{
		pieces[WHITE][loop].initialize(wIn[loop]);
		pieces[BLACK][loop].initialize(bIn[loop]);
	}

	genBoard();

	findKings();

	countMaterial();
}

Scene::Scene(const Scene & sIn)
{
	initializeMembers();

	css -> rule50 = sIn.getNHalfMoves();
	css -> nFullMoves = sIn.getNFullMoves();

	const ChessPiece *const wIn = sIn.getPieces(WHITE);
	const ChessPiece *const bIn = sIn.getPieces(BLACK);

	for(int loop=0; loop<16; loop++)
	{
		pieces[WHITE][loop].initialize(wIn[loop]);
		pieces[BLACK][loop].initialize(bIn[loop]);
	}

	genBoard();

	findKings();

	countMaterial();
}

void Scene::ConvertFEN(const char * const restrict fen, const char * const restrict castlingAvailability, const char *const restrict halfMoveCounter, const char *const restrict fullMoveCounter)
{
	css -> rule50 = atoi(halfMoveCounter);
	css -> nFullMoves = atoi(fullMoveCounter);

	for(int i=0; i<16; i++) {
		pieces[0][i].setAlive(false);
		pieces[1][i].setAlive(false);
	}

	int np[2] = { 0, 0 };

	int y = 7;
	int x = 0;

	unsigned long int len = strlen(fen);
	for(unsigned long int index=0; index<len; index++) {
		int currentChar = fen[index];
		ChessPieceType objectType = NONE;
		PlayerColor objectColor = WHITE;
		int currentX = x;
		int currentY = y;

		if (islower(currentChar))
			objectColor = BLACK;

		currentChar = toupper(currentChar);
		if (currentChar == 'P') {
			objectType = PAWN;
			x++;
		}
		else if (currentChar == 'N') {
			objectType = KNIGHT;
			x++;
		}
		else if (currentChar == 'B') {
			objectType = BISHOP;
			x++;
		}
		else if (currentChar == 'R') {
			objectType = ROOK;
			x++;
		}
		else if (currentChar == 'Q') {
			objectType = QUEEN;
			x++;
		}
		else if (currentChar == 'K') {
			objectType = KING;
			x++;
		}
		else if (isdigit(currentChar)) {
			x += currentChar - '0';
		}
		else if (currentChar == '/') {
			x = 0;
			y--;
		}
		else {
			my_assert(false, myformat("ConvertFEN: invalid character %c (%d)", currentChar, currentChar).c_str());
		}

		if (objectType != NONE) {
			my_assert(objectColor == WHITE || objectColor == BLACK, "Scene::ConvertFEN check color of object");

			bool firstMove = false;

			if (objectColor == WHITE && y == 1 && objectType == PAWN)
				firstMove = true;
			else if (objectColor == BLACK && y == 6 && objectType == PAWN)
				firstMove = true;

			int i = 0;
			for(; i<16; i++) {
				if (pieces[objectColor][i].getType() == objectType && pieces[objectColor][i].getAlive() == false) {
					pieces[objectColor][i].setAlive(true);
					pieces[objectColor][i].setXY(currentX, currentY);
					pieces[objectColor][i].setFirstMove(firstMove);
					break;
				}
			}

			if (i == 16) { // promotion?
				for(i = 0; i<16; i++) {
					if (pieces[objectColor][i].getType() == PAWN && pieces[objectColor][i].getAlive() == false) {
						pieces[objectColor][i].setType(objectType);
						pieces[objectColor][i].setAlive(true);
						pieces[objectColor][i].setXY(currentX, currentY);
						pieces[objectColor][i].setFirstMove(firstMove);
						break;
					}
				}
			}

			if (i == 16) {
				error_exit(false, "FEN init issue");
			}
		}
	}

	genBoard();

	unsigned long int caLen = strlen(castlingAvailability);
	for(unsigned long int index=0; index<caLen; index++) {
		int currentChar = castlingAvailability[index];

		if (currentChar == '-')
			break;

		if (currentChar == 'K') { // white king side => right
			y = 0;
			x = 7;
		}
		else if (currentChar == 'Q') { // white queen side => left
			y = 0;
			x = 0;
		}
		else if (currentChar == 'k') { // black king side => right
			y = 7;
			x = 7;
		}
		else if (currentChar == 'q') { // black queen side => left
			y = 7;
			x = 0;
		}
		else {
			my_assert(false, "ConvertFEN: invalid castling-availability string");
		}

		if (b.getAt(x, y))
			b.getAt(x, y) -> setFirstMove();
		// if one can still castle with a rook, the king should also be able to castle
		if (b.getAt(4, y))
			b.getAt(4, y) -> setFirstMove();
	}
}

Scene::Scene(const std::string & fenIn)
{
	initializeMembers();

	setupPieces();

	std::vector<std::string> *const parts = split(fenIn, " ");
	std::string board = parts -> at(0);
	std::string color = parts -> at(1);
	std::string castling = parts -> at(2);
	std::string ep = parts -> at(3);

	SceneState state;
	memset(&state, 0x00, sizeof state);
	ss.push_back(state);
	css = &ss.back();

	ConvertFEN(board.c_str(), castling.c_str(), parts -> size() >= 5 ? parts -> at(4).c_str() : "1", parts -> size() >= 6 ? parts -> at(5).c_str() : "1");

	if (ep != "-")
	{
		int dx = ep.at(0) - 'a';
		int dy = ep.at(1) - '1';

		Move *const epMove = &css -> lastMove.lastOpponentMove;
		epMove -> u.m.fx = epMove -> u.m.tx = dx;
		epMove -> u.m.fy = dy == 2 ? 1 : 6;
		epMove -> u.m.ty = dy == 2 ? 3 : 4;
		epMove -> u.m.mt = REGULAR;
		epMove -> u.m.promoteTo = NONE;
		epMove -> u.m.isCatchMove = true;

		css -> lastMove.lastOpponentColor = color == "w" ? BLACK : WHITE;
	}

	delete parts;

	findKings();

	countMaterial();
}

Scene::~Scene()
{
	while(ss.size())
		popState();

	while(!moveLists.empty())
	{
		delete moveLists.at(0);
		moveLists.erase(moveLists.begin() + 0);
	}
}

bool Scene::compareStateTo(const SceneState *const other)
{
	if (css -> lastMove.lastOpponentColor != other -> lastMove.lastOpponentColor)
		return false;

	if (isValidMove(css -> lastMove.lastOpponentMove) != isValidMove(other -> lastMove.lastOpponentMove))
		return false;

	if (moveCompareTo(css -> lastMove.lastOpponentMove, other -> lastMove.lastOpponentMove) == false)
		return false;

	if (css -> isKingPinnedBitmap != other -> isKingPinnedBitmap)
		return false;

	if (css -> rule50 != other -> rule50)
		return false;

	if (css -> nFullMoves != other -> nFullMoves)
		return false;

	return true;
}

bool Scene::compareTo(Scene & other)
{
#if 0
	if (b.compareTo(*other.getBoardConst()) == false)
		return false;

	ChessPiece * const *otherWhite = other.getPieces(WHITE);
	ChessPiece * const *otherBlack = other.getPieces(BLACK);

	for(int idx=0; idx<16; idx++)
	{
		if (pieces[WHITE][idx].compareTo(*otherWhite[idx]) == false)
			return false;

		if (pieces[BLACK][idx].compareTo(*otherBlack[idx]) == false)
			return false;
	}

// FIXME	if (compareStateTo(&other.state) == false)
//		return false;

	return true;
#endif
	return false;
}

void Scene::pushState() 
{ 
	ss.push_back(ss.back());
	css = &ss.back();

	mlStateDepth++;

	try {
		moveLists.at(mlStateDepth) -> lists[WHITE].clear();
		moveLists.at(mlStateDepth) -> lists[BLACK].clear();
	}
	catch(std::out_of_range & o) {
		// if 2 or more are missing: that would be a bug
		allocateMoveListPair();
	}

	css -> lastMove.lastOpponentMove = invalidMove;

	css -> isKingPinnedBitmap = 0;

	css -> ccheckedField[0] = css -> ccheckedField[1] = false;

	for(int y=0; y<8; y++) {
		for(int x=0; x<8; x++)
			css -> tcs[y * 8 + x] = 0;
	}

//	memset(css -> tcs, 0x00, sizeof(css -> tcs));
} 

void Scene::popState() 
{ 
	my_assert(ss.size() > 0, "Scene::popState: state stack underflow"); 

	//my_assert(mlStateDepth > 0, "Scene::popState: moveLists stack underflow");
	mlStateDepth--;

#ifdef _DEBUG
	css -> lastMove.lastOpponentColor = WHITE;
	css -> lastMove.lastOpponentMove.u.m.fx = css -> lastMove.lastOpponentMove.u.m.fy = css -> lastMove.lastOpponentMove.u.m.tx = css -> lastMove.lastOpponentMove.u.m.ty = 1;
	css -> lastMove.lastOpponentMove.u.m.mt = EN_PASSANT;
	css -> lastMove.lastOpponentMove.u.m.promoteTo = NONE;
#endif

	ss.pop_back(); 
	css = &ss.back();
} 

bool Scene::getIsProtected(const ChessPiece *const restrict cp) const
{
	PlayerColor c = cp -> getColor();

	int x, y;
	cp -> getXY(&x, &y);

	return (css -> tcs[y * 8 + x] & idmask[c]) != 0;
}

bool Scene::getIsUnderAttack(const ChessPiece *const restrict cp) const
{
	PlayerColor c = cp -> getColor();

	int x, y;
	cp -> getXY(&x, &y);

	return (css -> tcs[y * 8 + x] & idmask[opponentColor(c)]) != 0;
}

bool Scene::getIsUnderAttackByPawn(const ChessPiece *const restrict cp) const
{
	PlayerColor c = cp -> getColor();

	int x, y;
	cp -> getXY(&x, &y);

	return (css -> tcs[y * 8 + x] & idmaskpawn[opponentColor(c)]) != 0;
}

int Scene::countUnderAttack(const PlayerColor c) const
{
	return 0; // FIXME
}

int Scene::countProtected(const PlayerColor c) const
{
	return 0; // FIXME
}

void Scene::doMove(const Move & m, BoardState *const cbs)
{
	pushState();

	ChessPiece *const lmo = b.getAt(m.u.m.fx, m.u.m.fy);
	const bool isPawn = lmo -> getType() == PAWN;
	if (isPawn || m.u.m.isCatchMove)
		css -> rule50 = 0;
	else
		css -> rule50++;

	const PlayerColor c = lmo -> getColor();
	if (c == BLACK)
		css -> nFullMoves++;

	bool checkEndGame = false;
	if (m.u.m.isCatchMove) {
		const ChessPiece *const victim = b.findVictim(m);

		PlayerColor vc = victim -> getColor();
		ChessPieceType vt = victim -> getType();

		css -> materialCounts[vc][vt]--;

		if (vt == QUEEN) {
			if (vc == WHITE)
				css -> wQMoved = true;
			else
				css -> bQMoved = true;
		}

		checkEndGame = true;
	}

	if (isPawn) {
		if (abs(m.u.m.ty - m.u.m.fy) == 2)
		{
			css -> lastMove.lastOpponentMove = m;
			css -> lastMove.lastOpponentColor = c;
		}
		else if (m.u.m.promoteTo != NONE) {
			css -> materialCounts[c][PAWN]--;
			css -> materialCounts[c][m.u.m.promoteTo]++;
			checkEndGame = true;
		}
	}

	// middle game: both queens have moved
	if (lmo -> getType() == QUEEN) {
		if (c == WHITE)
			css -> wQMoved = true;
		else
			css -> bQMoved = true;
	}

	// end game
	if (checkEndGame) {
		// FIXME tuning
		css -> isEndGame = (css -> materialCounts[WHITE][PAWN] + css -> materialCounts[WHITE][KNIGHT] * 3 + css -> materialCounts[WHITE][BISHOP] * 3 + css -> materialCounts[WHITE][ROOK] * 5 + css -> materialCounts[WHITE][QUEEN] * 9 + css -> materialCounts[BLACK][PAWN] + css -> materialCounts[BLACK][KNIGHT] * 3 + css -> materialCounts[BLACK][BISHOP] * 3 + css -> materialCounts[BLACK][ROOK] * 5 + css -> materialCounts[BLACK][QUEEN] * 9) <= 12;
	}

	b.doMove(m, cbs);

#ifdef _DEBUG
{
	int mc[2][6];
	memcpy(mc, css -> materialCounts, sizeof css -> materialCounts);
	memset(css -> materialCounts, 0x00, sizeof css -> materialCounts);
	countMaterial(getPieces(WHITE), css -> materialCounts[WHITE]);
	countMaterial(getPieces(BLACK), css -> materialCounts[BLACK]);
	for(int C=0; C<2; C++) {
		for(int t=0; t<6; t++) {
			if (mc[C][t] != css -> materialCounts[C][t]) {
				printf("%d %d\n", mc[C][t], css -> materialCounts[C][t]);
				printf("%s\n", moveToString(css -> lastMove.lastOpponentMove).c_str());
				printf("%s\n", genFen(this, false, c).c_str());
			}
		}
	}
	my_assert(memcmp(mc, css -> materialCounts, sizeof css -> materialCounts) == 0, "material counts");

	my_assert(css -> materialCounts[WHITE][PAWN] >= 0 && css -> materialCounts[WHITE][PAWN] <= 8, "material counts");
	int extra = 8 - css -> materialCounts[WHITE][PAWN];
	extra -= std::max(css -> materialCounts[WHITE][ROOK] - 2, 0);
	extra -= std::max(css -> materialCounts[WHITE][BISHOP] - 2, 0);
	extra -= std::max(css -> materialCounts[WHITE][KNIGHT] - 2, 0);
	extra -= std::max(css -> materialCounts[WHITE][QUEEN] - 2, 0);
	my_assert(extra >= 0, "material counts");
	my_assert(css -> materialCounts[WHITE][KING] == 1, "material counts");
}
#endif
}

void Scene::undoMove()
{
	b.undoMove();

	popState();
}

PlayerColor Scene::getLastOpponentMoveColor() const
{
	return css -> lastMove.lastOpponentColor;
}

const Move * Scene::getLastOpponentMove() const
{
	if (isValidMove(css -> lastMove.lastOpponentMove))
		return &css -> lastMove.lastOpponentMove;

	return NULL;
}

int count1(const uint32_t bits)
{
	return __builtin_popcount(bits);
}

void Scene::generateMovelists()
{
	my_assert(kings[0] -> getAlive(), "Scene::generateMovelists: white king must be alive");
	my_assert(kings[1] -> getAlive(), "Scene::generateMovelists: black king must be alive");

	css -> isKingPinnedBitmap = 0;

	gml_pars_t pars;
	pars.b = &b;
	pars.isKingPinned = &css -> isKingPinnedBitmap;
	pars.tcs = css -> tcs;

	for(int c=0; c<2; c++) {
		moveLists.at(mlStateDepth) -> lists[c].clear();

		pars.ml = &moveLists.at(mlStateDepth) -> lists[c];
		pars.lastOpponentMove = css -> lastMove.lastOpponentColor == PlayerColor(1 - c) ? css -> lastMove.lastOpponentMove : invalidMove;
		pars.mking = kings[c];
		pars.oking = kings[opponentColor(c)];

		for(int idx=0; idx<16; idx++)
		{
#ifdef _DEBUG
			pieces[c][idx].prepareForGenerateMoveList();
#endif

			if (pieces[c][idx].getAlive())
				pieces[c][idx].generateMoveList(&pars);
		}
	}

	my_assert(count1(css -> isKingPinnedBitmap) <= 24, "Scene::generateMovelists");
}

VectorType<Move> *Scene::getMoveList(PlayerColor c) const
{
	return &moveLists.at(mlStateDepth) -> lists[c];
}

ChessPiece *Scene::getKing(const PlayerColor c) const
{
	return kings[c];
}

bool Scene::isKingUnderAttack(const PlayerColor kingColor) const
{
	return getIsUnderAttack(getKing(kingColor));
}

// c == under attack by whom
bool Scene::isFieldUnderAttack(const PlayerColor c, const int x, const int y) const
{
	return !!(css -> tcs[y * 8 + x] & idmask[c]);
}

bool Scene::isOpponentPawnAt(const int x, const int y, const PlayerColor hisColor) const
{
	const ChessPiece *const currentObject = b.getAt(x, y);

	return currentObject != NULL && currentObject -> getType() == PAWN && currentObject -> getColor() == hisColor;
}

bool Scene::validateMoves(const PlayerColor color, const size_t moveIndex)
{
	my_assert(kings[0] -> getAlive(), "Scene::validateMoves: white king must be alive");
	my_assert(kings[1] -> getAlive(), "Scene::validateMoves: black king must be alive");

	/* get the reference to the list of moves for the color selected */
	VectorType<Move> *const checkMoves = getMoveList(color);

	bool kingUnderAttack = isKingUnderAttack(color);

	const ChessPiece *const king = getKing(color);

	PlayerColor oc = opponentColor(color);
	const ChessPiece *const hisPieces = getPieces(oc);

	BoardState bs;

	/* check every move 'color' can do */
	const Move currentMove = checkMoves -> at(moveIndex);
	//printf("%s: KUA %d %d,%d | %d\n", moveToString(currentMove).c_str(), kingUnderAttack, king -> getX(), king -> getY(), color);
	//dumpCell(&css -> tcs, king -> getX(), king -> getY(), opponentColor(color));

	const ChessPiece * const moveObject = b.getAt(currentMove.u.m.fx, currentMove.u.m.fy);

	bool isKingPinned = moveObject -> getIsKingPinned(&css -> isKingPinnedBitmap);

	bool needThoroughValidityCheck =
		moveObject == king ||
		isKingPinned ||
		currentMove.u.m.mt == EN_PASSANT ||
		kingUnderAttack;
	// needThoroughValidityCheck = true;

	// slows down 10% in start position
	// if (moveObject -> getType() == BISHOP && isKingPinned) {
	// 	remove = true;
	// 	goto skip_check;
	// }

	/* calculate moves of opponent AFTER we did the move we're checking */
	bool remove = false;

	if (unlikely(needThoroughValidityCheck)) {
		uint32_t bits = (css -> tcs[king -> getY() * 8 + king -> getX()] & idmask[oc]);
		if (oc == BLACK)
			bits >>= 16;

		doMove(currentMove, &bs);

		// do only a search for opponent (oc); can he capture
		// my king
		const isUnderAttackBy_pars_t ibp = { &b, king -> getX(), king -> getY() };
#ifdef _DEBUG
		int n = 0;
		int nt[6] = { 0, 0, 0, 0, 0, 0 };
#endif

		for(int i=0; i<16; i++) {
			if ((bits & (1 << i)) && hisPieces[i].getAlive() && hisPieces[i].isUnderAttackByThis(&ibp)) {
				remove = true;
				undoMove();
				goto r;
			}
		}

		for(int i=0; i<16; i++) {
			if (hisPieces[i].getAlive() && hisPieces[i].isUnderAttackByThis(&ibp)) {
				//printf("removed by uat\n");
				remove = true;
				//printf("%s by %c%c\n", moveToString(currentMove).c_str(), hisPieces[i] -> getX() + 'a', hisPieces[i] -> getY() + '1');
#ifdef _DEBUG
				n++;
				nt[hisPieces[i].getType()]++;
#else
				break;
#endif
			}
		}

#ifdef _DEBUG
		if (remove) {
			my_assert(n > 0, "iuabt");
			my_assert(nt[PAWN] >= 0 && nt[PAWN] <= 2, "iuabt");
			my_assert(nt[KING] >= 0 && nt[KING] <= 1, "iuabt");
			my_assert(nt[KNIGHT] >= 0 && nt[KNIGHT] <= 8, "iuabt");
			my_assert(nt[QUEEN] >= 0 && nt[QUEEN] <= 8, "iuabt");
			my_assert(nt[BISHOP] >= 0 && nt[BISHOP] <= 4, "iuabt");
			my_assert(nt[ROOK] >= 0 && nt[ROOK] <= 4, "iuabt");


			{
				generateMovelists();

				my_assert(isKingUnderAttack(color), "under attack");
			}

		}
#endif

		undoMove();
	}

	if (unlikely(currentMove.u.m.mt == CASTLING && remove == false))
	{
		my_assert((currentMove.u.m.fy == 0 && color == WHITE) || (currentMove.u.m.fy == 7 && color == BLACK), "Scene::validateMoves: castling validity (1)");
		my_assert(currentMove.u.m.fy == currentMove.u.m.ty, "Scene::validateMoves: castling validity (2)");

		if (kingUnderAttack) {
			//printf("removed by kua %s (%d,%d)\n", moveToString(currentMove).c_str(), getKing(color) -> getX(), getKing(color) -> getY());
			remove = true;
		}
		else
		{  
			//printf("removed by fua %s\n", moveToString(currentMove).c_str());
			// LONG castling = 2
			int checkField = currentMove.u.m.tx == 2 ? 3 : 5;
			int pawnLine = color == WHITE ? 1 : 6;

			int i = checkField == 3;
			if (!css -> ccheckedField[i]) {
				css -> ccheckedField[i] = true;
				css -> cfieldUnderAttack[i] = isFieldUnderAttack(oc, checkField, currentMove.u.m.fy);
			}

			remove = css -> cfieldUnderAttack[i] || isOpponentPawnAt(checkField - 1, pawnLine, oc) || isOpponentPawnAt(checkField + 1, pawnLine, oc);
		}
	}

r:
	return !remove;
}

void Scene::partialValidateMoves(const PlayerColor color)
{
	VectorType<Move> *list = getMoveList(color);

	for(size_t i=0; i<list -> size();) {
		if (validateMoves(color, i))
			break;

		list -> erase(i);
	}
}

Board * Scene::getBoard()
{
	return &b;
}

const ChessPiece * Scene::getPieces(const PlayerColor c) const
{
	return pieces[c];
}

bool Scene::isCheckMate(const PlayerColor side) const
{
	return getMoveList(side) -> empty() && isKingUnderAttack(side) == true;
}

bool Scene::isStaleMate(const PlayerColor side) const
{
	return getMoveList(side) -> empty() && isKingUnderAttack(side) == false;
}

void Scene::countMaterial(const ChessPiece * const restrict objs, int *const restrict counts)
{
	for(int i=0; i<16; i++) {
		if (objs[i].getAlive())
			counts[objs[i].getType()]++;
	}
}

bool Scene::isInsufficientMaterialDraw(int A[], int B[])
{
	// K+B v K, K+N v K, K v K

	if (A[5] || B[5] || // pawn
			A[2] || B[2] || // rook
			A[1] || B[1]) // queen
		return false;

	if (A[3] + B[3] +       // knight
			A[4] + B[4] > 1) // bishop
		return false;

	return true;
}

bool Scene::isInsufficientMaterialDraw()
{
	return isInsufficientMaterialDraw(css -> materialCounts[WHITE], css -> materialCounts[BLACK]);
}

bool Scene::isEndGame() const
{
	return css -> isEndGame;
}

bool Scene::isMiddleGame() const
{
	return css -> wQMoved && css -> bQMoved;
}

void Scene::getCastlingRights(bool *const restrict wl, bool *const restrict ws, bool *const restrict bl, bool *const restrict bs) const
{
	*ws = *wl = *bs = *bl = false;

	if (b.getAt(4, 0) && b.getAt(4, 0) -> getFirstMove())
	{
		if (b.getAt(7, 0) && b.getAt(7, 0) -> getFirstMove())
			*ws = true;

		if (b.getAt(0, 0) && b.getAt(0, 0) -> getFirstMove())
			*wl = true;
	}

	if (b.getAt(4, 7) && b.getAt(4, 7) -> getFirstMove())
	{
		if (b.getAt(7, 7) && b.getAt(7, 7) -> getFirstMove())
			*bs = true;

		if (b.getAt(0, 7) && b.getAt(0, 7) -> getFirstMove())
			*bl = true;
	}
}


void Scene::getKingPinnedCount(int *const whitePinnedN, int *const blackPinnedN) const
{
	*whitePinnedN = count1(css -> isKingPinnedBitmap & 0xffff);
	*blackPinnedN = count1(css -> isKingPinnedBitmap & 0xffff0000);
}

bool Scene::getSEE(const int x, const int y, const ChessPiece *const attacker) const
{
	const ChessPiece *const victim = b.getAt(x, y);

	const PlayerColor c = attacker -> getColor();

	// bit 0...15: if set, then a white piece can walk over and/or attack this field, 16...31 black
	const uint32_t bits = css -> tcs[y * 8 + x];

	int see_val = victim -> getEvalVal();
	int trophy_val = attacker -> getEvalVal();

	PlayerColor sc = opponentColor(c);
	int so[] = { 0, 16 }; // where to start to search in the bit pattern. so[0] is for white
	constexpr int soe[] = { 16, 32 }; // where to stop searching. soe[0] is for white

	for(; so[sc] < soe[sc];) {
		// find a 1 bit
		while((bits & (1 << so[sc])) == 0) {
			// if end of pattern, stop searching
			if (++so[sc] == soe[sc])
				break;
		}

		if (so[sc] >= soe[sc])
			break;

		if (sc == c)
			see_val += trophy_val;
		else
			see_val -= trophy_val;


		// get eval val. bit number is an index in pieces[color][index] pointing to the
		// piece.
		trophy_val = sc == WHITE ? pieces[sc][so[sc]].getEvalVal() : pieces[sc][so[sc] - 16].getEvalVal();

		if (sc == c) {
			if (see_val - trophy_val >= 0)
				return true;

			if (see_val < 0)
				return false;
		}
		else {
			if (see_val >= 0)
				return true;

			if (see_val + trophy_val < 0)
				return false;
		}

		// on to the next bit!
		so[sc]++;

		sc = opponentColor(sc);
	}

	return see_val >= 0;
}

void Scene::setNullMove(Move *const nmm, int *const hm)
{
	*nmm = css -> lastMove.lastOpponentMove;
	css -> lastMove.lastOpponentMove = invalidMove;
	css -> rule50 = 0;
}

void Scene::unsetNullMove(const Move & org, const int hm)
{
	css -> lastMove.lastOpponentMove = org;
	css -> rule50 = hm;
}

int Scene::gamePhase()
{
	const int num_knights = css -> materialCounts[WHITE][KNIGHT] + css -> materialCounts[BLACK][KNIGHT];
	const int num_bishops = css -> materialCounts[WHITE][BISHOP] + css -> materialCounts[BLACK][BISHOP];
	const int num_rooks   = css -> materialCounts[WHITE][ROOK]   + css -> materialCounts[BLACK][ROOK];
	const int num_queens  = css -> materialCounts[WHITE][QUEEN]  + css -> materialCounts[BLACK][QUEEN];

	// from https://www.chessprogramming.org/Tapered_Eval
	const int knight_phase = 1;
	const int bishop_phase = 1;
	const int rook_phase   = 2;
	const int queen_phase  = 4;

	const int total_phase = knight_phase * 4 + bishop_phase * 4 + rook_phase * 4 + queen_phase * 2;

	int phase = total_phase;

	phase -= knight_phase * num_knights;
	phase -= bishop_phase * num_bishops;
	phase -= rook_phase   * num_rooks;
	phase -= queen_phase  * num_queens;

	return (phase * 256 + (total_phase / 2)) / total_phase;
}
