// (C) Folkert van Heusden. Not to be distributed nor open-source.
// $Revision: 4915 $
#include "Tweaks.h"
#include "config.h"
#include <cstdint>
#include <stdlib.h>
#include <string.h>
#ifdef USE_MMAP
#include <sys/mman.h>
#endif

#include "Exception.h"
#include "Tpt.h"
#include "Utils.h"
#include "IO.h"
#include "error.h"

char REV_TPT[] = "Tpt $Revision: 4915 $";

bool getUint64_t(FILE *const fh, uint64_t *const dummy)
{
	return fread(dummy, sizeof(uint64_t), 1, fh) == 1;
}

bool putUint64_t(FILE *const fh, const uint64_t v)
{
	return fwrite(&v, sizeof(uint64_t), 1, fh) == 1;
}

// use polyglot randoms from BookPolyglot.cpp
#define GB (1024 * 1024 * 1024)

Tpt::Tpt(const std::string & nameIn, IO *const ioIn, const uint64_t nBytes, const int nThreads) : name(nameIn), io(ioIn), age(0)
{
	entries = NULL;

	setSize(nBytes, nThreads);
}

Tpt::~Tpt()
{
#ifdef USE_MMAP
	if (isMmap)
		munmap(entries, nBytesTruncated);
	else
#endif
		free(entries);
}

void Tpt::setSize(const uint64_t nBytes, const int nThreads)
{
	if (entries) {
#ifdef USE_MMAP
		if (isMmap)
			munmap(entries, nBytesTruncated);
		else
#endif
			free(entries);
	}

	nEntries = nBytes / sizeof(tptHashGroup);
	nBytesTruncated = nEntries * sizeof(tptHashGroup);

	isMmap = false;
#ifdef USE_MMAP
#ifdef __ia64__
	void *addr = 0x8000000000000000UL;
#else
	void *addr = NULL;
#endif
	// http://stackoverflow.com/questions/2688466/why-mallocmemset-is-slower-than-calloc says the zeroing only takes place when a page is first written to
	int mmap_flags = MAP_PRIVATE | MAP_ANONYMOUS;
	int extra_mmap_flags = 0;

#ifdef MAP_HUGE_SHIFT
	mmap_flags |= MAP_HUGETLB;

#if !defined(MAP_HUGE_1GB) || !defined(MAP_HUGE_2MB)
	#define MAP_HUGE_2MB    (21 << MAP_HUGE_SHIFT)
	#define MAP_HUGE_1GB    (30 << MAP_HUGE_SHIFT)
#endif
	log(ll_ok, "%s size in bytes: %ld", name.c_str(), nBytesTruncated);
	int remainder = nBytesTruncated % GB;
	if (remainder >= MMAP_1GB_THRESHOLD || remainder == 0)
	{
		log(ll_ok, "Attempting 1GB memory pages for %s", name.c_str());
		extra_mmap_flags |= MAP_HUGE_1GB;
	}
	else
	{
		log(ll_ok, "Attempting 2MB memory pages for %s", name.c_str());
		extra_mmap_flags |= MAP_HUGE_2MB;
	}
#endif

	uint8_t *temp = (uint8_t *)mmap(addr, nBytesTruncated, PROT_READ | PROT_WRITE, mmap_flags | extra_mmap_flags, 0, 0);
	if (temp != (uint8_t *)-1)
		isMmap = true;
	else
	{
		temp = NULL;

		if (extra_mmap_flags)
		{
			log(ll_err, "Cannot allocate TPT space using MMAP non default settings for %s; trying default instead", name.c_str());

			temp = (uint8_t *)mmap(addr, nBytesTruncated, PROT_READ | PROT_WRITE, mmap_flags, 0, 0);
			if (temp != (uint8_t *)-1)
				isMmap = true;
			else
			{
				log(ll_err, "Mmap with default settings for %s failed: %s", name.c_str(), strerror(errno));
				temp = NULL;
			}
		}

		if (!temp)
		{
			log(ll_err, "Cannot allocate TPT space using MMAP for %s: %s (using malloc instead, linux users check /proc/sys/vm/nr_hugepages)", name.c_str(), strerror(errno));
			temp = (uint8_t *)malloc(nBytesTruncated);
		}
	}
#else
	uint8_t *temp = (uint8_t *)malloc(nBytesTruncated);
#endif

	if (!temp)
		error_exit(true, "Cannot allocate memory for %s", name.c_str());

	entries = (tptHashGroup *)temp;

	void *curAffinity = getCPUAffinity();

	// http://talkchess.com/forum/viewtopic.php?t=58830
	// memset the tpt in nThreads blocks while setting cpu affinity so that
	// each thread gets its own block of memory (think numa here)
	size_t bytesPerThread = nBytesTruncated / std::max(1, nThreads);
	for(int i=0; i<std::max(1, nThreads); i++)
	{
		setCPUAffinity(i);

		size_t offset = i * bytesPerThread;
		size_t curNBytes = (i == nThreads - 1) ? nBytesTruncated - offset : bytesPerThread;
		memset(&temp[offset], 0x00, curNBytes);
	}

	resetCPUAffinity(curAffinity);
}

void Tpt::registerIO(IO *const ioIn)
{
	io = ioIn;
}

uint64_t Tpt::findCastlingRightsHash(const Board *const b)
{
	uint64_t hash = 0;

        if (b -> getAt(4, 0) && b -> getAt(4, 0) -> getFirstMove()) 
        { 
                if (b -> getAt(0, 0) && b -> getAt(0, 0) -> getFirstMove()) 
                        hash ^= Random64[768]; 

                if (b -> getAt(7, 0) && b -> getAt(7, 0) -> getFirstMove()) 
                        hash ^= Random64[768 + 1]; 
        } 

        if (b -> getAt(4, 7) && b -> getAt(4, 7) -> getFirstMove()) 
        { 
                if (b -> getAt(0, 7) && b -> getAt(0, 7) -> getFirstMove()) 
                        hash ^= Random64[768 + 2]; 

                if (b -> getAt(7, 7) && b -> getAt(7, 7) -> getFirstMove()) 
                        hash ^= Random64[768 + 3]; 
        } 

	return hash;
}

zobrist_t Tpt::updateZobristNullMove(const zobrist_t & in, const Board *const b, const PlayerColor c)
{
	zobrist_t nh = { in.newHash ^ in.cleanup, 0x00, in.pawnNewHash };

	if (c == WHITE)
	{
		const uint64_t apply = Random64[780];
		nh.newHash ^= apply;
		nh.cleanup ^= apply;
	}

	uint64_t castlingHash = findCastlingRightsHash(b);
	nh.newHash ^= castlingHash;
	nh.cleanup ^= castlingHash;

	return nh;
}

zobrist_t Tpt::updateZobrist(const zobrist_t & in, const BoardState & bs, const Board *const b, const PlayerColor c, const Move epMove)
{
	zobrist_t nh = { in.newHash ^ in.cleanup, 0x00, in.pawnNewHash };

	if (unlikely(bs.mt == CASTLING)) {
		int pos1_1 = -1, pos2_1 = -1, pos1_2 = -1, pos2_2 = -1;
		const int y_offset = bs.fy * 8;

		if (bs.tx == 6) // short castling
		{
			pos1_1 = y_offset + 7; // room
			pos2_1 = y_offset + 5;

			pos1_2 = y_offset + 4; // king
			pos2_2 = y_offset + 6;
		}
		else
		{
			pos1_1 = y_offset + 0; // room
			pos2_1 = y_offset + 3;

			pos1_2 = y_offset + 4; // king
			pos2_2 = y_offset + 2;
		}

		nh.newHash ^= calcZobristDo(bs.p1, pos1_2); // remove king
		nh.newHash ^= calcZobristDo(bs.p1, pos2_2); // put king

		nh.newHash ^= calcZobristDo(bs.p2, pos1_1); // remove rook
		nh.newHash ^= calcZobristDo(bs.p2, pos2_1); // put rook
	}
	else {
		// remove piece from 'from-location'
		int pos1 = bs.fy * 8 + bs.fx;

		ChessPieceType fromType = bs.p1 -> getType();
		PlayerColor fromColor = bs.p1 -> getColor();
		nh.newHash ^= calcZobristDo(pos1, bs.isPromotion ? PAWN : fromType, fromColor);

		int pos2 = bs.ty * 8 + bs.tx;

		// remove piece from destination location
		if (bs.p2) {
			nh.newHash ^= calcZobristDo(bs.p2, pos2);

			if (bs.p2 -> getType() == PAWN)
				nh.pawnNewHash ^= calcZobristDo(bs.p2, pos2);
		}

		// put 'from-piece' on destination location
		if (unlikely(bs.mt == EN_PASSANT)) {
			int putX = bs.tx;
			int putY = c == WHITE ? 2 : 5;

			pos2 = putY * 8 + putX;
		}

		nh.newHash ^= calcZobristDo(bs.p1, pos2);

		if (bs.p1 -> getType() == PAWN) {
			nh.pawnNewHash ^= calcZobristDo(bs.p1, pos1);
			nh.pawnNewHash ^= calcZobristDo(bs.p1, pos2);
		}
		else if (bs.isPromotion) {
			nh.pawnNewHash ^= calcZobristDo(pos1, PAWN, fromColor);
		}
	}

	if (c == WHITE) {
		const uint64_t apply = Random64[780];
		nh.newHash ^= apply;
		nh.cleanup ^= apply;
	}

	uint64_t castlingHash = findCastlingRightsHash(b);
	nh.newHash ^= castlingHash;
	nh.cleanup ^= castlingHash;

	if (isValidMove(epMove))
	{
		// only if b -> getAt(epMove.u.m.tx - 1, ty) == pawn & that color is c or at + 1
		const ChessPiece *const left = epMove.u.m.tx ? b -> getAt(epMove.u.m.tx - 1, epMove.u.m.ty) : NULL;
		const ChessPiece *const right = epMove.u.m.tx < 7 ? b -> getAt(epMove.u.m.tx + 1, epMove.u.m.ty) : NULL;

		if ((left && left -> getType() == PAWN && left -> getColor() == c) || (right && right -> getType() == PAWN && right -> getColor() == c)) {
			const uint64_t apply = Random64[772 + epMove.u.m.tx];
			nh.newHash ^= apply;
			nh.cleanup ^= apply;
		}
	}

	return nh;
}

zobrist_t Tpt::calcZobrist(const Board * const b, const PlayerColor c, const Move epMove)
{
	zobrist_t nh = { 0x00, 0x00, 0x00 };

	int offset = 0;
	for(int y=0; y<8; y++) {
		for(int x=0; x<8; x++, offset++) {
			const ChessPiece *const cur = b -> getAt(x, y);

			// it starts with # empty == # with pieces but it'll
			// get less
			if (unlikely(cur != NULL)) {
				nh.newHash ^= calcZobristDo(cur, offset);

				if (cur -> getType() == PAWN)
					nh.pawnNewHash ^= calcZobristDo(cur, offset);
			}
		}
	}

	if (c == WHITE)
	{
		const uint64_t apply = Random64[780];
		nh.newHash ^= apply;
		nh.cleanup ^= apply;
	}

	uint64_t castlingHash = findCastlingRightsHash(b);
	nh.newHash ^= castlingHash;
	nh.cleanup ^= castlingHash;

	// FIXME move into a function/seperate method
	if (isValidMove(epMove))
	{
		const ChessPiece *const left = epMove.u.m.tx ? b -> getAt(epMove.u.m.tx - 1, epMove.u.m.ty) : NULL;
		const ChessPiece *const right = epMove.u.m.tx < 7 ? b -> getAt(epMove.u.m.tx + 1, epMove.u.m.ty) : NULL;

		if ((left && left -> getType() == PAWN && left -> getColor() == c) || (right && right -> getType() == PAWN && right -> getColor() == c)) {
			const uint64_t apply = Random64[772 + epMove.u.m.tx];
			nh.newHash ^= apply;
			nh.cleanup ^= apply;
		}
	}

	return nh;
}

uint64_t Tpt::calcZobristDo(const int pos, const int type, const PlayerColor c)
{
	my_assert(type < 6, "calcZobristDo type in range");
	my_assert(type < 64, "calcZobristDo pos in range");

	return Random64[(type * 2 + (c == WHITE)) * 64 + pos];
}
uint64_t Tpt::calcZobristDo(const ChessPiece *const obj, const int pos)
{
	return calcZobristDo(pos, obj -> getType(), obj -> getColor());
}

bool Tpt::lookup(const uint64_t hash, tptEntry *const te)
{
	unsigned long int index = hash % nEntries;

	tptEntry *const e = entries[index].entries;

	int bestDepth = -1;
	bool hit = false;

	for(int i=0; i<N_TE_PER_HASH_GROUP; i++)
	{
		tptEntry *const cur = &e[i];

#ifdef THREADING
		if ((cur -> hash ^ cur -> data_.data) == hash && (cur -> data_._data.depth > bestDepth || (cur -> data_._data.depth == bestDepth && cur -> data_._data.age == age))) {
			// FIXME update age
#else
		if (cur -> hash == hash && (cur -> data_._data.depth > bestDepth || (cur -> data_._data.depth == bestDepth && cur -> data_._data.age == age))) {
			cur -> data_._data.age = age;
#endif
			bestDepth = cur -> data_._data.depth;
			te -> data_.data = cur -> data_.data;
			hit = true;
			my_assert(cur -> data_._data.flags != NOTVALID, "tpt::lookup");
			my_assert(cur -> data_._data.value >= -10000, "tpt::lookup");
			my_assert(cur -> data_._data.value <= 10000, "tpt::lookup");
		}
	}

	return hit;
}

void Tpt::store(const uint64_t hash, const tptEntryFlag f, const int d, const int value, const Move::_u_::_m_ & m, const bool useNm)
{
	my_assert(d >= 0, "Tpt::store depth must be >= 0");

	unsigned long int index = hash % nEntries; // FIXME is biased at the bottom of the list

	tptEntry *const e = entries[index].entries;

	int useSubIndex = -1, minDepth = 999, mdi = -1;

        for(int i=0; i<N_TE_PER_HASH_GROUP; i++)
        {
#ifdef THREADING
		if ((e[i].hash ^ e[i].data_.data) == hash) {
#else
		if (e[i].hash == hash) {
#endif
			if (e[i].data_._data.depth > d)
				return;
			if (f!=EXACT && e[i].data_._data.depth==d)
				return;

			useSubIndex = i;

			break;
		}

		if (e[i].data_._data.age != age)
			useSubIndex = i;
		else if (e[i].data_._data.depth < minDepth) {
			minDepth = e[i].data_._data.depth;
			mdi = i;
		}
        }

	if (useSubIndex == -1)
		useSubIndex = mdi;

	my_assert(useSubIndex >= 0 && useSubIndex < N_TE_PER_HASH_GROUP, "Tpt::store");

#ifdef THREADING
	tptEntry *const cur = &e[useSubIndex];
	tptEntry::u n;
	n._data.value = int16_t(value);
	n._data.depth = uint8_t(d);
	n._data.flags = f;
	n._data.age = age;
	n._data.m = m;
	n._data.nm = useNm;

	cur -> hash = hash ^ n.data;
	cur -> data_.data = n.data;
#else
	tptEntry *const cur = &e[useSubIndex];
	cur -> hash = hash;
	cur -> data_._data.value = int16_t(value);
	cur -> data_._data.depth = uint8_t(d);
	cur -> data_._data.flags = f;
	cur -> data_._data.age = age;
	cur -> data_._data.m = m;
	cur -> data_._data.nm = useNm;
#endif
}

void dumpTptEntry(const tptEntry *const p)
{
	printf("%lx (%lx) %d %d %d %d\n", p -> hash, p -> hash ^ p -> data_.data, p -> data_._data.depth, p -> data_._data.flags, p -> data_._data.age, p -> data_._data.nm);
}

int Tpt::hashFullPermill()
{
	int p = 0;

	for(int idx=0; idx<1000/N_TE_PER_HASH_GROUP; idx++) {
		for(int i=0; i<N_TE_PER_HASH_GROUP; i++) {
			if (entries[idx].entries[i].data_._data.age)
				p++;
		}
	}

	return p;
}

void Tpt::verify()
{
	int fail = 0, fail2 = 0, fail3 = 0;

	for(long unsigned int index=0; index<nEntries; index++) {
		tptEntry *const e = entries[index].entries;

		for(int i=0; i<N_TE_PER_HASH_GROUP; i++) {
			tptEntry *const curi = &e[i];
			uint64_t hi = curi -> hash ^ curi -> data_.data;

			for(int j=i + 1; j<N_TE_PER_HASH_GROUP; j++) {
				tptEntry *const curj = &e[j];
				uint64_t hj = curj -> hash ^ curj -> data_.data;

				if (hi == hj && hi != 0) {
					fail2 += curi -> data_._data.depth > curj -> data_._data.depth;
					fail3 += curi -> data_._data.depth == curj -> data_._data.depth && curj -> data_._data.age == age;
					fail++;
					goto next;
				}
			}
		}
next:
		{}
	}

	printf("FAIL COUNT: %d %f\n", fail, fail * 100.0 / nEntries);
	printf("FAIL2COUNT: %d %f\n", fail2, fail2 * 100.0 / nEntries);
	printf("FAIL3COUNT: %d %f\n", fail3, fail3 * 100.0 / nEntries);
}

void Tpt::reset()
{
	memset(entries, 0x00, nEntries * sizeof(tptHashGroup));
}
