// (C) Folkert van Heusden. AGPL v3.0.
// $Revision: 4541 $
#include "Tweaks.h"
#include <pthread.h>
#include <stdio.h>

#include <random>
#include <stdarg.h>
#include <stdlib.h>
#include <string>
#include <vector>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>

#include "config.h"
#include "Exception.h"
#include "PlayerColor.h"
#include "ChessPiece.h"
#include "Scene.h"
#include "IO.h"
#include "error.h"

std::string myformat(const char *const fmt, ...)
{
	char *buffer = NULL;
        va_list ap;

        va_start(ap, fmt);
        (void)vasprintf(&buffer, fmt, ap);
        va_end(ap);

	std::string result = buffer;
	free(buffer);

	return result;
}

ChessPiece *createNewChessPiece(const ChessPieceType cpt, const PlayerColor c, const int x, const int y, const bool isFirstMove)
{
	switch(cpt)
	{
		case PAWN:
			return new ChessPiece(c, x, y, isFirstMove, PAWN);

		case ROOK:
			return new ChessPiece(c, x, y, isFirstMove, ROOK);

		case BISHOP:
			return new ChessPiece(c, x, y, isFirstMove, BISHOP);

		case KNIGHT:
			return new ChessPiece(c, x, y, isFirstMove, KNIGHT);

		case QUEEN:
			return new ChessPiece(c, x, y, isFirstMove, QUEEN);

		case KING:
			return new ChessPiece(c, x, y, isFirstMove, KING);

		case NONE:
			my_assert(false, "createNewChessPiece: \"NONE\" is not a valid type");
			break;

		default:
			my_assert(false, "createNewChessPiece: unknown type");
			break;
	}

	return NULL;
}

std::string genFen(const Scene * const s, const bool shortOutput, const PlayerColor c)
{
	const Board *const b = s -> getBoardConst();

	const Move *const lom = s -> getLastOpponentMove();
	const ChessPiece *const lastMoveObject = lom != NULL ? b -> getAt(lom -> u.m.tx, lom -> u.m.ty) : NULL;
	char output[128];
#ifdef _DEBUG
	memset(output, 0xee, sizeof output);
#endif
	int o_offset = 0;

	for(int y=7; y>=0; y--) {
		int empty = '0';

		for(int x=0; x<8; x++) {
			ChessPiece *const cur = b -> getAt(x, y);

			if (likely(cur == NULL))
				empty++;
			else {  
				if (empty > '0') {
					output[o_offset++] = empty;
					empty = '0';
				}

				output[o_offset++] = cur -> getTypeChar(true);
			}
		}

		if (empty > '0')
			output[o_offset++] = empty;

		if (y != 0)
			output[o_offset++] = '/';
	}

	output[o_offset++] = ' ';
	output[o_offset++] = c == WHITE ? 'w' : 'b';
	output[o_offset++] = ' ';

	const int castle_offset = o_offset;
	if (b -> getAt(4, 0) && b -> getAt(4, 0) -> getFirstMove())
	{
		if (b -> getAt(7, 0) && b -> getAt(7, 0) -> getFirstMove())
			output[o_offset++] = 'K';

		if (b -> getAt(0, 0) && b -> getAt(0, 0) -> getFirstMove())
			output[o_offset++] = 'Q';
	}

	if (b -> getAt(4, 7) && b -> getAt(4, 7) -> getFirstMove())
	{
		if (b -> getAt(7, 7) && b -> getAt(7, 7) -> getFirstMove())
			output[o_offset++] = 'k';

		if (b -> getAt(0, 7) && b -> getAt(0, 7) -> getFirstMove())
			output[o_offset++] = 'q';
	}

	if (castle_offset == o_offset)
		output[o_offset++] = '-';

	output[o_offset++] = ' ';

	if (lastMoveObject && lastMoveObject -> getType() == PAWN)
	{
		output[o_offset++] = 'a' + lom -> u.m.tx;
		output[o_offset++] = lom -> u.m.ty == 3 ? '3' : '6';
	}
	else
	{
		output[o_offset++] = '-';
	}

	if (!shortOutput)
		o_offset += sprintf(&output[o_offset], " %d %d", s -> getNHalfMoves(), s -> getNFullMoves());

	return std::string(output, o_offset);
}

uint64_t get_ts_ms()
{
	struct timeval tv = { 0, 0 };

	gettimeofday(&tv, NULL);

	return tv.tv_sec * 1000 + tv.tv_usec / 1000;
}

std::string trim(std::string in)
{
	size_t len = in.size();
	size_t index=0;
	while(index<len && isspace(in.at(index)))
		index++;
	if (index == len)
		return "";
	size_t start = index;

	size_t end = len - 1;
	while(end > start && isspace(in.at(index)))
		end--;

	return in.substr(start, end - start + 1);
}

std::string replace(std::string in, char from, char to)
{
	std::string out = in;
	char to_str[] = { to, 0x00 };

	size_t len = out.size();
	for(size_t index=0; index<len; index++)
	{
		if (out.at(index) == from)
			out.replace(index, 1, to_str, 1);
	}

	return out;
}

std::vector<std::string> * split(std::string in, std::string splitter)
{
	std::vector<std::string> *out = new std::vector<std::string>;
	size_t splitter_size = splitter.size();

	for(;;)
	{
		size_t pos = in.find(splitter);
		if (pos == std::string::npos)
			break;

		std::string before = in.substr(0, pos);
		out -> push_back(before);

		size_t bytes_left = in.size() - (pos + splitter_size);
		if (bytes_left == 0)
		{
			out -> push_back("");
			return out;
		}

		in = in.substr(pos + splitter_size);
	}

	if (in.size() > 0)
		out -> push_back(in);

	return out;
}

ssize_t read_retry(int fd, void *data, size_t len)
{
        char *buf = static_cast<char *>(data);
        size_t nread = 0;

        while (len > 0) {
                ssize_t ret = read(fd, buf, len);

                if (ret == -1)
		{
			if (errno == EINTR || errno == EAGAIN)
				continue;

                        return ret;
		}

                if (ret == 0)
                        break;

                len -= ret;
                buf += ret;
                nread += ret;
        }

        return nread;
}

uint64_t genRnd()
{
	std::random_device rd;
	std::uniform_int_distribution<uint8_t> dist(0, 255);

	uint64_t temp;
	uint8_t *p = (uint8_t *)&temp;
	for(int i=0; i<8; i++)
		p[i] = dist(rd);

	return temp;
}

// http://stackoverflow.com/questions/466204/rounding-up-to-nearest-power-of-2
unsigned int upper_power_of_two(unsigned int v)
{
	v--;
	v |= v >> 1;
	v |= v >> 2;
	v |= v >> 4;
	v |= v >> 8;
	v |= v >> 16;
	v++;

	return v;
}

// http://stackoverflow.com/questions/14409466/simple-hash-functions
uint32_t adler32(const void *const buf, const size_t buflength)
{
	const uint8_t *buffer = (const uint8_t*)buf;

	uint32_t s1 = 1;
	uint32_t s2 = 0;

	for(size_t n = 0; n < buflength; n++)
	{
		s1 = (s1 + buffer[n]) % 65521;
		s2 = (s2 + s1) % 65521;
	}     

	return (s2 << 16) | s1;
}

static ChessPieceType CharToChessObjectType(const char objectChar)
{
	switch(objectChar)
	{
		case 'P':
		case 'p':
			return PAWN;

		case 'B':
		case 'b':
			return BISHOP;

		case 'R':
		case 'r':
			return ROOK;

		case 'N':
		case 'n':
			return KNIGHT;

		case 'Q':
		case 'q':
			return QUEEN;

		case 'K':
		case 'k':
			return KING;

		default:
			return NONE;
	}
}

bool SANToMove(const std::string & sanMove, const Scene & s, const PlayerColor myColor, const VectorType<Move> *const validMoves, Move *const m)
{
	std::string smwork = sanMove;
	ChessPieceType type = PAWN, promoteTo = NONE;
	uint8_t baseLine = myColor == WHITE ? 0 : 7;
	std::string from, to;
	size_t x = std::string::npos;
	int fx = -1, fy = -1, tx = -1, ty = -1;

	/* '+' at the end means check */
	size_t garbageIndex = smwork.find('+');
	if (garbageIndex != std::string::npos)
		smwork = smwork.substr(0, garbageIndex);

	/* '#' at the end means check-mate */
	garbageIndex = smwork.find('#');
	if (garbageIndex != std::string::npos)
		smwork = smwork.substr(0, garbageIndex);

	/* o-o/o-o-o (castling) */
	if (strcasecmp(smwork.c_str(), "o-o") == 0) {
		// FIXME: check if castling is allowed
		m -> u.m.fx = 4;
		m -> u.m.tx = 6;
		m -> u.m.fy = m -> u.m.ty = baseLine;
		m -> u.m.mt = CASTLING;
		m -> u.m.promoteTo = NONE;
		return true;
	}
	else if (strcasecmp(smwork.c_str(), "o-o-o") == 0) {
		// FIXME: check if castling is allowed
		m -> u.m.fx = 4;
		m -> u.m.tx = 2;
		m -> u.m.fy = m -> u.m.ty = baseLine;
		m -> u.m.mt = CASTLING;
		m -> u.m.promoteTo = NONE;
		return true;
	}

	// do not crash on valid moves
	if (smwork.empty())
		return false;

	/* the object-type is in the first character of the string,
	 * it can be recognized by that it is an uppercase character.
	 * if it is not there (the first char is lowercase), it is
	 * a pawn
	 */
	if (isupper(smwork.at(0))) {
		type = CharToChessObjectType(smwork.at(0));

		if (type == NONE)
			return false;

		smwork = smwork.substr(1);

		if (smwork.empty())
			return false;
	}

	/* if the object type is a pawn, there might be a '=Q/B/N/R'
	 * behind the string indicating the object it promotes to
	 */
	if (type == PAWN) {
		size_t is = smwork.find('=');

		if (is != std::string::npos) {
			promoteTo = CharToChessObjectType(smwork.at(is + 1));

			if (promoteTo == NONE)
				return false;

			smwork = smwork.substr(0, is);
		}
		else {
			size_t lastCharIndex = smwork.length() - 1;
			char lastCharacter = smwork.at(lastCharIndex);

			if (isupper(lastCharacter)) {
				promoteTo = CharToChessObjectType(lastCharacter);

				if (promoteTo == NONE)
					return false;

				smwork = smwork.substr(0, lastCharIndex);
			}
		}

		if (smwork.empty())
			return false;
	}

	/* check for capture move & split move into from/to string */
	x = smwork.find('x');
	if (x == std::string::npos)
		x = smwork.find('X');

	if (x != std::string::npos) {
		from = smwork.substr(0, x);
		to = smwork.substr(x + 1);
	}
	else {
		size_t moveLen = smwork.length();

		if (moveLen > 2) {
			from = smwork.substr(0, moveLen - 2);
			to = smwork.substr(moveLen - 2);
		}
		else {
			from.clear();
			to = smwork;
		}
	}

	/* convert to-string to a position */
	tx = to.at(0) - 'a';
	ty = to.at(1) - '1';

	/* convert from-string, if available */
	if (from.empty()) {
		fx = fy = -1;
	}
	else if (from.length() == 2) {
		fx = from.at(0) - 'a';
		fy = from.at(1) - '1';
	}
	else if (from.length() == 1) {
		char fromChar = from.at(0);

		if (isdigit(fromChar)) {
			fx = -1;
			fy = fromChar - '1';
		}
		else {
			fx = fromChar - 'a';
			fy = -1;
		}
	}

	const Board *const b = s.getBoardConst();

	/* now find the move in the movelist */
	for(size_t i=0; i<validMoves -> size(); i++)
	{
		const Move currentMove = validMoves -> at(i);

		if (currentMove.u.m.tx != tx || currentMove.u.m.ty != ty)
			continue;

		if (fx != -1 && currentMove.u.m.fx != fx) 
			continue;

		if (fy != -1 && currentMove.u.m.fy != fy)
			continue;

		// determing the type is tricky
		if (b -> getAt(currentMove.u.m.fx, currentMove.u.m.fy) -> getType() != type)
			continue;

		if (currentMove.u.m.promoteTo != promoteTo)
			continue;

		*m = currentMove;

		return true;
	}

	return false;
}

void * getCPUAffinity()
{
#if defined(__CYGWIN__) || defined(__APPLE__) || defined(ANDROID)
	return NULL;
#else
	cpu_set_t *set = new cpu_set_t;

	pthread_getaffinity_np(pthread_self(), sizeof(cpu_set_t), set);

	return set;
#endif
}

void setCPUAffinity(const int i)
{
#if !defined(__CYGWIN__) && !defined(__APPLE__) && !defined(ANDROID)
	cpu_set_t set;
	CPU_ZERO(&set);
	CPU_SET(i, &set);

	pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t), &set);
#endif
}

void resetCPUAffinity(void *p)
{
#if !defined(__CYGWIN__) && !defined(__APPLE__) && !defined(ANDROID)
	cpu_set_t *set = (cpu_set_t *)p;

	pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t), set);

	delete set;
#endif
}

rw_lock *alloc_rw_lock()
{
#if defined(USE_PTHREAD_RW_LOCK) || defined(__APPLE__) || defined(NO_SHARED_MUTEX)
	pthread_rwlock_t *lock = new pthread_rwlock_t;
	pthread_rwlock_init(lock, NULL);
	return lock;
#else
	return new std::shared_timed_mutex();
#endif
}

void free_rw_lock(rw_lock *const l)
{
#if defined(USE_PTHREAD_RW_LOCK) || defined(__APPLE__) || defined(NO_SHARED_MUTEX)
	pthread_rwlock_destroy(l);
#endif
	delete l;
}

std::string mytolower(const std::string & in)
{
	size_t len = in.size();

	std::string out(len, ' ');

	for(size_t index=0; index<len; index++)
		out[index] = char(tolower(in[index]));

	return out;
}

bool yesOrNo(const std::string & in)
{
	std::string inu = mytolower(in);

	return inu == "yes" || inu == "true";
}

uint64_t int_from_mem(const uint8_t *p, int l)
{
	uint64_t t = 0;

	while(l--) {
		t <<= 8;
		t |= *p++;
	}

	return t;
}

std::string * getLineFromSTDIN() {
	char buffer[4096] = { 0 };

	if (!fgets(buffer, sizeof buffer, stdin))
		return NULL;

	char *cr = strchr(buffer, '\r');
	if (cr)
		*cr = 0x00;

	char *lf = strchr(buffer, '\n');
	if (lf)
		*lf = 0x00;

	std::string out = trim(std::string(buffer));

	return new std::string(out);
}

int countBits1(const uint32_t in) 
{
	return __builtin_popcount(in);
}

int WRITE(const int fd, const uint8_t *p, size_t l)
{
	int n = 0;

	while(l) {
		int rc = write(fd, p, l);

		if (rc <= 0)
			return rc;

		p += rc;
		l -= rc;
		n += rc;
	}

	return n;
}

int start_listen(int portnr)
{
	struct sockaddr_in6 server_addr;
	int server_addr_len = 0;
	int fd = socket(AF_INET6, SOCK_STREAM, 0);
	if (fd == -1)
		error_exit(true, "failed creating socket");

	int reuse_addr = 1;
	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *)&reuse_addr, sizeof(reuse_addr)) == -1)
		error_exit(true, "setsockopt(SO_REUSEADDR) failed");

#ifdef TCP_FASTOPEN
	int qlen = SOMAXCONN;
	setsockopt(fd, SOL_TCP, TCP_FASTOPEN, &qlen, sizeof qlen);
#endif

	server_addr_len = sizeof(server_addr);
	memset((char *)&server_addr, 0x00, server_addr_len);
	server_addr.sin6_family = AF_INET6;
	server_addr.sin6_port = htons(portnr);
	server_addr.sin6_addr = in6addr_any;

	if (bind(fd, (struct sockaddr *)&server_addr, server_addr_len) == -1)
		error_exit(true, "bind() failed");

	if (listen(fd, SOMAXCONN) == -1)
		error_exit(true, "listen() failed");

	return fd;
}

std::string get_endpoint_name(int fd)
{
	char buffer[4096] = { "?" };
	struct sockaddr_in6 addr;
	socklen_t addr_len = sizeof addr;

	if (getpeername(fd, (struct sockaddr *)&addr, &addr_len) == -1)
		snprintf(buffer, sizeof buffer, "[FAILED TO FIND NAME OF %d: %s (1)]", fd, strerror(errno));
	else
	{
		char buffer2[4096];

		if (inet_ntop(AF_INET6, &addr.sin6_addr, buffer2, sizeof buffer2))
			snprintf(buffer, sizeof buffer, "[%s]:%d", buffer2, ntohs(addr.sin6_port));
		else
			snprintf(buffer, sizeof buffer, "[FAILED TO FIND NAME OF %d: %s (1)]", fd, strerror(errno));
	}

	return std::string(buffer);
}
