/*
 * fathom.c
 * (C) 2015 basil, all rights reserved,
 *
 * adjusted by folkert@vanheusden.com into a library
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <assert.h>
#include <getopt.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

extern "C" {
#include "tbprobe.h"
}
#include "fathom.h"

#define NAME "SYZYGY-ETB"

#define BOARD_RANK_1            0x00000000000000FFull
#define BOARD_FILE_A            0x8080808080808080ull
#define square(r, f)            (8 * (r) + (f))
#define rank(s)                 ((s) >> 3)
#define file(s)                 ((s) & 0x07)
#define board(s)                ((uint64_t)1 << (s))

#ifdef __GNUG__
        #define likely(x)       __builtin_expect((x), 1)
        #define unlikely(x)     __builtin_expect((x), 0)
	#define __PRAGMA_PACKED__ __attribute__ ((__packed__))
#else
        #define likely(x) (x)
        #define unlikely(x) (x)
	#define __PRAGMA_PACKED__
#endif

#if 0
static const char *wdl_to_str[5] =
{
	"0-1",
	"1/2-1/2",
	"1/2-1/2",
	"1/2-1/2",
	"1-0"
};
#endif

struct pos
{
	uint64_t white;
	uint64_t black;
	uint64_t kings;
	uint64_t queens;
	uint64_t rooks;
	uint64_t bishops;
	uint64_t knights;
	uint64_t pawns;
	uint8_t castling;
	uint8_t rule50;
	uint8_t ep;
	bool turn;
	uint16_t move;
};

/*
 * Parse a FEN string.
 */
#define CASE_TO_INDEX(c) ((c & 32) >> 5)

static bool parse_FEN(struct pos *pos, const char *fen)
{
	uint64_t col[2] = { 0, 0 };
	uint64_t kings, queens, rooks, bishops, knights, pawns;
	kings = queens = rooks = bishops = knights = pawns = 0;
	bool turn;
	unsigned rule50 = 0, move = 1;
	unsigned ep = 0;
	unsigned castling = 0;
	char c;
	int r, f;
	uint64_t *lu[256];

	lu[(int)'K'] = lu[(int)'k'] = &kings;
	lu[(int)'Q'] = lu[(int)'q'] = &queens;
	lu[(int)'R'] = lu[(int)'r'] = &rooks;
	lu[(int)'B'] = lu[(int)'b'] = &bishops;
	lu[(int)'N'] = lu[(int)'n'] = &knights;
	lu[(int)'P'] = lu[(int)'p'] = &pawns;

	if (fen == NULL)
		goto fen_parse_error;

	for (r = 7; r >= 0; r--)
	{
		for (f = 0; f <= 7; f++, fen++)
		{
			unsigned s = (r * 8) + f;
			uint64_t b = board(s);
			const char c = *fen;
			if (likely(c <= '9'))
				f += c - '0' - 1;
			else
			{
				col[CASE_TO_INDEX(c)] |= b;
				(*lu[(int)c]) |= b;
			}
		}
		if (r == 0)
			break;
		if (unlikely(*fen++ != '/'))
			goto fen_parse_error;
	}
	c = *fen++;
	if (unlikely(c != ' '))
		goto fen_parse_error;
	c = *fen++;
	if (unlikely(c != 'w' && c != 'b'))
		goto fen_parse_error;
	turn = (c == 'w');
	c = *fen++;
	if (unlikely(c != ' '))
		goto fen_parse_error;
	c = *fen++;
	if (c != '-')
	{
		do
		{
			switch (c)
			{
				case 'K':
					castling |= TB_CASTLING_K; break;
				case 'Q':
					castling |= TB_CASTLING_Q; break;
				case 'k':
					castling |= TB_CASTLING_k; break;
				case 'q':
					castling |= TB_CASTLING_q; break;
				default:
					goto fen_parse_error;
			}
			c = *fen++;
		}
		while (c != ' ');
		fen--;
	}
	c = *fen++;
	if (unlikely(c != ' '))
		goto fen_parse_error;
	c = *fen++;
	if (c >= 'a' && c <= 'h')
	{
		unsigned file = c - 'a';
		c = *fen++;
		if (c != '3' && c != '6')
			goto fen_parse_error;
		unsigned rank = c - '1';
		ep = square(rank, file);
		if (rank == 2 && turn)
			goto fen_parse_error;
		if (rank == 5 && !turn)
			goto fen_parse_error;
		if (rank == 2 && ((tb_pawn_attacks(ep, true) & (col[CASE_TO_INDEX('k')] & pawns)) == 0)) /* black */
			ep = 0;
		if (rank == 5 && ((tb_pawn_attacks(ep, false) & (col[CASE_TO_INDEX('K')] & pawns)) == 0)) /* white */
			ep = 0;
	}
	else if (unlikely(c != '-'))
		goto fen_parse_error;
	c = *fen++;
	if (unlikely(c != ' '))
		goto fen_parse_error;
	char clk[4];
	clk[0] = *fen++;
	if (unlikely(clk[0] < '0' || clk[0] > '9'))
		goto fen_parse_error;
	clk[1] = *fen++;
	if (clk[1] != ' ')
	{
		if (unlikely(clk[1] < '0' || clk[1] > '9'))
			goto fen_parse_error;
		clk[2] = *fen++;
		if (clk[2] != ' ')
		{
			if (unlikely(clk[2] < '0' || clk[2] > '9'))
				goto fen_parse_error;
			c = *fen++;
			if (unlikely(c != ' '))
				goto fen_parse_error;
			clk[3] = '\0';
		}
		else
			clk[2] = '\0';
	}
	else
		clk[1] = '\0';
	rule50 = atoi(clk);
	move = atoi(fen);

	pos->white = col[CASE_TO_INDEX('K')];
	pos->black = col[CASE_TO_INDEX('k')];
	pos->kings = kings;
	pos->queens = queens;
	pos->rooks = rooks;
	pos->bishops = bishops;
	pos->knights = knights;
	pos->pawns = pawns;
	pos->castling = castling;
	pos->rule50 = rule50;
	pos->ep = ep;
	pos->turn = turn;
	pos->move = move;
	return true;

fen_parse_error:
	return false;
}

/*
 * Convert a move into a string.
 */
static char * move_to_str(const struct pos *const pos, const unsigned move)
{
	char *out = (char *)malloc(6), *str = out;
	unsigned from     = TB_GET_FROM(move);
	unsigned to       = TB_GET_TO(move);
	unsigned promotes = TB_GET_PROMOTES(move);

	*str++ = 'a' + file(from);
	*str++ = '1' + rank(from);
	*str++ = 'a' + file(to);
	*str++ = '1' + rank(to);

	if (promotes != TB_PROMOTES_NONE)
	{
		switch (promotes)
		{
			case TB_PROMOTES_QUEEN:
				*str++ = 'Q'; break;
			case TB_PROMOTES_ROOK:
				*str++ = 'R'; break;
			case TB_PROMOTES_BISHOP:
				*str++ = 'B'; break;
			case TB_PROMOTES_KNIGHT:
				*str++ = 'N'; break;
		}
	}

	*str++ = '\0';

	return out;
}

static bool se_valid = false;

bool init_syzygy_etb(const char *const path)
{
	if (se_valid) // alread initialized
		return false;

	tb_init(path);

	if (TB_LARGEST == 0)
	{
		fprintf(stderr, "Failed loading syzygy-etb from %s\n", path);
		return false;
	}

	se_valid = true;

	return true;
}

int get_syzygy_etb_piece_count()
{
	return TB_LARGEST;
}

pthread_mutex_t se_lock = PTHREAD_MUTEX_INITIALIZER;

std::vector<std::string> select_a_move(const struct pos *const pos, const unsigned *const results, const int wdl)
{
	std::vector<std::string> out;

	for(unsigned i = 0; results[i] != TB_RESULT_FAILED; i++)
	{
		if (TB_GET_WDL(results[i]) < wdl)
			continue;

		char *s = move_to_str(pos, results[i]);
		out.push_back(s);
		free(s);
	}

	return out;
}

void query_syzygy_etb(const char *const fen, syzygy_result_t *const sr, std::vector<std::string> *const moves, int *const depth, const bool nonblock, const bool quick)
{
	moves -> clear();

	*sr = se_fail;

	if (pthread_mutex_trylock(&se_lock))
		return;

	for(;;)
	{
		if (!se_valid)
			break;

		// (1) parse the FEN:
		struct pos pos0;
		struct pos *pos = &pos0;
		if (!parse_FEN(pos, fen))
		{
			fprintf(stderr, NAME " error: unable to parse FEN string \"%s\"\n", fen);
exit(1);
			break;
		}

		// (2) probe the TB:
		unsigned int tpc = tb_pop_count(pos->white | pos->black);
		if (tpc > TB_LARGEST)
		{
			fprintf(stderr, NAME " error: unable to probe tablebase; FEN string \"%s\" has too many pieces (max=%u)\n", fen, TB_LARGEST);
exit(1);
			break;
		}

		unsigned results[TB_MAX_MOVES], res = -1;
		if (quick)
			res = tb_probe_wdl(pos->white, pos->black, pos->kings, pos->queens, pos->rooks, pos->bishops, pos->knights, pos->pawns, pos->rule50, pos->castling, pos->ep, pos->turn);
		else
			res = tb_probe_root(pos->white, pos->black, pos->kings, pos->queens, pos->rooks, pos->bishops, pos->knights, pos->pawns, pos->rule50, pos->castling, pos->ep, pos->turn, results);
		if (res == TB_RESULT_FAILED)
		{
#ifdef _DEBUG
// FIXME		fprintf(stderr, NAME " error: unable to probe tablebase; position invalid, illegal or not in tablebase\n");
#endif
			break;
		}

		// (3) Output:
		unsigned wdl = TB_GET_WDL(res);
		if (quick) {
			*depth = 0;
		}
		else
		{
			*depth = abs(int(TB_GET_DTZ(res)));

			*moves = select_a_move(pos, results, TB_CURSED_WIN);

			if (moves -> empty())
				*moves = select_a_move(pos, results, TB_DRAW);
		}

		if (wdl == TB_WIN || wdl == TB_CURSED_WIN)
		{
			*sr = se_win;
			break;
		}

		if (wdl == TB_DRAW)
		{
			*sr = se_draw;
			break;
		}

		if (wdl == TB_LOSS || wdl == TB_BLESSED_LOSS)
		{
			*sr = se_lost;
			break;
		}

		fprintf(stderr, "Unexpected return-code from table base: %d\n", wdl);

		break;
	}

	pthread_mutex_unlock(&se_lock);
}
