#include "ws.h"
#ifdef WSJS
#include <fstream>
#include <stdio.h>
#include "jansson.h"
#include "Utils.h"

void add_json_key_value(json_t *const target, const std::string & key, const std::string & value)
{
	json_t *jval = json_string(value.c_str());

	json_object_set_new(target, key.c_str(), jval);
}

void add_json_key_value(json_t *const target, const std::string & key, const long int value)
{
	json_t *jval = json_string(myformat("%ld", value).c_str());

	json_object_set_new(target, key.c_str(), jval);
}

void add_json_key_value(json_t *const target, const std::string & key, const ULI value)
{
	json_t *jval = json_string(myformat("%lu", value).c_str());

	json_object_set_new(target, key.c_str(), jval);
}

void add_json_key_value(json_t *const target, const std::string & key, const double value)
{
	json_t *jval = json_string(myformat("%f", value).c_str());

	json_object_set_new(target, key.c_str(), jval);
}

wsserver::wsserver(const int portIn) : port(portIn)
{
	// set up access channels to only log interesting things
	m_endpoint.clear_access_channels(websocketpp::log::alevel::all);
//	m_endpoint.set_access_channels(websocketpp::log::alevel::access_core);
//	m_endpoint.set_access_channels(websocketpp::log::alevel::app);

	// Initialize the Asio transport policy
	m_endpoint.init_asio();

	// Bind the handlers we are using
	using websocketpp::lib::placeholders::_1;
	using websocketpp::lib::bind;
	m_endpoint.set_open_handler(bind(&wsserver::on_open, this, _1));
	m_endpoint.set_close_handler(bind(&wsserver::on_close, this, _1));
	m_endpoint.set_http_handler(bind(&wsserver::on_http, this, _1));

	ws_thread_handle = new std::thread(std::ref(*this));

	stop_flag = false;
	worker = new std::thread(&wsserver::workerMethod, this);
}

wsserver::~wsserver()
{
	log(ll_ok, "Shutting down WebSockets");

	stop_flag = true;
	update("", false);

	m_timer -> cancel();

	m_endpoint.stop();
	// FIXME check ec

	con_lock.lock();
	for(con_list::iterator it = m_connections.begin(); it != m_connections.end(); ++it) {
		m_endpoint.pause_reading(*it);
		m_endpoint.close(*it, websocketpp::close::status::going_away, "");
	}
	con_lock.unlock();

	ws_thread_handle -> join();
	delete ws_thread_handle;

	worker -> join();
	delete worker;

	log(ll_ok, "WebSockets shut down");
}

void wsserver::workerMethod()
{
	log(ll_ok, "wsserver::workerMethod started");

	for(;!stop_flag;) {
		std::string json = work.get();

		try {
			con_lock.lock();
			for(con_list::iterator it = m_connections.begin(); it != m_connections.end(); ++it)
				m_endpoint.send(*it, json, websocketpp::frame::opcode::text);
			con_lock.unlock();
		}
		catch (websocketpp::exception const & e) {
			log(ll_err, "WebSockets (wsserver::update): %s", e.what());
		}
	}

	log(ll_ok, "wsserver::workerMethod terminating");
}

void wsserver::set_timer()
{
	m_timer = m_endpoint.set_timer(2500,
		websocketpp::lib::bind(
				&wsserver::on_timer,
				this,
				websocketpp::lib::placeholders::_1
			)
		);
}

void wsserver::on_timer(websocketpp::lib::error_code const & ec)
{
	json_t *j = json_object();

	add_json_key_value(j, "type", "ping");
	add_json_key_value(j, "n_watchers", (ULI)getNWatchers());

	update(json_dumps(j, 0), false);

	json_decref(j);

	set_timer();
}

void wsserver::operator()()
{
	log(ll_ok, "WebSockets endpoint started, listening on %d", port);

	m_endpoint.set_reuse_addr(true);
	m_endpoint.listen(port);

	m_endpoint.start_accept();

	set_timer();

	try {
		m_endpoint.run();
	}
	catch (websocketpp::exception const & e) {
		log(ll_err, "WebSockets (wsserver::operator()): %s", e.what());
	}

	log(ll_ok, "WebSockets endpoint terminating");
}

void wsserver::update(const std::string & json, const bool remember)
{
	if (remember)
		latest_json = json;

	work.put(json);
}

void wsserver::on_http(connection_hdl hdl)
{
	server::connection_ptr con = m_endpoint.get_con_from_hdl(hdl);

	std::string filename = con->get_resource();
	if (filename.empty())
		return;

	con->set_body("nope");
}

void wsserver::on_message(server* s, websocketpp::connection_hdl hdl, message_ptr msg)
{
}

size_t wsserver::getNWatchers()
{
	con_lock.lock();
	size_t n = m_connections.size();
	con_lock.unlock();

	return n;
}

void wsserver::on_open(connection_hdl hdl)
{
	server::connection_ptr con = m_endpoint.get_con_from_hdl(hdl);
	log(ll_ok, "WebSockets connected with %s", con -> get_remote_endpoint().c_str());

	con_lock.lock();
	m_connections.insert(hdl);
	con_lock.unlock();

	if (!def.empty())
		m_endpoint.send(hdl, def, websocketpp::frame::opcode::text);

	if (!latest_json.empty())
		m_endpoint.send(hdl, latest_json, websocketpp::frame::opcode::text);
}

void wsserver::on_close(connection_hdl hdl)
{
	con_lock.lock();
	m_connections.erase(hdl);
	con_lock.unlock();

	server::connection_ptr con = m_endpoint.get_con_from_hdl(hdl);
	log(ll_ok, "WebSockets connected with %s closed", con -> get_remote_endpoint().c_str());
}

#endif
